return {
 ["hashes"]={
  { "fontloader-2023-12-28.lua", "f33a8946b4d6a005e76276706db83059" },
  { "fontloader-util-fil.lua", "d0accf4180201efd567816a25ad5e037" },
  { "fontloader-util-str.lua", "f9cf43057eb25bbf466821908686fc3d" },
  { "fontloader-l-unicode.lua", "f368874f11f3c525de5864fd413e4b24" },
  { "fontloader-l-math.lua", "e19c80b7e700a69df942d4df2e14aa28" },
  { "fontloader-l-boolean.lua", "abe28515dd33e8f6c416c09bca351cf8" },
  { "fontloader-l-file.lua", "525b1c6af6f94d93ae929bfbe980d656" },
  { "fontloader-l-io.lua", "3396a8644bae37e49b2cfc0c388c78f8" },
  { "fontloader-l-table.lua", "baa3aea43dbc57e6fb5ff7850129218c" },
  { "fontloader-l-string.lua", "3e97812b1d8eff26f19837ab00813e71" },
  { "fontloader-l-function.lua", "a7e68a9703c35238729da41a474e951b" },
  { "fontloader-l-lpeg.lua", "8315a56e6bc7b2ede3f70c49bf2b5af5" },
  { "fontloader-l-lua.lua", "420bc442e6d81f4624728a6c56fcd175" },
  { "fontloader-fonts-gbn.lua", "34e4543a02f6fbc8c2ade896cb6dc7df" },
  { "fontloader-fonts-lig.lua", "fbd251eea3810a43a8d5542319361d68" },
  { "fontloader-fonts-ext.lua", "32013cbc5d5d336be8b1d1e5879d86c4" },
  { "fontloader-fonts-def.lua", "f435e0875f203f343157baeff876ec9c" },
  { "fontloader-fonts-tfm.lua", "f0eb5e2a8068b17ad401bb7efdba1630" },
  { "fontloader-fonts-enc.lua", "c24fd6b4f34f87b9c4ff5e6193ce2656" },
  { "fontloader-fonts-mis.lua", "17e967c9ec4d001deefd43ddf25e98f7" },
  { "fontloader-basics-chr.lua", "ce4afac7ba19471f3c8f90800839f315" },
  { "fontloader-basics-nod.lua", "1750af43148def9e7a75c6900a96f9dc" },
  { "fontloader-font-imp-effects.lua", "d05997c1826355548ec9aec1346f9f23" },
  { "fontloader-font-imp-italics.lua", "3cc555a1fe00dc45449f1e4a47a5c990" },
  { "fontloader-font-imp-ligatures.lua", "250f6f4488ee770c23177afb2ff75676" },
  { "fontloader-font-imp-tex.lua", "9882d382700e714cace08b7f09820958" },
  { "fontloader-font-shp.lua", "a174ac26390bd0828805747500df0077" },
  { "fontloader-font-def.lua", "cec0e49aee881c4cbd36d541f0162c6f" },
  { "fontloader-font-lua.lua", "1fbfdf7b689b2bdfd0e3bb9bf74ce136" },
  { "fontloader-font-afk.lua", "b7bef80aa4c18357cbfa0ae4df4d4046" },
  { "fontloader-font-one.lua", "b7fb0389462a49bc1f41bebd2da85401" },
  { "fontloader-font-onr.lua", "a3f85c886cf22fb27760d98603d51ffd" },
  { "fontloader-font-ocl.lua", "2a377351ae23c339764ec8ce2e8e9164" },
  { "fontloader-font-osd.lua", "837039b0dca88edd3f4b54222f7cf54c" },
  { "fontloader-font-otc.lua", "0ad931eec679ebeceed5d4f02f4ba31b" },
  { "fontloader-font-ots.lua", "2ba92e909a45b68592f7271d8190f390" },
  { "fontloader-font-ota.lua", "7407b709ae7b2cdf36f68d5c98d51d12" },
  { "fontloader-font-oup.lua", "aaff9c28233bbea56e6d27ae50e04ff0" },
  { "fontloader-font-otj.lua", "b2c165bd6771a57633f992d07384b739" },
  { "fontloader-font-oto.lua", "b148d66f642aaa3d0ddc427f71d6e02d" },
  { "fontloader-font-otl.lua", "84499d355f82829841b8e304952a9d5d" },
  { "fontloader-font-ott.lua", "427bdc77dd187158eafb79b7193400dc" },
  { "fontloader-font-oti.lua", "309a75f9c14b77d87e94eba827dc4e71" },
  { "fontloader-font-dsp.lua", "d07a0663bafeb7603c5456639afaf2e8" },
  { "fontloader-font-ttf.lua", "ff94e130d7c71edd899c4bae7745b7db" },
  { "fontloader-font-cff.lua", "999da544079698ce70178fd84bcdd67a" },
  { "fontloader-font-otr.lua", "2958a21bebcf0579f4e11f0646d1b402" },
  { "fontloader-font-vfc.lua", "a81c29eda98cd62cbebdb6c93544b50d" },
  { "fontloader-font-map.lua", "b1444428c238a17d2a998ff1a4769183" },
  { "fontloader-font-cid.lua", "22b0367742fb253deef84ef7ccf5e8de" },
  { "fontloader-font-con.lua", "ebdd0ca857deafdc0152c7a4f1ed2490" },
  { "fontloader-font-ini.lua", "16b855c98602e4a3e9df39512916a427" },
  { "fontloader-data-con.lua", "b4ab51aaeab872d86d827f3e216bf05e" },
  { "fontloader-reference.lua", "23d03b682af6417aa08995f2ec10de22" },
  { "fontloader-basics-gen.lua", "578c251c496bed783075709e94c2be26" },
  { "luaotfload-glyphlist.lua", "beb741fb48fc25ceb0edfed023c06a94" },
  { "luaotfload-characters.lua", "76415ca3ede312cad6e09038323489ae" },
  { "luaotfload-dvi.lua", "c21eaeeaf37d6fb5b01709a36fbc9c43" },
  { "luaotfload-tounicode.lua", "21fe79d2526f3c3d789a9cc0366347ab" },
  { "luaotfload-unicode.lua", "4272855a25780b80472f3b58edd95e33" },
  { "luaotfload-resolvers.lua", "4f8521dc226d08512950d1d5e8aab8c2" },
  { "luaotfload-parsers.lua", "2e9a5d75a9ee0e5f9a1e1c8323cd2de0" },
  { "luaotfload-fallback.lua", "e05de4261a8752342848719813588180" },
  { "luaotfload-szss.lua", "e251628fef2f987a01578e9cc4c82588" },
  { "luaotfload-scripts.lua", "6327dd2af89cf4683693118b025b3eaa" },
  { "luaotfload-multiscript.lua", "43be714858bc37c09930d53a42882afa" },
  { "luaotfload-loaders.lua", "2f1994b70766eb36ef058a1c9b6b68b6" },
  { "luaotfload-harf-plug.lua", "6257d6011bc38cd0b21efba6f7385ef4" },
  { "luaotfload-harf-define.lua", "bfd8d8d2badb2ccd8a4a06fcbfafa74c" },
  { "luaotfload-notdef.lua", "fa3fca3ed4bd42155506aa25b43e28e8" },
  { "luaotfload-embolden.lua", "805f2fb9deea6465e6a234e8f8484b3c" },
  { "luaotfload-letterspace.lua", "74844a364041657273a550d05b195eaa" },
  { "luaotfload-features.lua", "53438c93b9fd5b6ed6dcf18f6489416a" },
  { "luaotfload-database.lua", "54d2b1150ded4b0a3a6b2f6f172be9aa" },
  { "luaotfload-configuration.lua", "c077eeb67481e5156cc70d6f5731637b" },
  { "luaotfload-colors.lua", "4e7012beae70afc8513d724e94d889f6" },
  { "luaotfload-auxiliary.lua", "08a4e0dae8c3826067aa833dde7ed899" },
  { "luaotfload-filelist.lua", "b2172cbee288902cc4944c0160bc9a02" },
  { "luaotfload-tool.lua", "e4be97c97c71d3f0d55b02188453df85" },
  { "luaotfload-diagnostics.lua", "664bb25ab022856ba173de7cc058848b" },
  { "luaotfload-log.lua", "743199ad7313baa02f88c938b567db1a" },
  { "luaotfload-init.lua", "f30eaefe7631651281cecba90b912e64" },
  { "luaotfload-main.lua", "b19c9cc34cf1d676c39f872cfb41aef6" },
 },
 ["notes"]={
  ["committer"]="Marcel Fabian Krüger <tex@2krueger.de>",
  ["description"]="v3.29/2024-12-03",
  ["loader"]="fontloader-2023-12-28.lua",
  ["revision"]="48139bd3b29cd45baf567b5f35b98ad2b9d7bccb",
  ["timestamp"]="2025-02-26 22:33:31 +0100",
 },
}