%%
%%  Copyright (C) 2020, 2021 by Xiangdong Zeng <xdzeng96@gmail.com>
%%
%%  This work may be distributed and/or modified under the
%%  conditions of the LaTeX Project Public License, either
%%  version 1.3c of this license or (at your option) any later
%%  version. The latest version of this license is in:
%%
%%    http://www.latex-project.org/lppl.txt
%%
%%  and version 1.3 or later is part of all distributions of
%%  LaTeX version 2005/12/01 or later.
%%
%%  This work has the LPPL maintenance status `maintained'.
%%
%%  The Current Maintainer of this work is Xiangdong Zeng.
%%
\ProvidesExplFile{emoji-table.def}
  {2021/07/17}{0.2.2}{Emoji support in (Lua)LaTeX}
\__emoji_group:n {Smileys~\&~Emotion}
\__emoji_subgroup:n {Face-smiling}
\__emoji_def:nnnnn {^^^^^^01f600} {grinning-face} {grinning} {Grinning~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f603} {grinning-face-with-big-eyes} {smiley} {Grinning~face~with~big~eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f604} {grinning-face-with-smiling-eyes} {smile} {Grinning~face~with~smiling~eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f601} {beaming-face-with-smiling-eyes} {grin} {Beaming~face~with~smiling~eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f606} {grinning-squinting-face} {laughing, satisfied} {Grinning~squinting~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f605} {grinning-face-with-sweat} {sweat-smile} {Grinning~face~with~sweat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f923} {rolling-on-the-floor-laughing} {rofl} {Rolling~on~the~floor~laughing} {3.0}
\__emoji_def:nnnnn {^^^^^^01f602} {face-with-tears-of-joy} {joy} {Face~with~tears~of~joy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f642} {slightly-smiling-face} {} {Slightly~smiling~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f643} {upside-down-face} {} {Upside-down~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f609} {winking-face} {wink} {Winking~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f60a} {smiling-face-with-smiling-eyes} {blush} {Smiling~face~with~smiling~eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f607} {smiling-face-with-halo} {innocent} {Smiling~face~with~halo} {1.0}
\__emoji_subgroup:n {Face-affection}
\__emoji_def:nnnnn {^^^^^^01f970} {smiling-face-with-hearts} {smiling-face-with-three-hearts} {Smiling~face~with~hearts} {11.0}
\__emoji_def:nnnnn {^^^^^^01f60d} {smiling-face-with-heart-eyes} {heart-eyes} {Smiling~face~with~heart-eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f929} {star-struck} {} {Star-struck} {5.0}
\__emoji_def:nnnnn {^^^^^^01f618} {face-blowing-a-kiss} {kissing-heart} {Face~blowing~a~kiss} {0.6}
\__emoji_def:nnnnn {^^^^^^01f617} {kissing-face} {kissing} {Kissing~face} {1.0}
\__emoji_def:nnnnn {^^^^263a^^^^fe0f} {smiling-face} {relaxed} {Smiling~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f61a} {kissing-face-with-closed-eyes} {kissing-closed-eyes} {Kissing~face~with~closed~eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f619} {kissing-face-with-smiling-eyes} {kissing-smiling-eyes} {Kissing~face~with~smiling~eyes} {1.0}
\__emoji_def:nnnnn {^^^^^^01f972} {smiling-face-with-tear} {} {Smiling~face~with~tear} {13.0}
\__emoji_subgroup:n {Face-tongue}
\__emoji_def:nnnnn {^^^^^^01f60b} {face-savoring-food} {yum} {Face~savoring~food} {0.6}
\__emoji_def:nnnnn {^^^^^^01f61b} {face-with-tongue} {stuck-out-tongue} {Face~with~tongue} {1.0}
\__emoji_def:nnnnn {^^^^^^01f61c} {winking-face-with-tongue} {stuck-out-tongue-winking-eye} {Winking~face~with~tongue} {0.6}
\__emoji_def:nnnnn {^^^^^^01f92a} {zany-face} {} {Zany~face} {5.0}
\__emoji_def:nnnnn {^^^^^^01f61d} {squinting-face-with-tongue} {stuck-out-tongue-closed-eyes} {Squinting~face~with~tongue} {0.6}
\__emoji_def:nnnnn {^^^^^^01f911} {money-mouth-face} {} {Money-mouth~face} {1.0}
\__emoji_subgroup:n {Face-hand}
\__emoji_def:nnnnn {^^^^^^01f917} {hugging-face} {hugs} {Hugging~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f92d} {face-with-hand-over-mouth} {hand-over-mouth} {Face~with~hand~over~mouth} {5.0}
\__emoji_def:nnnnn {^^^^^^01f92b} {shushing-face} {} {Shushing~face} {5.0}
\__emoji_def:nnnnn {^^^^^^01f914} {thinking-face} {thinking} {Thinking~face} {1.0}
\__emoji_subgroup:n {Face-neutral-skeptical}
\__emoji_def:nnnnn {^^^^^^01f910} {zipper-mouth-face} {} {Zipper-mouth~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f928} {face-with-raised-eyebrow} {raised-eyebrow} {Face~with~raised~eyebrow} {5.0}
\__emoji_def:nnnnn {^^^^^^01f610} {neutral-face} {} {Neutral~face} {0.7}
\__emoji_def:nnnnn {^^^^^^01f611} {expressionless-face} {expressionless} {Expressionless~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f636} {face-without-mouth} {no-mouth} {Face~without~mouth} {1.0}
\__emoji_def:nnnnn {^^^^^^01f636^^^^200d^^^^^^01f32b^^^^fe0f} {face-in-clouds} {} {Face~in~clouds} {13.1}
\__emoji_def:nnnnn {^^^^^^01f60f} {smirking-face} {smirk} {Smirking~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f612} {unamused-face} {unamused} {Unamused~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f644} {face-with-rolling-eyes} {roll-eyes} {Face~with~rolling~eyes} {1.0}
\__emoji_def:nnnnn {^^^^^^01f62c} {grimacing-face} {grimacing} {Grimacing~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f62e^^^^200d^^^^^^01f4a8} {face-exhaling} {} {Face~exhaling} {13.1}
\__emoji_def:nnnnn {^^^^^^01f925} {lying-face} {} {Lying~face} {3.0}
\__emoji_subgroup:n {Face-sleepy}
\__emoji_def:nnnnn {^^^^^^01f60c} {relieved-face} {relieved} {Relieved~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f614} {pensive-face} {pensive} {Pensive~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f62a} {sleepy-face} {sleepy} {Sleepy~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f924} {drooling-face} {} {Drooling~face} {3.0}
\__emoji_def:nnnnn {^^^^^^01f634} {sleeping-face} {sleeping} {Sleeping~face} {1.0}
\__emoji_subgroup:n {Face-unwell}
\__emoji_def:nnnnn {^^^^^^01f637} {face-with-medical-mask} {mask} {Face~with~medical~mask} {0.6}
\__emoji_def:nnnnn {^^^^^^01f912} {face-with-thermometer} {} {Face~with~thermometer} {1.0}
\__emoji_def:nnnnn {^^^^^^01f915} {face-with-head-bandage} {} {Face~with~head-bandage} {1.0}
\__emoji_def:nnnnn {^^^^^^01f922} {nauseated-face} {} {Nauseated~face} {3.0}
\__emoji_def:nnnnn {^^^^^^01f92e} {face-vomiting} {vomiting-face} {Face~vomiting} {5.0}
\__emoji_def:nnnnn {^^^^^^01f927} {sneezing-face} {} {Sneezing~face} {3.0}
\__emoji_def:nnnnn {^^^^^^01f975} {hot-face} {} {Hot~face} {11.0}
\__emoji_def:nnnnn {^^^^^^01f976} {cold-face} {} {Cold~face} {11.0}
\__emoji_def:nnnnn {^^^^^^01f974} {woozy-face} {} {Woozy~face} {11.0}
\__emoji_def:nnnnn {^^^^^^01f635} {knocked-out-face} {dizzy-face} {Knocked-out~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f635^^^^200d^^^^^^01f4ab} {face-with-spiral-eyes} {} {Face~with~spiral~eyes} {13.1}
\__emoji_def:nnnnn {^^^^^^01f92f} {exploding-head} {} {Exploding~head} {5.0}
\__emoji_subgroup:n {Face-hat}
\__emoji_def:nnnnn {^^^^^^01f920} {cowboy-hat-face} {} {Cowboy~hat~face} {3.0}
\__emoji_def:nnnnn {^^^^^^01f973} {partying-face} {} {Partying~face} {11.0}
\__emoji_def:nnnnn {^^^^^^01f978} {disguised-face} {} {Disguised~face} {13.0}
\__emoji_subgroup:n {Face-glasses}
\__emoji_def:nnnnn {^^^^^^01f60e} {smiling-face-with-sunglasses} {} {Smiling~face~with~sunglasses} {1.0}
\__emoji_def:nnnnn {^^^^^^01f913} {nerd-face} {} {Nerd~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9d0} {face-with-monocle} {monocle-face} {Face~with~monocle} {5.0}
\__emoji_subgroup:n {Face-concerned}
\__emoji_def:nnnnn {^^^^^^01f615} {confused-face} {confused} {Confused~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f61f} {worried-face} {worried} {Worried~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f641} {slightly-frowning-face} {} {Slightly~frowning~face} {1.0}
\__emoji_def:nnnnn {^^^^2639^^^^fe0f} {frowning-face} {} {Frowning~face} {0.7}
\__emoji_def:nnnnn {^^^^^^01f62e} {face-with-open-mouth} {open-mouth} {Face~with~open~mouth} {1.0}
\__emoji_def:nnnnn {^^^^^^01f62f} {hushed-face} {hushed} {Hushed~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f632} {astonished-face} {astonished} {Astonished~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f633} {flushed-face} {flushed} {Flushed~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f97a} {pleading-face} {} {Pleading~face} {11.0}
\__emoji_def:nnnnn {^^^^^^01f626} {frowning-face-with-open-mouth} {frowning} {Frowning~face~with~open~mouth} {1.0}
\__emoji_def:nnnnn {^^^^^^01f627} {anguished-face} {anguished} {Anguished~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f628} {fearful-face} {fearful} {Fearful~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f630} {anxious-face-with-sweat} {cold-sweat} {Anxious~face~with~sweat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f625} {sad-but-relieved-face} {disappointed-relieved} {Sad~but~relieved~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f622} {crying-face} {cry} {Crying~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f62d} {loudly-crying-face} {sob} {Loudly~crying~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f631} {face-screaming-in-fear} {scream} {Face~screaming~in~fear} {0.6}
\__emoji_def:nnnnn {^^^^^^01f616} {confounded-face} {confounded} {Confounded~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f623} {persevering-face} {persevere} {Persevering~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f61e} {disappointed-face} {disappointed} {Disappointed~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f613} {downcast-face-with-sweat} {sweat} {Downcast~face~with~sweat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f629} {weary-face} {weary} {Weary~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f62b} {tired-face} {} {Tired~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f971} {yawning-face} {} {Yawning~face} {12.0}
\__emoji_subgroup:n {Face-negative}
\__emoji_def:nnnnn {^^^^^^01f624} {face-with-steam-from-nose} {triumph} {Face~with~steam~from~nose} {0.6}
\__emoji_def:nnnnn {^^^^^^01f621} {pouting-face} {pout, rage} {Pouting~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f620} {angry-face} {angry} {Angry~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f92c} {face-with-symbols-on-mouth} {cursing-face} {Face~with~symbols~on~mouth} {5.0}
\__emoji_def:nnnnn {^^^^^^01f608} {smiling-face-with-horns} {smiling-imp} {Smiling~face~with~horns} {1.0}
\__emoji_def:nnnnn {^^^^^^01f47f} {angry-face-with-horns} {imp} {Angry~face~with~horns} {0.6}
\__emoji_def:nnnnn {^^^^^^01f480} {skull} {} {Skull} {0.6}
\__emoji_def:nnnnn {^^^^2620^^^^fe0f} {skull-and-crossbones} {} {Skull~and~crossbones} {1.0}
\__emoji_subgroup:n {Face-costume}
\__emoji_def:nnnnn {^^^^^^01f4a9} {pile-of-poo} {hankey, poop, shit} {Pile~of~poo} {0.6}
\__emoji_def:nnnnn {^^^^^^01f921} {clown-face} {} {Clown~face} {3.0}
\__emoji_def:nnnnn {^^^^^^01f479} {ogre} {japanese-ogre} {Ogre} {0.6}
\__emoji_def:nnnnn {^^^^^^01f47a} {goblin} {japanese-goblin} {Goblin} {0.6}
\__emoji_def:nnnnn {^^^^^^01f47b} {ghost} {} {Ghost} {0.6}
\__emoji_def:nnnnn {^^^^^^01f47d} {alien} {} {Alien} {0.6}
\__emoji_def:nnnnn {^^^^^^01f47e} {alien-monster} {space-invader} {Alien~monster} {0.6}
\__emoji_def:nnnnn {^^^^^^01f916} {robot} {} {Robot} {1.0}
\__emoji_subgroup:n {Cat-face}
\__emoji_def:nnnnn {^^^^^^01f63a} {grinning-cat} {smiley-cat} {Grinning~cat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f638} {grinning-cat-with-smiling-eyes} {smile-cat} {Grinning~cat~with~smiling~eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f639} {cat-with-tears-of-joy} {joy-cat} {Cat~with~tears~of~joy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f63b} {smiling-cat-with-heart-eyes} {heart-eyes-cat} {Smiling~cat~with~heart-eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f63c} {cat-with-wry-smile} {smirk-cat} {Cat~with~wry~smile} {0.6}
\__emoji_def:nnnnn {^^^^^^01f63d} {kissing-cat} {} {Kissing~cat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f640} {weary-cat} {scream-cat} {Weary~cat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f63f} {crying-cat} {crying-cat-face} {Crying~cat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f63e} {pouting-cat} {} {Pouting~cat} {0.6}
\__emoji_subgroup:n {Monkey-face}
\__emoji_def:nnnnn {^^^^^^01f648} {see-no-evil-monkey} {see-no-evil} {See-no-evil~monkey} {0.6}
\__emoji_def:nnnnn {^^^^^^01f649} {hear-no-evil-monkey} {hear-no-evil} {Hear-no-evil~monkey} {0.6}
\__emoji_def:nnnnn {^^^^^^01f64a} {speak-no-evil-monkey} {speak-no-evil} {Speak-no-evil~monkey} {0.6}
\__emoji_subgroup:n {Emotion}
\__emoji_def:nnnnn {^^^^^^01f48b} {kiss-mark} {} {Kiss~mark} {0.6}
\__emoji_def:nnnnn {^^^^^^01f48c} {love-letter} {} {Love~letter} {0.6}
\__emoji_def:nnnnn {^^^^^^01f498} {heart-with-arrow} {cupid} {Heart~with~arrow} {0.6}
\__emoji_def:nnnnn {^^^^^^01f49d} {heart-with-ribbon} {gift-heart} {Heart~with~ribbon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f496} {sparkling-heart} {} {Sparkling~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f497} {growing-heart} {heartpulse} {Growing~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f493} {beating-heart} {heartbeat} {Beating~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f49e} {revolving-hearts} {} {Revolving~hearts} {0.6}
\__emoji_def:nnnnn {^^^^^^01f495} {two-hearts} {} {Two~hearts} {0.6}
\__emoji_def:nnnnn {^^^^^^01f49f} {heart-decoration} {} {Heart~decoration} {0.6}
\__emoji_def:nnnnn {^^^^2763^^^^fe0f} {heart-exclamation} {heavy-heart-exclamation} {Heart~exclamation} {1.0}
\__emoji_def:nnnnn {^^^^^^01f494} {broken-heart} {} {Broken~heart} {0.6}
\__emoji_def:nnnnn {^^^^2764^^^^fe0f^^^^200d^^^^^^01f525} {heart-on-fire} {} {Heart~on~fire} {13.1}
\__emoji_def:nnnnn {^^^^2764^^^^fe0f^^^^200d^^^^^^01fa79} {mending-heart} {} {Mending~heart} {13.1}
\__emoji_def:nnnnn {^^^^2764^^^^fe0f} {red-heart} {heart} {Red~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9e1} {orange-heart} {} {Orange~heart} {5.0}
\__emoji_def:nnnnn {^^^^^^01f49b} {yellow-heart} {} {Yellow~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f49a} {green-heart} {} {Green~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f499} {blue-heart} {} {Blue~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f49c} {purple-heart} {} {Purple~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f90e} {brown-heart} {} {Brown~heart} {12.0}
\__emoji_def:nnnnn {^^^^^^01f5a4} {black-heart} {} {Black~heart} {3.0}
\__emoji_def:nnnnn {^^^^^^01f90d} {white-heart} {} {White~heart} {12.0}
\__emoji_def:nnnnn {^^^^^^01f4af} {hundred-points} {100} {Hundred~points} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4a2} {anger-symbol} {anger} {Anger~symbol} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4a5} {collision} {boom} {Collision} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ab} {dizzy} {} {Dizzy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4a6} {sweat-droplets} {sweat-drops} {Sweat~droplets} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4a8} {dashing-away} {dash} {Dashing~away} {0.6}
\__emoji_def:nnnnn {^^^^^^01f573^^^^fe0f} {hole} {} {Hole} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4a3} {bomb} {} {Bomb} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ac} {speech-balloon} {} {Speech~balloon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f441^^^^fe0f^^^^200d^^^^^^01f5e8^^^^fe0f} {eye-in-speech-bubble} {eye-speech-bubble} {Eye~in~speech~bubble} {2.0}
\__emoji_def:nnnnn {^^^^^^01f5e8^^^^fe0f} {left-speech-bubble} {} {Left~speech~bubble} {2.0}
\__emoji_def:nnnnn {^^^^^^01f5ef^^^^fe0f} {right-anger-bubble} {} {Right~anger~bubble} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4ad} {thought-balloon} {} {Thought~balloon} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4a4} {zzz} {} {Zzz} {0.6}
\__emoji_group:n {People~\&~Body}
\__emoji_subgroup:n {Hand-fingers-open}
\__emoji_def:nnnnn {^^^^^^01f44b} {waving-hand} {wave} {Waving~hand} {0.6}
\__emoji_def:nnnnn {^^^^^^01f44b^^^^^^01f3fb} {waving-hand-light-skin-tone} {} {Waving~hand:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44b^^^^^^01f3fc} {waving-hand-medium-light-skin-tone} {} {Waving~hand:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44b^^^^^^01f3fd} {waving-hand-medium-skin-tone} {} {Waving~hand:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44b^^^^^^01f3fe} {waving-hand-medium-dark-skin-tone} {} {Waving~hand:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44b^^^^^^01f3ff} {waving-hand-dark-skin-tone} {} {Waving~hand:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f91a} {raised-back-of-hand} {} {Raised~back~of~hand} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91a^^^^^^01f3fb} {raised-back-of-hand-light-skin-tone} {} {Raised~back~of~hand:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91a^^^^^^01f3fc} {raised-back-of-hand-medium-light-skin-tone} {} {Raised~back~of~hand:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91a^^^^^^01f3fd} {raised-back-of-hand-medium-skin-tone} {} {Raised~back~of~hand:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91a^^^^^^01f3fe} {raised-back-of-hand-medium-dark-skin-tone} {} {Raised~back~of~hand:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91a^^^^^^01f3ff} {raised-back-of-hand-dark-skin-tone} {} {Raised~back~of~hand:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f590^^^^fe0f} {hand-with-fingers-splayed} {raised-hand-with-fingers-splayed} {Hand~with~fingers~splayed} {0.7}
\__emoji_def:nnnnn {^^^^^^01f590^^^^^^01f3fb} {hand-with-fingers-splayed-light-skin-tone} {} {Hand~with~fingers~splayed:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f590^^^^^^01f3fc} {hand-with-fingers-splayed-medium-light-skin-tone} {} {Hand~with~fingers~splayed:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f590^^^^^^01f3fd} {hand-with-fingers-splayed-medium-skin-tone} {} {Hand~with~fingers~splayed:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f590^^^^^^01f3fe} {hand-with-fingers-splayed-medium-dark-skin-tone} {} {Hand~with~fingers~splayed:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f590^^^^^^01f3ff} {hand-with-fingers-splayed-dark-skin-tone} {} {Hand~with~fingers~splayed:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270b} {raised-hand} {hand} {Raised~hand} {0.6}
\__emoji_def:nnnnn {^^^^270b^^^^^^01f3fb} {raised-hand-light-skin-tone} {} {Raised~hand:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270b^^^^^^01f3fc} {raised-hand-medium-light-skin-tone} {} {Raised~hand:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270b^^^^^^01f3fd} {raised-hand-medium-skin-tone} {} {Raised~hand:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270b^^^^^^01f3fe} {raised-hand-medium-dark-skin-tone} {} {Raised~hand:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270b^^^^^^01f3ff} {raised-hand-dark-skin-tone} {} {Raised~hand:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f596} {vulcan-salute} {} {Vulcan~salute} {1.0}
\__emoji_def:nnnnn {^^^^^^01f596^^^^^^01f3fb} {vulcan-salute-light-skin-tone} {} {Vulcan~salute:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f596^^^^^^01f3fc} {vulcan-salute-medium-light-skin-tone} {} {Vulcan~salute:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f596^^^^^^01f3fd} {vulcan-salute-medium-skin-tone} {} {Vulcan~salute:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f596^^^^^^01f3fe} {vulcan-salute-medium-dark-skin-tone} {} {Vulcan~salute:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f596^^^^^^01f3ff} {vulcan-salute-dark-skin-tone} {} {Vulcan~salute:~dark~skin~tone} {1.0}
\__emoji_subgroup:n {Hand-fingers-partial}
\__emoji_def:nnnnn {^^^^^^01f44c} {ok-hand} {} {OK~hand} {0.6}
\__emoji_def:nnnnn {^^^^^^01f44c^^^^^^01f3fb} {ok-hand-light-skin-tone} {} {OK~hand:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44c^^^^^^01f3fc} {ok-hand-medium-light-skin-tone} {} {OK~hand:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44c^^^^^^01f3fd} {ok-hand-medium-skin-tone} {} {OK~hand:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44c^^^^^^01f3fe} {ok-hand-medium-dark-skin-tone} {} {OK~hand:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44c^^^^^^01f3ff} {ok-hand-dark-skin-tone} {} {OK~hand:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f90c} {pinched-fingers} {} {Pinched~fingers} {13.0}
\__emoji_def:nnnnn {^^^^^^01f90c^^^^^^01f3fb} {pinched-fingers-light-skin-tone} {} {Pinched~fingers:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f90c^^^^^^01f3fc} {pinched-fingers-medium-light-skin-tone} {} {Pinched~fingers:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f90c^^^^^^01f3fd} {pinched-fingers-medium-skin-tone} {} {Pinched~fingers:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f90c^^^^^^01f3fe} {pinched-fingers-medium-dark-skin-tone} {} {Pinched~fingers:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f90c^^^^^^01f3ff} {pinched-fingers-dark-skin-tone} {} {Pinched~fingers:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f90f} {pinching-hand} {} {Pinching~hand} {12.0}
\__emoji_def:nnnnn {^^^^^^01f90f^^^^^^01f3fb} {pinching-hand-light-skin-tone} {} {Pinching~hand:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f90f^^^^^^01f3fc} {pinching-hand-medium-light-skin-tone} {} {Pinching~hand:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f90f^^^^^^01f3fd} {pinching-hand-medium-skin-tone} {} {Pinching~hand:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f90f^^^^^^01f3fe} {pinching-hand-medium-dark-skin-tone} {} {Pinching~hand:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f90f^^^^^^01f3ff} {pinching-hand-dark-skin-tone} {} {Pinching~hand:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^270c^^^^fe0f} {victory-hand} {v} {Victory~hand} {0.6}
\__emoji_def:nnnnn {^^^^270c^^^^^^01f3fb} {victory-hand-light-skin-tone} {} {Victory~hand:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270c^^^^^^01f3fc} {victory-hand-medium-light-skin-tone} {} {Victory~hand:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270c^^^^^^01f3fd} {victory-hand-medium-skin-tone} {} {Victory~hand:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270c^^^^^^01f3fe} {victory-hand-medium-dark-skin-tone} {} {Victory~hand:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270c^^^^^^01f3ff} {victory-hand-dark-skin-tone} {} {Victory~hand:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f91e} {crossed-fingers} {} {Crossed~fingers} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91e^^^^^^01f3fb} {crossed-fingers-light-skin-tone} {} {Crossed~fingers:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91e^^^^^^01f3fc} {crossed-fingers-medium-light-skin-tone} {} {Crossed~fingers:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91e^^^^^^01f3fd} {crossed-fingers-medium-skin-tone} {} {Crossed~fingers:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91e^^^^^^01f3fe} {crossed-fingers-medium-dark-skin-tone} {} {Crossed~fingers:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91e^^^^^^01f3ff} {crossed-fingers-dark-skin-tone} {} {Crossed~fingers:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91f} {love-you-gesture} {} {Love-you~gesture} {5.0}
\__emoji_def:nnnnn {^^^^^^01f91f^^^^^^01f3fb} {love-you-gesture-light-skin-tone} {} {Love-you~gesture:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f91f^^^^^^01f3fc} {love-you-gesture-medium-light-skin-tone} {} {Love-you~gesture:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f91f^^^^^^01f3fd} {love-you-gesture-medium-skin-tone} {} {Love-you~gesture:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f91f^^^^^^01f3fe} {love-you-gesture-medium-dark-skin-tone} {} {Love-you~gesture:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f91f^^^^^^01f3ff} {love-you-gesture-dark-skin-tone} {} {Love-you~gesture:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f918} {sign-of-the-horns} {metal} {Sign~of~the~horns} {1.0}
\__emoji_def:nnnnn {^^^^^^01f918^^^^^^01f3fb} {sign-of-the-horns-light-skin-tone} {} {Sign~of~the~horns:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f918^^^^^^01f3fc} {sign-of-the-horns-medium-light-skin-tone} {} {Sign~of~the~horns:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f918^^^^^^01f3fd} {sign-of-the-horns-medium-skin-tone} {} {Sign~of~the~horns:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f918^^^^^^01f3fe} {sign-of-the-horns-medium-dark-skin-tone} {} {Sign~of~the~horns:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f918^^^^^^01f3ff} {sign-of-the-horns-dark-skin-tone} {} {Sign~of~the~horns:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f919} {call-me-hand} {} {Call~me~hand} {3.0}
\__emoji_def:nnnnn {^^^^^^01f919^^^^^^01f3fb} {call-me-hand-light-skin-tone} {} {Call~me~hand:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f919^^^^^^01f3fc} {call-me-hand-medium-light-skin-tone} {} {Call~me~hand:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f919^^^^^^01f3fd} {call-me-hand-medium-skin-tone} {} {Call~me~hand:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f919^^^^^^01f3fe} {call-me-hand-medium-dark-skin-tone} {} {Call~me~hand:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f919^^^^^^01f3ff} {call-me-hand-dark-skin-tone} {} {Call~me~hand:~dark~skin~tone} {3.0}
\__emoji_subgroup:n {Hand-single-finger}
\__emoji_def:nnnnn {^^^^^^01f448} {backhand-index-pointing-left} {point-left} {Backhand~index~pointing~left} {0.6}
\__emoji_def:nnnnn {^^^^^^01f448^^^^^^01f3fb} {backhand-index-pointing-left-light-skin-tone} {} {Backhand~index~pointing~left:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f448^^^^^^01f3fc} {backhand-index-pointing-left-medium-light-skin-tone} {} {Backhand~index~pointing~left:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f448^^^^^^01f3fd} {backhand-index-pointing-left-medium-skin-tone} {} {Backhand~index~pointing~left:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f448^^^^^^01f3fe} {backhand-index-pointing-left-medium-dark-skin-tone} {} {Backhand~index~pointing~left:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f448^^^^^^01f3ff} {backhand-index-pointing-left-dark-skin-tone} {} {Backhand~index~pointing~left:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f449} {backhand-index-pointing-right} {point-right} {Backhand~index~pointing~right} {0.6}
\__emoji_def:nnnnn {^^^^^^01f449^^^^^^01f3fb} {backhand-index-pointing-right-light-skin-tone} {} {Backhand~index~pointing~right:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f449^^^^^^01f3fc} {backhand-index-pointing-right-medium-light-skin-tone} {} {Backhand~index~pointing~right:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f449^^^^^^01f3fd} {backhand-index-pointing-right-medium-skin-tone} {} {Backhand~index~pointing~right:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f449^^^^^^01f3fe} {backhand-index-pointing-right-medium-dark-skin-tone} {} {Backhand~index~pointing~right:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f449^^^^^^01f3ff} {backhand-index-pointing-right-dark-skin-tone} {} {Backhand~index~pointing~right:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f446} {backhand-index-pointing-up} {point-up-2} {Backhand~index~pointing~up} {0.6}
\__emoji_def:nnnnn {^^^^^^01f446^^^^^^01f3fb} {backhand-index-pointing-up-light-skin-tone} {} {Backhand~index~pointing~up:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f446^^^^^^01f3fc} {backhand-index-pointing-up-medium-light-skin-tone} {} {Backhand~index~pointing~up:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f446^^^^^^01f3fd} {backhand-index-pointing-up-medium-skin-tone} {} {Backhand~index~pointing~up:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f446^^^^^^01f3fe} {backhand-index-pointing-up-medium-dark-skin-tone} {} {Backhand~index~pointing~up:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f446^^^^^^01f3ff} {backhand-index-pointing-up-dark-skin-tone} {} {Backhand~index~pointing~up:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f595} {middle-finger} {fu} {Middle~finger} {1.0}
\__emoji_def:nnnnn {^^^^^^01f595^^^^^^01f3fb} {middle-finger-light-skin-tone} {} {Middle~finger:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f595^^^^^^01f3fc} {middle-finger-medium-light-skin-tone} {} {Middle~finger:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f595^^^^^^01f3fd} {middle-finger-medium-skin-tone} {} {Middle~finger:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f595^^^^^^01f3fe} {middle-finger-medium-dark-skin-tone} {} {Middle~finger:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f595^^^^^^01f3ff} {middle-finger-dark-skin-tone} {} {Middle~finger:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f447} {backhand-index-pointing-down} {point-down} {Backhand~index~pointing~down} {0.6}
\__emoji_def:nnnnn {^^^^^^01f447^^^^^^01f3fb} {backhand-index-pointing-down-light-skin-tone} {} {Backhand~index~pointing~down:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f447^^^^^^01f3fc} {backhand-index-pointing-down-medium-light-skin-tone} {} {Backhand~index~pointing~down:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f447^^^^^^01f3fd} {backhand-index-pointing-down-medium-skin-tone} {} {Backhand~index~pointing~down:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f447^^^^^^01f3fe} {backhand-index-pointing-down-medium-dark-skin-tone} {} {Backhand~index~pointing~down:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f447^^^^^^01f3ff} {backhand-index-pointing-down-dark-skin-tone} {} {Backhand~index~pointing~down:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^261d^^^^fe0f} {index-pointing-up} {point-up} {Index~pointing~up} {0.6}
\__emoji_def:nnnnn {^^^^261d^^^^^^01f3fb} {index-pointing-up-light-skin-tone} {} {Index~pointing~up:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^261d^^^^^^01f3fc} {index-pointing-up-medium-light-skin-tone} {} {Index~pointing~up:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^261d^^^^^^01f3fd} {index-pointing-up-medium-skin-tone} {} {Index~pointing~up:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^261d^^^^^^01f3fe} {index-pointing-up-medium-dark-skin-tone} {} {Index~pointing~up:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^261d^^^^^^01f3ff} {index-pointing-up-dark-skin-tone} {} {Index~pointing~up:~dark~skin~tone} {1.0}
\__emoji_subgroup:n {Hand-fingers-closed}
\__emoji_def:nnnnn {^^^^^^01f44d} {thumbs-up} {+1, thumbsup} {Thumbs~up} {0.6}
\__emoji_def:nnnnn {^^^^^^01f44d^^^^^^01f3fb} {thumbs-up-light-skin-tone} {} {Thumbs~up:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44d^^^^^^01f3fc} {thumbs-up-medium-light-skin-tone} {} {Thumbs~up:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44d^^^^^^01f3fd} {thumbs-up-medium-skin-tone} {} {Thumbs~up:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44d^^^^^^01f3fe} {thumbs-up-medium-dark-skin-tone} {} {Thumbs~up:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44d^^^^^^01f3ff} {thumbs-up-dark-skin-tone} {} {Thumbs~up:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44e} {thumbs-down} {-1, thumbsdown} {Thumbs~down} {0.6}
\__emoji_def:nnnnn {^^^^^^01f44e^^^^^^01f3fb} {thumbs-down-light-skin-tone} {} {Thumbs~down:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44e^^^^^^01f3fc} {thumbs-down-medium-light-skin-tone} {} {Thumbs~down:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44e^^^^^^01f3fd} {thumbs-down-medium-skin-tone} {} {Thumbs~down:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44e^^^^^^01f3fe} {thumbs-down-medium-dark-skin-tone} {} {Thumbs~down:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44e^^^^^^01f3ff} {thumbs-down-dark-skin-tone} {} {Thumbs~down:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270a} {raised-fist} {fist, fist-raised} {Raised~fist} {0.6}
\__emoji_def:nnnnn {^^^^270a^^^^^^01f3fb} {raised-fist-light-skin-tone} {} {Raised~fist:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270a^^^^^^01f3fc} {raised-fist-medium-light-skin-tone} {} {Raised~fist:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270a^^^^^^01f3fd} {raised-fist-medium-skin-tone} {} {Raised~fist:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270a^^^^^^01f3fe} {raised-fist-medium-dark-skin-tone} {} {Raised~fist:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270a^^^^^^01f3ff} {raised-fist-dark-skin-tone} {} {Raised~fist:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44a} {oncoming-fist} {facepunch, fist-oncoming, punch} {Oncoming~fist} {0.6}
\__emoji_def:nnnnn {^^^^^^01f44a^^^^^^01f3fb} {oncoming-fist-light-skin-tone} {} {Oncoming~fist:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44a^^^^^^01f3fc} {oncoming-fist-medium-light-skin-tone} {} {Oncoming~fist:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44a^^^^^^01f3fd} {oncoming-fist-medium-skin-tone} {} {Oncoming~fist:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44a^^^^^^01f3fe} {oncoming-fist-medium-dark-skin-tone} {} {Oncoming~fist:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44a^^^^^^01f3ff} {oncoming-fist-dark-skin-tone} {} {Oncoming~fist:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f91b} {left-facing-fist} {fist-left} {Left-facing~fist} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91b^^^^^^01f3fb} {left-facing-fist-light-skin-tone} {} {Left-facing~fist:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91b^^^^^^01f3fc} {left-facing-fist-medium-light-skin-tone} {} {Left-facing~fist:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91b^^^^^^01f3fd} {left-facing-fist-medium-skin-tone} {} {Left-facing~fist:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91b^^^^^^01f3fe} {left-facing-fist-medium-dark-skin-tone} {} {Left-facing~fist:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91b^^^^^^01f3ff} {left-facing-fist-dark-skin-tone} {} {Left-facing~fist:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91c} {right-facing-fist} {fist-right} {Right-facing~fist} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91c^^^^^^01f3fb} {right-facing-fist-light-skin-tone} {} {Right-facing~fist:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91c^^^^^^01f3fc} {right-facing-fist-medium-light-skin-tone} {} {Right-facing~fist:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91c^^^^^^01f3fd} {right-facing-fist-medium-skin-tone} {} {Right-facing~fist:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91c^^^^^^01f3fe} {right-facing-fist-medium-dark-skin-tone} {} {Right-facing~fist:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f91c^^^^^^01f3ff} {right-facing-fist-dark-skin-tone} {} {Right-facing~fist:~dark~skin~tone} {3.0}
\__emoji_subgroup:n {Hands}
\__emoji_def:nnnnn {^^^^^^01f44f} {clapping-hands} {clap} {Clapping~hands} {0.6}
\__emoji_def:nnnnn {^^^^^^01f44f^^^^^^01f3fb} {clapping-hands-light-skin-tone} {} {Clapping~hands:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44f^^^^^^01f3fc} {clapping-hands-medium-light-skin-tone} {} {Clapping~hands:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44f^^^^^^01f3fd} {clapping-hands-medium-skin-tone} {} {Clapping~hands:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44f^^^^^^01f3fe} {clapping-hands-medium-dark-skin-tone} {} {Clapping~hands:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f44f^^^^^^01f3ff} {clapping-hands-dark-skin-tone} {} {Clapping~hands:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64c} {raising-hands} {raised-hands} {Raising~hands} {0.6}
\__emoji_def:nnnnn {^^^^^^01f64c^^^^^^01f3fb} {raising-hands-light-skin-tone} {} {Raising~hands:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64c^^^^^^01f3fc} {raising-hands-medium-light-skin-tone} {} {Raising~hands:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64c^^^^^^01f3fd} {raising-hands-medium-skin-tone} {} {Raising~hands:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64c^^^^^^01f3fe} {raising-hands-medium-dark-skin-tone} {} {Raising~hands:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64c^^^^^^01f3ff} {raising-hands-dark-skin-tone} {} {Raising~hands:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f450} {open-hands} {} {Open~hands} {0.6}
\__emoji_def:nnnnn {^^^^^^01f450^^^^^^01f3fb} {open-hands-light-skin-tone} {} {Open~hands:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f450^^^^^^01f3fc} {open-hands-medium-light-skin-tone} {} {Open~hands:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f450^^^^^^01f3fd} {open-hands-medium-skin-tone} {} {Open~hands:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f450^^^^^^01f3fe} {open-hands-medium-dark-skin-tone} {} {Open~hands:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f450^^^^^^01f3ff} {open-hands-dark-skin-tone} {} {Open~hands:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f932} {palms-up-together} {} {Palms~up~together} {5.0}
\__emoji_def:nnnnn {^^^^^^01f932^^^^^^01f3fb} {palms-up-together-light-skin-tone} {} {Palms~up~together:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f932^^^^^^01f3fc} {palms-up-together-medium-light-skin-tone} {} {Palms~up~together:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f932^^^^^^01f3fd} {palms-up-together-medium-skin-tone} {} {Palms~up~together:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f932^^^^^^01f3fe} {palms-up-together-medium-dark-skin-tone} {} {Palms~up~together:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f932^^^^^^01f3ff} {palms-up-together-dark-skin-tone} {} {Palms~up~together:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f91d} {handshake} {} {Handshake} {3.0}
\__emoji_def:nnnnn {^^^^^^01f64f} {folded-hands} {pray} {Folded~hands} {0.6}
\__emoji_def:nnnnn {^^^^^^01f64f^^^^^^01f3fb} {folded-hands-light-skin-tone} {} {Folded~hands:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64f^^^^^^01f3fc} {folded-hands-medium-light-skin-tone} {} {Folded~hands:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64f^^^^^^01f3fd} {folded-hands-medium-skin-tone} {} {Folded~hands:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64f^^^^^^01f3fe} {folded-hands-medium-dark-skin-tone} {} {Folded~hands:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64f^^^^^^01f3ff} {folded-hands-dark-skin-tone} {} {Folded~hands:~dark~skin~tone} {1.0}
\__emoji_subgroup:n {Hand-prop}
\__emoji_def:nnnnn {^^^^270d^^^^fe0f} {writing-hand} {} {Writing~hand} {0.7}
\__emoji_def:nnnnn {^^^^270d^^^^^^01f3fb} {writing-hand-light-skin-tone} {} {Writing~hand:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270d^^^^^^01f3fc} {writing-hand-medium-light-skin-tone} {} {Writing~hand:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270d^^^^^^01f3fd} {writing-hand-medium-skin-tone} {} {Writing~hand:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270d^^^^^^01f3fe} {writing-hand-medium-dark-skin-tone} {} {Writing~hand:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^270d^^^^^^01f3ff} {writing-hand-dark-skin-tone} {} {Writing~hand:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f485} {nail-polish} {nail-care} {Nail~polish} {0.6}
\__emoji_def:nnnnn {^^^^^^01f485^^^^^^01f3fb} {nail-polish-light-skin-tone} {} {Nail~polish:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f485^^^^^^01f3fc} {nail-polish-medium-light-skin-tone} {} {Nail~polish:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f485^^^^^^01f3fd} {nail-polish-medium-skin-tone} {} {Nail~polish:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f485^^^^^^01f3fe} {nail-polish-medium-dark-skin-tone} {} {Nail~polish:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f485^^^^^^01f3ff} {nail-polish-dark-skin-tone} {} {Nail~polish:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f933} {selfie} {} {Selfie} {3.0}
\__emoji_def:nnnnn {^^^^^^01f933^^^^^^01f3fb} {selfie-light-skin-tone} {} {Selfie:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f933^^^^^^01f3fc} {selfie-medium-light-skin-tone} {} {Selfie:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f933^^^^^^01f3fd} {selfie-medium-skin-tone} {} {Selfie:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f933^^^^^^01f3fe} {selfie-medium-dark-skin-tone} {} {Selfie:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f933^^^^^^01f3ff} {selfie-dark-skin-tone} {} {Selfie:~dark~skin~tone} {3.0}
\__emoji_subgroup:n {Body-parts}
\__emoji_def:nnnnn {^^^^^^01f4aa} {flexed-biceps} {muscle} {Flexed~biceps} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4aa^^^^^^01f3fb} {flexed-biceps-light-skin-tone} {} {Flexed~biceps:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4aa^^^^^^01f3fc} {flexed-biceps-medium-light-skin-tone} {} {Flexed~biceps:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4aa^^^^^^01f3fd} {flexed-biceps-medium-skin-tone} {} {Flexed~biceps:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4aa^^^^^^01f3fe} {flexed-biceps-medium-dark-skin-tone} {} {Flexed~biceps:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4aa^^^^^^01f3ff} {flexed-biceps-dark-skin-tone} {} {Flexed~biceps:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9be} {mechanical-arm} {} {Mechanical~arm} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9bf} {mechanical-leg} {} {Mechanical~leg} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9b5} {leg} {} {Leg} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b5^^^^^^01f3fb} {leg-light-skin-tone} {} {Leg:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b5^^^^^^01f3fc} {leg-medium-light-skin-tone} {} {Leg:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b5^^^^^^01f3fd} {leg-medium-skin-tone} {} {Leg:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b5^^^^^^01f3fe} {leg-medium-dark-skin-tone} {} {Leg:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b5^^^^^^01f3ff} {leg-dark-skin-tone} {} {Leg:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b6} {foot} {} {Foot} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b6^^^^^^01f3fb} {foot-light-skin-tone} {} {Foot:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b6^^^^^^01f3fc} {foot-medium-light-skin-tone} {} {Foot:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b6^^^^^^01f3fd} {foot-medium-skin-tone} {} {Foot:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b6^^^^^^01f3fe} {foot-medium-dark-skin-tone} {} {Foot:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b6^^^^^^01f3ff} {foot-dark-skin-tone} {} {Foot:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f442} {ear} {} {Ear} {0.6}
\__emoji_def:nnnnn {^^^^^^01f442^^^^^^01f3fb} {ear-light-skin-tone} {} {Ear:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f442^^^^^^01f3fc} {ear-medium-light-skin-tone} {} {Ear:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f442^^^^^^01f3fd} {ear-medium-skin-tone} {} {Ear:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f442^^^^^^01f3fe} {ear-medium-dark-skin-tone} {} {Ear:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f442^^^^^^01f3ff} {ear-dark-skin-tone} {} {Ear:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9bb} {ear-with-hearing-aid} {} {Ear~with~hearing~aid} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9bb^^^^^^01f3fb} {ear-with-hearing-aid-light-skin-tone} {} {Ear~with~hearing~aid:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9bb^^^^^^01f3fc} {ear-with-hearing-aid-medium-light-skin-tone} {} {Ear~with~hearing~aid:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9bb^^^^^^01f3fd} {ear-with-hearing-aid-medium-skin-tone} {} {Ear~with~hearing~aid:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9bb^^^^^^01f3fe} {ear-with-hearing-aid-medium-dark-skin-tone} {} {Ear~with~hearing~aid:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9bb^^^^^^01f3ff} {ear-with-hearing-aid-dark-skin-tone} {} {Ear~with~hearing~aid:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f443} {nose} {} {Nose} {0.6}
\__emoji_def:nnnnn {^^^^^^01f443^^^^^^01f3fb} {nose-light-skin-tone} {} {Nose:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f443^^^^^^01f3fc} {nose-medium-light-skin-tone} {} {Nose:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f443^^^^^^01f3fd} {nose-medium-skin-tone} {} {Nose:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f443^^^^^^01f3fe} {nose-medium-dark-skin-tone} {} {Nose:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f443^^^^^^01f3ff} {nose-dark-skin-tone} {} {Nose:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9e0} {brain} {} {Brain} {5.0}
\__emoji_def:nnnnn {^^^^^^01fac0} {anatomical-heart} {} {Anatomical~heart} {13.0}
\__emoji_def:nnnnn {^^^^^^01fac1} {lungs} {} {Lungs} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9b7} {tooth} {} {Tooth} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b4} {bone} {} {Bone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f440} {eyes} {} {Eyes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f441^^^^fe0f} {eye} {} {Eye} {0.7}
\__emoji_def:nnnnn {^^^^^^01f445} {tongue} {} {Tongue} {0.6}
\__emoji_def:nnnnn {^^^^^^01f444} {mouth} {lips} {Mouth} {0.6}
\__emoji_subgroup:n {Person}
\__emoji_def:nnnnn {^^^^^^01f476} {baby} {} {Baby} {0.6}
\__emoji_def:nnnnn {^^^^^^01f476^^^^^^01f3fb} {baby-light-skin-tone} {} {Baby:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f476^^^^^^01f3fc} {baby-medium-light-skin-tone} {} {Baby:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f476^^^^^^01f3fd} {baby-medium-skin-tone} {} {Baby:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f476^^^^^^01f3fe} {baby-medium-dark-skin-tone} {} {Baby:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f476^^^^^^01f3ff} {baby-dark-skin-tone} {} {Baby:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9d2} {child} {} {Child} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d2^^^^^^01f3fb} {child-light-skin-tone} {} {Child:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d2^^^^^^01f3fc} {child-medium-light-skin-tone} {} {Child:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d2^^^^^^01f3fd} {child-medium-skin-tone} {} {Child:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d2^^^^^^01f3fe} {child-medium-dark-skin-tone} {} {Child:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d2^^^^^^01f3ff} {child-dark-skin-tone} {} {Child:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f466} {boy} {} {Boy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f466^^^^^^01f3fb} {boy-light-skin-tone} {} {Boy:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f466^^^^^^01f3fc} {boy-medium-light-skin-tone} {} {Boy:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f466^^^^^^01f3fd} {boy-medium-skin-tone} {} {Boy:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f466^^^^^^01f3fe} {boy-medium-dark-skin-tone} {} {Boy:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f466^^^^^^01f3ff} {boy-dark-skin-tone} {} {Boy:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f467} {girl} {} {Girl} {0.6}
\__emoji_def:nnnnn {^^^^^^01f467^^^^^^01f3fb} {girl-light-skin-tone} {} {Girl:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f467^^^^^^01f3fc} {girl-medium-light-skin-tone} {} {Girl:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f467^^^^^^01f3fd} {girl-medium-skin-tone} {} {Girl:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f467^^^^^^01f3fe} {girl-medium-dark-skin-tone} {} {Girl:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f467^^^^^^01f3ff} {girl-dark-skin-tone} {} {Girl:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9d1} {person} {adult} {Person} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb} {person-light-skin-tone} {} {Person:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc} {person-medium-light-skin-tone} {} {Person:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd} {person-medium-skin-tone} {} {Person:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe} {person-medium-dark-skin-tone} {} {Person:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff} {person-dark-skin-tone} {} {Person:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f471} {person-blond-hair} {blond-haired-person} {Person:~blond~hair} {0.6}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fb} {person-light-skin-tone-blond-hair} {} {Person:~light~skin~tone,~blond~hair} {1.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fc} {person-medium-light-skin-tone-blond-hair} {} {Person:~medium-light~skin~tone,~blond~hair} {1.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fd} {person-medium-skin-tone-blond-hair} {} {Person:~medium~skin~tone,~blond~hair} {1.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fe} {person-medium-dark-skin-tone-blond-hair} {} {Person:~medium-dark~skin~tone,~blond~hair} {1.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3ff} {person-dark-skin-tone-blond-hair} {} {Person:~dark~skin~tone,~blond~hair} {1.0}
\__emoji_def:nnnnn {^^^^^^01f468} {man} {} {Man} {0.6}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb} {man-light-skin-tone} {} {Man:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc} {man-medium-light-skin-tone} {} {Man:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd} {man-medium-skin-tone} {} {Man:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe} {man-medium-dark-skin-tone} {} {Man:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff} {man-dark-skin-tone} {} {Man:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9d4} {person-beard} {bearded-person} {Person:~beard} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fb} {person-light-skin-tone-beard} {} {Person:~light~skin~tone,~beard} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fc} {person-medium-light-skin-tone-beard} {} {Person:~medium-light~skin~tone,~beard} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fd} {person-medium-skin-tone-beard} {} {Person:~medium~skin~tone,~beard} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fe} {person-medium-dark-skin-tone-beard} {} {Person:~medium-dark~skin~tone,~beard} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3ff} {person-dark-skin-tone-beard} {} {Person:~dark~skin~tone,~beard} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^200d^^^^2642^^^^fe0f} {man-beard} {} {Man:~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-light-skin-tone-beard} {} {Man:~light~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-medium-light-skin-tone-beard} {} {Man:~medium-light~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-medium-skin-tone-beard} {} {Man:~medium~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-medium-dark-skin-tone-beard} {} {Man:~medium-dark~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-dark-skin-tone-beard} {} {Man:~dark~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^200d^^^^2640^^^^fe0f} {woman-beard} {} {Woman:~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-light-skin-tone-beard} {} {Woman:~light~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-medium-light-skin-tone-beard} {} {Woman:~medium-light~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-medium-skin-tone-beard} {} {Woman:~medium~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-medium-dark-skin-tone-beard} {} {Woman:~medium-dark~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d4^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-dark-skin-tone-beard} {} {Woman:~dark~skin~tone,~beard} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f9b0} {man-red-hair} {red-haired-man} {Man:~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f9b0} {man-light-skin-tone-red-hair} {} {Man:~light~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f9b0} {man-medium-light-skin-tone-red-hair} {} {Man:~medium-light~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f9b0} {man-medium-skin-tone-red-hair} {} {Man:~medium~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f9b0} {man-medium-dark-skin-tone-red-hair} {} {Man:~medium-dark~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f9b0} {man-dark-skin-tone-red-hair} {} {Man:~dark~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f9b1} {man-curly-hair} {curly-haired-man} {Man:~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f9b1} {man-light-skin-tone-curly-hair} {} {Man:~light~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f9b1} {man-medium-light-skin-tone-curly-hair} {} {Man:~medium-light~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f9b1} {man-medium-skin-tone-curly-hair} {} {Man:~medium~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f9b1} {man-medium-dark-skin-tone-curly-hair} {} {Man:~medium-dark~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f9b1} {man-dark-skin-tone-curly-hair} {} {Man:~dark~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f9b3} {man-white-hair} {white-haired-man} {Man:~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f9b3} {man-light-skin-tone-white-hair} {} {Man:~light~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f9b3} {man-medium-light-skin-tone-white-hair} {} {Man:~medium-light~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f9b3} {man-medium-skin-tone-white-hair} {} {Man:~medium~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f9b3} {man-medium-dark-skin-tone-white-hair} {} {Man:~medium-dark~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f9b3} {man-dark-skin-tone-white-hair} {} {Man:~dark~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f9b2} {man-bald} {bald-man} {Man:~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f9b2} {man-light-skin-tone-bald} {} {Man:~light~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f9b2} {man-medium-light-skin-tone-bald} {} {Man:~medium-light~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f9b2} {man-medium-skin-tone-bald} {} {Man:~medium~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f9b2} {man-medium-dark-skin-tone-bald} {} {Man:~medium-dark~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f9b2} {man-dark-skin-tone-bald} {} {Man:~dark~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469} {woman} {} {Woman} {0.6}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb} {woman-light-skin-tone} {} {Woman:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc} {woman-medium-light-skin-tone} {} {Woman:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd} {woman-medium-skin-tone} {} {Woman:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe} {woman-medium-dark-skin-tone} {} {Woman:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff} {woman-dark-skin-tone} {} {Woman:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f9b0} {woman-red-hair} {red-haired-woman} {Woman:~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f9b0} {woman-light-skin-tone-red-hair} {} {Woman:~light~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f9b0} {woman-medium-light-skin-tone-red-hair} {} {Woman:~medium-light~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f9b0} {woman-medium-skin-tone-red-hair} {} {Woman:~medium~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f9b0} {woman-medium-dark-skin-tone-red-hair} {} {Woman:~medium-dark~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f9b0} {woman-dark-skin-tone-red-hair} {} {Woman:~dark~skin~tone,~red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f9b0} {person-red-hair} {} {Person:~red~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f9b0} {person-light-skin-tone-red-hair} {} {Person:~light~skin~tone,~red~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f9b0} {person-medium-light-skin-tone-red-hair} {} {Person:~medium-light~skin~tone,~red~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f9b0} {person-medium-skin-tone-red-hair} {} {Person:~medium~skin~tone,~red~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f9b0} {person-medium-dark-skin-tone-red-hair} {} {Person:~medium-dark~skin~tone,~red~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f9b0} {person-dark-skin-tone-red-hair} {} {Person:~dark~skin~tone,~red~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f9b1} {woman-curly-hair} {curly-haired-woman} {Woman:~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f9b1} {woman-light-skin-tone-curly-hair} {} {Woman:~light~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f9b1} {woman-medium-light-skin-tone-curly-hair} {} {Woman:~medium-light~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f9b1} {woman-medium-skin-tone-curly-hair} {} {Woman:~medium~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f9b1} {woman-medium-dark-skin-tone-curly-hair} {} {Woman:~medium-dark~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f9b1} {woman-dark-skin-tone-curly-hair} {} {Woman:~dark~skin~tone,~curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f9b1} {person-curly-hair} {} {Person:~curly~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f9b1} {person-light-skin-tone-curly-hair} {} {Person:~light~skin~tone,~curly~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f9b1} {person-medium-light-skin-tone-curly-hair} {} {Person:~medium-light~skin~tone,~curly~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f9b1} {person-medium-skin-tone-curly-hair} {} {Person:~medium~skin~tone,~curly~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f9b1} {person-medium-dark-skin-tone-curly-hair} {} {Person:~medium-dark~skin~tone,~curly~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f9b1} {person-dark-skin-tone-curly-hair} {} {Person:~dark~skin~tone,~curly~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f9b3} {woman-white-hair} {white-haired-woman} {Woman:~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f9b3} {woman-light-skin-tone-white-hair} {} {Woman:~light~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f9b3} {woman-medium-light-skin-tone-white-hair} {} {Woman:~medium-light~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f9b3} {woman-medium-skin-tone-white-hair} {} {Woman:~medium~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f9b3} {woman-medium-dark-skin-tone-white-hair} {} {Woman:~medium-dark~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f9b3} {woman-dark-skin-tone-white-hair} {} {Woman:~dark~skin~tone,~white~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f9b3} {person-white-hair} {} {Person:~white~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f9b3} {person-light-skin-tone-white-hair} {} {Person:~light~skin~tone,~white~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f9b3} {person-medium-light-skin-tone-white-hair} {} {Person:~medium-light~skin~tone,~white~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f9b3} {person-medium-skin-tone-white-hair} {} {Person:~medium~skin~tone,~white~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f9b3} {person-medium-dark-skin-tone-white-hair} {} {Person:~medium-dark~skin~tone,~white~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f9b3} {person-dark-skin-tone-white-hair} {} {Person:~dark~skin~tone,~white~hair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f9b2} {woman-bald} {bald-woman} {Woman:~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f9b2} {woman-light-skin-tone-bald} {} {Woman:~light~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f9b2} {woman-medium-light-skin-tone-bald} {} {Woman:~medium-light~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f9b2} {woman-medium-skin-tone-bald} {} {Woman:~medium~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f9b2} {woman-medium-dark-skin-tone-bald} {} {Woman:~medium-dark~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f9b2} {woman-dark-skin-tone-bald} {} {Woman:~dark~skin~tone,~bald} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f9b2} {person-bald} {} {Person:~bald} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f9b2} {person-light-skin-tone-bald} {} {Person:~light~skin~tone,~bald} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f9b2} {person-medium-light-skin-tone-bald} {} {Person:~medium-light~skin~tone,~bald} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f9b2} {person-medium-skin-tone-bald} {} {Person:~medium~skin~tone,~bald} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f9b2} {person-medium-dark-skin-tone-bald} {} {Person:~medium-dark~skin~tone,~bald} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f9b2} {person-dark-skin-tone-bald} {} {Person:~dark~skin~tone,~bald} {12.1}
\__emoji_def:nnnnn {^^^^^^01f471^^^^200d^^^^2640^^^^fe0f} {woman-blond-hair} {blond-haired-woman, blonde-woman} {Woman:~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-light-skin-tone-blond-hair} {} {Woman:~light~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-medium-light-skin-tone-blond-hair} {} {Woman:~medium-light~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-medium-skin-tone-blond-hair} {} {Woman:~medium~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-medium-dark-skin-tone-blond-hair} {} {Woman:~medium-dark~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-dark-skin-tone-blond-hair} {} {Woman:~dark~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^200d^^^^2642^^^^fe0f} {man-blond-hair} {blond-haired-man} {Man:~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-light-skin-tone-blond-hair} {} {Man:~light~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-medium-light-skin-tone-blond-hair} {} {Man:~medium-light~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-medium-skin-tone-blond-hair} {} {Man:~medium~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-medium-dark-skin-tone-blond-hair} {} {Man:~medium-dark~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f471^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-dark-skin-tone-blond-hair} {} {Man:~dark~skin~tone,~blond~hair} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d3} {older-person} {older-adult} {Older~person} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d3^^^^^^01f3fb} {older-person-light-skin-tone} {} {Older~person:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d3^^^^^^01f3fc} {older-person-medium-light-skin-tone} {} {Older~person:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d3^^^^^^01f3fd} {older-person-medium-skin-tone} {} {Older~person:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d3^^^^^^01f3fe} {older-person-medium-dark-skin-tone} {} {Older~person:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d3^^^^^^01f3ff} {older-person-dark-skin-tone} {} {Older~person:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f474} {old-man} {older-man} {Old~man} {0.6}
\__emoji_def:nnnnn {^^^^^^01f474^^^^^^01f3fb} {old-man-light-skin-tone} {} {Old~man:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f474^^^^^^01f3fc} {old-man-medium-light-skin-tone} {} {Old~man:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f474^^^^^^01f3fd} {old-man-medium-skin-tone} {} {Old~man:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f474^^^^^^01f3fe} {old-man-medium-dark-skin-tone} {} {Old~man:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f474^^^^^^01f3ff} {old-man-dark-skin-tone} {} {Old~man:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f475} {old-woman} {older-woman} {Old~woman} {0.6}
\__emoji_def:nnnnn {^^^^^^01f475^^^^^^01f3fb} {old-woman-light-skin-tone} {} {Old~woman:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f475^^^^^^01f3fc} {old-woman-medium-light-skin-tone} {} {Old~woman:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f475^^^^^^01f3fd} {old-woman-medium-skin-tone} {} {Old~woman:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f475^^^^^^01f3fe} {old-woman-medium-dark-skin-tone} {} {Old~woman:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f475^^^^^^01f3ff} {old-woman-dark-skin-tone} {} {Old~woman:~dark~skin~tone} {1.0}
\__emoji_subgroup:n {Person-gesture}
\__emoji_def:nnnnn {^^^^^^01f64d} {person-frowning} {frowning-person} {Person~frowning} {0.6}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fb} {person-frowning-light-skin-tone} {} {Person~frowning:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fc} {person-frowning-medium-light-skin-tone} {} {Person~frowning:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fd} {person-frowning-medium-skin-tone} {} {Person~frowning:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fe} {person-frowning-medium-dark-skin-tone} {} {Person~frowning:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3ff} {person-frowning-dark-skin-tone} {} {Person~frowning:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^200d^^^^2642^^^^fe0f} {man-frowning} {frowning-man} {Man~frowning} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-frowning-light-skin-tone} {} {Man~frowning:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-frowning-medium-light-skin-tone} {} {Man~frowning:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-frowning-medium-skin-tone} {} {Man~frowning:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-frowning-medium-dark-skin-tone} {} {Man~frowning:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-frowning-dark-skin-tone} {} {Man~frowning:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^200d^^^^2640^^^^fe0f} {woman-frowning} {frowning-woman} {Woman~frowning} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-frowning-light-skin-tone} {} {Woman~frowning:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-frowning-medium-light-skin-tone} {} {Woman~frowning:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-frowning-medium-skin-tone} {} {Woman~frowning:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-frowning-medium-dark-skin-tone} {} {Woman~frowning:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64d^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-frowning-dark-skin-tone} {} {Woman~frowning:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e} {person-pouting} {} {Person~pouting} {0.6}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fb} {person-pouting-light-skin-tone} {} {Person~pouting:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fc} {person-pouting-medium-light-skin-tone} {} {Person~pouting:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fd} {person-pouting-medium-skin-tone} {} {Person~pouting:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fe} {person-pouting-medium-dark-skin-tone} {} {Person~pouting:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3ff} {person-pouting-dark-skin-tone} {} {Person~pouting:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^200d^^^^2642^^^^fe0f} {man-pouting} {pouting-man} {Man~pouting} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-pouting-light-skin-tone} {} {Man~pouting:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-pouting-medium-light-skin-tone} {} {Man~pouting:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-pouting-medium-skin-tone} {} {Man~pouting:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-pouting-medium-dark-skin-tone} {} {Man~pouting:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-pouting-dark-skin-tone} {} {Man~pouting:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^200d^^^^2640^^^^fe0f} {woman-pouting} {pouting-woman} {Woman~pouting} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-pouting-light-skin-tone} {} {Woman~pouting:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-pouting-medium-light-skin-tone} {} {Woman~pouting:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-pouting-medium-skin-tone} {} {Woman~pouting:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-pouting-medium-dark-skin-tone} {} {Woman~pouting:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64e^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-pouting-dark-skin-tone} {} {Woman~pouting:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645} {person-gesturing-no} {no-good} {Person~gesturing~NO} {0.6}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fb} {person-gesturing-no-light-skin-tone} {} {Person~gesturing~NO:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fc} {person-gesturing-no-medium-light-skin-tone} {} {Person~gesturing~NO:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fd} {person-gesturing-no-medium-skin-tone} {} {Person~gesturing~NO:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fe} {person-gesturing-no-medium-dark-skin-tone} {} {Person~gesturing~NO:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3ff} {person-gesturing-no-dark-skin-tone} {} {Person~gesturing~NO:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^200d^^^^2642^^^^fe0f} {man-gesturing-no} {ng-man, no-good-man} {Man~gesturing~NO} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-gesturing-no-light-skin-tone} {} {Man~gesturing~NO:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-gesturing-no-medium-light-skin-tone} {} {Man~gesturing~NO:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-gesturing-no-medium-skin-tone} {} {Man~gesturing~NO:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-gesturing-no-medium-dark-skin-tone} {} {Man~gesturing~NO:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-gesturing-no-dark-skin-tone} {} {Man~gesturing~NO:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-no} {ng-woman, no-good-woman} {Woman~gesturing~NO} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-no-light-skin-tone} {} {Woman~gesturing~NO:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-no-medium-light-skin-tone} {} {Woman~gesturing~NO:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-no-medium-skin-tone} {} {Woman~gesturing~NO:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-no-medium-dark-skin-tone} {} {Woman~gesturing~NO:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f645^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-no-dark-skin-tone} {} {Woman~gesturing~NO:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646} {person-gesturing-ok} {ok-person} {Person~gesturing~OK} {0.6}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fb} {person-gesturing-ok-light-skin-tone} {} {Person~gesturing~OK:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fc} {person-gesturing-ok-medium-light-skin-tone} {} {Person~gesturing~OK:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fd} {person-gesturing-ok-medium-skin-tone} {} {Person~gesturing~OK:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fe} {person-gesturing-ok-medium-dark-skin-tone} {} {Person~gesturing~OK:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3ff} {person-gesturing-ok-dark-skin-tone} {} {Person~gesturing~OK:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^200d^^^^2642^^^^fe0f} {man-gesturing-ok} {ok-man} {Man~gesturing~OK} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-gesturing-ok-light-skin-tone} {} {Man~gesturing~OK:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-gesturing-ok-medium-light-skin-tone} {} {Man~gesturing~OK:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-gesturing-ok-medium-skin-tone} {} {Man~gesturing~OK:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-gesturing-ok-medium-dark-skin-tone} {} {Man~gesturing~OK:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-gesturing-ok-dark-skin-tone} {} {Man~gesturing~OK:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-ok} {ok-woman} {Woman~gesturing~OK} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-ok-light-skin-tone} {} {Woman~gesturing~OK:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-ok-medium-light-skin-tone} {} {Woman~gesturing~OK:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-ok-medium-skin-tone} {} {Woman~gesturing~OK:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-ok-medium-dark-skin-tone} {} {Woman~gesturing~OK:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f646^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-gesturing-ok-dark-skin-tone} {} {Woman~gesturing~OK:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481} {person-tipping-hand} {information-desk-person, tipping-hand-person} {Person~tipping~hand} {0.6}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fb} {person-tipping-hand-light-skin-tone} {} {Person~tipping~hand:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fc} {person-tipping-hand-medium-light-skin-tone} {} {Person~tipping~hand:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fd} {person-tipping-hand-medium-skin-tone} {} {Person~tipping~hand:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fe} {person-tipping-hand-medium-dark-skin-tone} {} {Person~tipping~hand:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3ff} {person-tipping-hand-dark-skin-tone} {} {Person~tipping~hand:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^200d^^^^2642^^^^fe0f} {man-tipping-hand} {sassy-man, tipping-hand-man} {Man~tipping~hand} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-tipping-hand-light-skin-tone} {} {Man~tipping~hand:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-tipping-hand-medium-light-skin-tone} {} {Man~tipping~hand:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-tipping-hand-medium-skin-tone} {} {Man~tipping~hand:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-tipping-hand-medium-dark-skin-tone} {} {Man~tipping~hand:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-tipping-hand-dark-skin-tone} {} {Man~tipping~hand:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^200d^^^^2640^^^^fe0f} {woman-tipping-hand} {sassy-woman, tipping-hand-woman} {Woman~tipping~hand} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-tipping-hand-light-skin-tone} {} {Woman~tipping~hand:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-tipping-hand-medium-light-skin-tone} {} {Woman~tipping~hand:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-tipping-hand-medium-skin-tone} {} {Woman~tipping~hand:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-tipping-hand-medium-dark-skin-tone} {} {Woman~tipping~hand:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f481^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-tipping-hand-dark-skin-tone} {} {Woman~tipping~hand:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b} {person-raising-hand} {raising-hand} {Person~raising~hand} {0.6}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fb} {person-raising-hand-light-skin-tone} {} {Person~raising~hand:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fc} {person-raising-hand-medium-light-skin-tone} {} {Person~raising~hand:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fd} {person-raising-hand-medium-skin-tone} {} {Person~raising~hand:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fe} {person-raising-hand-medium-dark-skin-tone} {} {Person~raising~hand:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3ff} {person-raising-hand-dark-skin-tone} {} {Person~raising~hand:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^200d^^^^2642^^^^fe0f} {man-raising-hand} {raising-hand-man} {Man~raising~hand} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-raising-hand-light-skin-tone} {} {Man~raising~hand:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-raising-hand-medium-light-skin-tone} {} {Man~raising~hand:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-raising-hand-medium-skin-tone} {} {Man~raising~hand:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-raising-hand-medium-dark-skin-tone} {} {Man~raising~hand:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-raising-hand-dark-skin-tone} {} {Man~raising~hand:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^200d^^^^2640^^^^fe0f} {woman-raising-hand} {raising-hand-woman} {Woman~raising~hand} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-raising-hand-light-skin-tone} {} {Woman~raising~hand:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-raising-hand-medium-light-skin-tone} {} {Woman~raising~hand:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-raising-hand-medium-skin-tone} {} {Woman~raising~hand:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-raising-hand-medium-dark-skin-tone} {} {Woman~raising~hand:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f64b^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-raising-hand-dark-skin-tone} {} {Woman~raising~hand:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9cf} {deaf-person} {} {Deaf~person} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fb} {deaf-person-light-skin-tone} {} {Deaf~person:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fc} {deaf-person-medium-light-skin-tone} {} {Deaf~person:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fd} {deaf-person-medium-skin-tone} {} {Deaf~person:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fe} {deaf-person-medium-dark-skin-tone} {} {Deaf~person:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3ff} {deaf-person-dark-skin-tone} {} {Deaf~person:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^200d^^^^2642^^^^fe0f} {deaf-man} {} {Deaf~man} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {deaf-man-light-skin-tone} {} {Deaf~man:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {deaf-man-medium-light-skin-tone} {} {Deaf~man:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {deaf-man-medium-skin-tone} {} {Deaf~man:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {deaf-man-medium-dark-skin-tone} {} {Deaf~man:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {deaf-man-dark-skin-tone} {} {Deaf~man:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^200d^^^^2640^^^^fe0f} {deaf-woman} {} {Deaf~woman} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {deaf-woman-light-skin-tone} {} {Deaf~woman:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {deaf-woman-medium-light-skin-tone} {} {Deaf~woman:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {deaf-woman-medium-skin-tone} {} {Deaf~woman:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {deaf-woman-medium-dark-skin-tone} {} {Deaf~woman:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cf^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {deaf-woman-dark-skin-tone} {} {Deaf~woman:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f647} {person-bowing} {bow} {Person~bowing} {0.6}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fb} {person-bowing-light-skin-tone} {} {Person~bowing:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fc} {person-bowing-medium-light-skin-tone} {} {Person~bowing:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fd} {person-bowing-medium-skin-tone} {} {Person~bowing:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fe} {person-bowing-medium-dark-skin-tone} {} {Person~bowing:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3ff} {person-bowing-dark-skin-tone} {} {Person~bowing:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^200d^^^^2642^^^^fe0f} {man-bowing} {bowing-man} {Man~bowing} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-bowing-light-skin-tone} {} {Man~bowing:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-bowing-medium-light-skin-tone} {} {Man~bowing:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-bowing-medium-skin-tone} {} {Man~bowing:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-bowing-medium-dark-skin-tone} {} {Man~bowing:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-bowing-dark-skin-tone} {} {Man~bowing:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^200d^^^^2640^^^^fe0f} {woman-bowing} {bowing-woman} {Woman~bowing} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-bowing-light-skin-tone} {} {Woman~bowing:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-bowing-medium-light-skin-tone} {} {Woman~bowing:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-bowing-medium-skin-tone} {} {Woman~bowing:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-bowing-medium-dark-skin-tone} {} {Woman~bowing:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f647^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-bowing-dark-skin-tone} {} {Woman~bowing:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926} {person-facepalming} {facepalm} {Person~facepalming} {3.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fb} {person-facepalming-light-skin-tone} {} {Person~facepalming:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fc} {person-facepalming-medium-light-skin-tone} {} {Person~facepalming:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fd} {person-facepalming-medium-skin-tone} {} {Person~facepalming:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fe} {person-facepalming-medium-dark-skin-tone} {} {Person~facepalming:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3ff} {person-facepalming-dark-skin-tone} {} {Person~facepalming:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^200d^^^^2642^^^^fe0f} {man-facepalming} {} {Man~facepalming} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-facepalming-light-skin-tone} {} {Man~facepalming:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-facepalming-medium-light-skin-tone} {} {Man~facepalming:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-facepalming-medium-skin-tone} {} {Man~facepalming:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-facepalming-medium-dark-skin-tone} {} {Man~facepalming:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-facepalming-dark-skin-tone} {} {Man~facepalming:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^200d^^^^2640^^^^fe0f} {woman-facepalming} {} {Woman~facepalming} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-facepalming-light-skin-tone} {} {Woman~facepalming:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-facepalming-medium-light-skin-tone} {} {Woman~facepalming:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-facepalming-medium-skin-tone} {} {Woman~facepalming:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-facepalming-medium-dark-skin-tone} {} {Woman~facepalming:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f926^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-facepalming-dark-skin-tone} {} {Woman~facepalming:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937} {person-shrugging} {shrug} {Person~shrugging} {3.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fb} {person-shrugging-light-skin-tone} {} {Person~shrugging:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fc} {person-shrugging-medium-light-skin-tone} {} {Person~shrugging:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fd} {person-shrugging-medium-skin-tone} {} {Person~shrugging:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fe} {person-shrugging-medium-dark-skin-tone} {} {Person~shrugging:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3ff} {person-shrugging-dark-skin-tone} {} {Person~shrugging:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^200d^^^^2642^^^^fe0f} {man-shrugging} {} {Man~shrugging} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-shrugging-light-skin-tone} {} {Man~shrugging:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-shrugging-medium-light-skin-tone} {} {Man~shrugging:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-shrugging-medium-skin-tone} {} {Man~shrugging:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-shrugging-medium-dark-skin-tone} {} {Man~shrugging:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-shrugging-dark-skin-tone} {} {Man~shrugging:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^200d^^^^2640^^^^fe0f} {woman-shrugging} {} {Woman~shrugging} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-shrugging-light-skin-tone} {} {Woman~shrugging:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-shrugging-medium-light-skin-tone} {} {Woman~shrugging:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-shrugging-medium-skin-tone} {} {Woman~shrugging:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-shrugging-medium-dark-skin-tone} {} {Woman~shrugging:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f937^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-shrugging-dark-skin-tone} {} {Woman~shrugging:~dark~skin~tone} {4.0}
\__emoji_subgroup:n {Person-role}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^2695^^^^fe0f} {health-worker} {} {Health~worker} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2695^^^^fe0f} {health-worker-light-skin-tone} {} {Health~worker:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2695^^^^fe0f} {health-worker-medium-light-skin-tone} {} {Health~worker:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2695^^^^fe0f} {health-worker-medium-skin-tone} {} {Health~worker:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2695^^^^fe0f} {health-worker-medium-dark-skin-tone} {} {Health~worker:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2695^^^^fe0f} {health-worker-dark-skin-tone} {} {Health~worker:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^2695^^^^fe0f} {man-health-worker} {} {Man~health~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2695^^^^fe0f} {man-health-worker-light-skin-tone} {} {Man~health~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2695^^^^fe0f} {man-health-worker-medium-light-skin-tone} {} {Man~health~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2695^^^^fe0f} {man-health-worker-medium-skin-tone} {} {Man~health~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2695^^^^fe0f} {man-health-worker-medium-dark-skin-tone} {} {Man~health~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2695^^^^fe0f} {man-health-worker-dark-skin-tone} {} {Man~health~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^2695^^^^fe0f} {woman-health-worker} {} {Woman~health~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2695^^^^fe0f} {woman-health-worker-light-skin-tone} {} {Woman~health~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2695^^^^fe0f} {woman-health-worker-medium-light-skin-tone} {} {Woman~health~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2695^^^^fe0f} {woman-health-worker-medium-skin-tone} {} {Woman~health~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2695^^^^fe0f} {woman-health-worker-medium-dark-skin-tone} {} {Woman~health~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2695^^^^fe0f} {woman-health-worker-dark-skin-tone} {} {Woman~health~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f393} {student} {} {Student} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f393} {student-light-skin-tone} {} {Student:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f393} {student-medium-light-skin-tone} {} {Student:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f393} {student-medium-skin-tone} {} {Student:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f393} {student-medium-dark-skin-tone} {} {Student:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f393} {student-dark-skin-tone} {} {Student:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f393} {man-student} {} {Man~student} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f393} {man-student-light-skin-tone} {} {Man~student:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f393} {man-student-medium-light-skin-tone} {} {Man~student:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f393} {man-student-medium-skin-tone} {} {Man~student:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f393} {man-student-medium-dark-skin-tone} {} {Man~student:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f393} {man-student-dark-skin-tone} {} {Man~student:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f393} {woman-student} {} {Woman~student} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f393} {woman-student-light-skin-tone} {} {Woman~student:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f393} {woman-student-medium-light-skin-tone} {} {Woman~student:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f393} {woman-student-medium-skin-tone} {} {Woman~student:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f393} {woman-student-medium-dark-skin-tone} {} {Woman~student:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f393} {woman-student-dark-skin-tone} {} {Woman~student:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f3eb} {teacher} {} {Teacher} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f3eb} {teacher-light-skin-tone} {} {Teacher:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f3eb} {teacher-medium-light-skin-tone} {} {Teacher:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f3eb} {teacher-medium-skin-tone} {} {Teacher:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f3eb} {teacher-medium-dark-skin-tone} {} {Teacher:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f3eb} {teacher-dark-skin-tone} {} {Teacher:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f3eb} {man-teacher} {} {Man~teacher} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f3eb} {man-teacher-light-skin-tone} {} {Man~teacher:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f3eb} {man-teacher-medium-light-skin-tone} {} {Man~teacher:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f3eb} {man-teacher-medium-skin-tone} {} {Man~teacher:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f3eb} {man-teacher-medium-dark-skin-tone} {} {Man~teacher:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f3eb} {man-teacher-dark-skin-tone} {} {Man~teacher:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f3eb} {woman-teacher} {} {Woman~teacher} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f3eb} {woman-teacher-light-skin-tone} {} {Woman~teacher:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f3eb} {woman-teacher-medium-light-skin-tone} {} {Woman~teacher:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f3eb} {woman-teacher-medium-skin-tone} {} {Woman~teacher:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f3eb} {woman-teacher-medium-dark-skin-tone} {} {Woman~teacher:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f3eb} {woman-teacher-dark-skin-tone} {} {Woman~teacher:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^2696^^^^fe0f} {judge} {} {Judge} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2696^^^^fe0f} {judge-light-skin-tone} {} {Judge:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2696^^^^fe0f} {judge-medium-light-skin-tone} {} {Judge:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2696^^^^fe0f} {judge-medium-skin-tone} {} {Judge:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2696^^^^fe0f} {judge-medium-dark-skin-tone} {} {Judge:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2696^^^^fe0f} {judge-dark-skin-tone} {} {Judge:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^2696^^^^fe0f} {man-judge} {} {Man~judge} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2696^^^^fe0f} {man-judge-light-skin-tone} {} {Man~judge:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2696^^^^fe0f} {man-judge-medium-light-skin-tone} {} {Man~judge:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2696^^^^fe0f} {man-judge-medium-skin-tone} {} {Man~judge:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2696^^^^fe0f} {man-judge-medium-dark-skin-tone} {} {Man~judge:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2696^^^^fe0f} {man-judge-dark-skin-tone} {} {Man~judge:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^2696^^^^fe0f} {woman-judge} {} {Woman~judge} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2696^^^^fe0f} {woman-judge-light-skin-tone} {} {Woman~judge:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2696^^^^fe0f} {woman-judge-medium-light-skin-tone} {} {Woman~judge:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2696^^^^fe0f} {woman-judge-medium-skin-tone} {} {Woman~judge:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2696^^^^fe0f} {woman-judge-medium-dark-skin-tone} {} {Woman~judge:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2696^^^^fe0f} {woman-judge-dark-skin-tone} {} {Woman~judge:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f33e} {farmer} {} {Farmer} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f33e} {farmer-light-skin-tone} {} {Farmer:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f33e} {farmer-medium-light-skin-tone} {} {Farmer:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f33e} {farmer-medium-skin-tone} {} {Farmer:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f33e} {farmer-medium-dark-skin-tone} {} {Farmer:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f33e} {farmer-dark-skin-tone} {} {Farmer:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f33e} {man-farmer} {} {Man~farmer} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f33e} {man-farmer-light-skin-tone} {} {Man~farmer:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f33e} {man-farmer-medium-light-skin-tone} {} {Man~farmer:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f33e} {man-farmer-medium-skin-tone} {} {Man~farmer:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f33e} {man-farmer-medium-dark-skin-tone} {} {Man~farmer:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f33e} {man-farmer-dark-skin-tone} {} {Man~farmer:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f33e} {woman-farmer} {} {Woman~farmer} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f33e} {woman-farmer-light-skin-tone} {} {Woman~farmer:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f33e} {woman-farmer-medium-light-skin-tone} {} {Woman~farmer:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f33e} {woman-farmer-medium-skin-tone} {} {Woman~farmer:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f33e} {woman-farmer-medium-dark-skin-tone} {} {Woman~farmer:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f33e} {woman-farmer-dark-skin-tone} {} {Woman~farmer:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f373} {cook} {} {Cook} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f373} {cook-light-skin-tone} {} {Cook:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f373} {cook-medium-light-skin-tone} {} {Cook:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f373} {cook-medium-skin-tone} {} {Cook:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f373} {cook-medium-dark-skin-tone} {} {Cook:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f373} {cook-dark-skin-tone} {} {Cook:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f373} {man-cook} {} {Man~cook} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f373} {man-cook-light-skin-tone} {} {Man~cook:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f373} {man-cook-medium-light-skin-tone} {} {Man~cook:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f373} {man-cook-medium-skin-tone} {} {Man~cook:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f373} {man-cook-medium-dark-skin-tone} {} {Man~cook:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f373} {man-cook-dark-skin-tone} {} {Man~cook:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f373} {woman-cook} {} {Woman~cook} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f373} {woman-cook-light-skin-tone} {} {Woman~cook:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f373} {woman-cook-medium-light-skin-tone} {} {Woman~cook:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f373} {woman-cook-medium-skin-tone} {} {Woman~cook:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f373} {woman-cook-medium-dark-skin-tone} {} {Woman~cook:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f373} {woman-cook-dark-skin-tone} {} {Woman~cook:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f527} {mechanic} {} {Mechanic} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f527} {mechanic-light-skin-tone} {} {Mechanic:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f527} {mechanic-medium-light-skin-tone} {} {Mechanic:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f527} {mechanic-medium-skin-tone} {} {Mechanic:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f527} {mechanic-medium-dark-skin-tone} {} {Mechanic:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f527} {mechanic-dark-skin-tone} {} {Mechanic:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f527} {man-mechanic} {} {Man~mechanic} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f527} {man-mechanic-light-skin-tone} {} {Man~mechanic:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f527} {man-mechanic-medium-light-skin-tone} {} {Man~mechanic:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f527} {man-mechanic-medium-skin-tone} {} {Man~mechanic:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f527} {man-mechanic-medium-dark-skin-tone} {} {Man~mechanic:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f527} {man-mechanic-dark-skin-tone} {} {Man~mechanic:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f527} {woman-mechanic} {} {Woman~mechanic} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f527} {woman-mechanic-light-skin-tone} {} {Woman~mechanic:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f527} {woman-mechanic-medium-light-skin-tone} {} {Woman~mechanic:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f527} {woman-mechanic-medium-skin-tone} {} {Woman~mechanic:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f527} {woman-mechanic-medium-dark-skin-tone} {} {Woman~mechanic:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f527} {woman-mechanic-dark-skin-tone} {} {Woman~mechanic:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f3ed} {factory-worker} {} {Factory~worker} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f3ed} {factory-worker-light-skin-tone} {} {Factory~worker:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f3ed} {factory-worker-medium-light-skin-tone} {} {Factory~worker:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f3ed} {factory-worker-medium-skin-tone} {} {Factory~worker:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f3ed} {factory-worker-medium-dark-skin-tone} {} {Factory~worker:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f3ed} {factory-worker-dark-skin-tone} {} {Factory~worker:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f3ed} {man-factory-worker} {} {Man~factory~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f3ed} {man-factory-worker-light-skin-tone} {} {Man~factory~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f3ed} {man-factory-worker-medium-light-skin-tone} {} {Man~factory~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f3ed} {man-factory-worker-medium-skin-tone} {} {Man~factory~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f3ed} {man-factory-worker-medium-dark-skin-tone} {} {Man~factory~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f3ed} {man-factory-worker-dark-skin-tone} {} {Man~factory~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f3ed} {woman-factory-worker} {} {Woman~factory~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f3ed} {woman-factory-worker-light-skin-tone} {} {Woman~factory~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f3ed} {woman-factory-worker-medium-light-skin-tone} {} {Woman~factory~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f3ed} {woman-factory-worker-medium-skin-tone} {} {Woman~factory~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f3ed} {woman-factory-worker-medium-dark-skin-tone} {} {Woman~factory~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f3ed} {woman-factory-worker-dark-skin-tone} {} {Woman~factory~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f4bc} {office-worker} {} {Office~worker} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f4bc} {office-worker-light-skin-tone} {} {Office~worker:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f4bc} {office-worker-medium-light-skin-tone} {} {Office~worker:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f4bc} {office-worker-medium-skin-tone} {} {Office~worker:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f4bc} {office-worker-medium-dark-skin-tone} {} {Office~worker:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f4bc} {office-worker-dark-skin-tone} {} {Office~worker:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f4bc} {man-office-worker} {} {Man~office~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f4bc} {man-office-worker-light-skin-tone} {} {Man~office~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f4bc} {man-office-worker-medium-light-skin-tone} {} {Man~office~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f4bc} {man-office-worker-medium-skin-tone} {} {Man~office~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f4bc} {man-office-worker-medium-dark-skin-tone} {} {Man~office~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f4bc} {man-office-worker-dark-skin-tone} {} {Man~office~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f4bc} {woman-office-worker} {} {Woman~office~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f4bc} {woman-office-worker-light-skin-tone} {} {Woman~office~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f4bc} {woman-office-worker-medium-light-skin-tone} {} {Woman~office~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f4bc} {woman-office-worker-medium-skin-tone} {} {Woman~office~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f4bc} {woman-office-worker-medium-dark-skin-tone} {} {Woman~office~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f4bc} {woman-office-worker-dark-skin-tone} {} {Woman~office~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f52c} {scientist} {} {Scientist} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f52c} {scientist-light-skin-tone} {} {Scientist:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f52c} {scientist-medium-light-skin-tone} {} {Scientist:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f52c} {scientist-medium-skin-tone} {} {Scientist:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f52c} {scientist-medium-dark-skin-tone} {} {Scientist:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f52c} {scientist-dark-skin-tone} {} {Scientist:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f52c} {man-scientist} {} {Man~scientist} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f52c} {man-scientist-light-skin-tone} {} {Man~scientist:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f52c} {man-scientist-medium-light-skin-tone} {} {Man~scientist:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f52c} {man-scientist-medium-skin-tone} {} {Man~scientist:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f52c} {man-scientist-medium-dark-skin-tone} {} {Man~scientist:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f52c} {man-scientist-dark-skin-tone} {} {Man~scientist:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f52c} {woman-scientist} {} {Woman~scientist} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f52c} {woman-scientist-light-skin-tone} {} {Woman~scientist:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f52c} {woman-scientist-medium-light-skin-tone} {} {Woman~scientist:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f52c} {woman-scientist-medium-skin-tone} {} {Woman~scientist:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f52c} {woman-scientist-medium-dark-skin-tone} {} {Woman~scientist:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f52c} {woman-scientist-dark-skin-tone} {} {Woman~scientist:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f4bb} {technologist} {} {Technologist} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f4bb} {technologist-light-skin-tone} {} {Technologist:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f4bb} {technologist-medium-light-skin-tone} {} {Technologist:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f4bb} {technologist-medium-skin-tone} {} {Technologist:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f4bb} {technologist-medium-dark-skin-tone} {} {Technologist:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f4bb} {technologist-dark-skin-tone} {} {Technologist:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f4bb} {man-technologist} {} {Man~technologist} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f4bb} {man-technologist-light-skin-tone} {} {Man~technologist:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f4bb} {man-technologist-medium-light-skin-tone} {} {Man~technologist:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f4bb} {man-technologist-medium-skin-tone} {} {Man~technologist:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f4bb} {man-technologist-medium-dark-skin-tone} {} {Man~technologist:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f4bb} {man-technologist-dark-skin-tone} {} {Man~technologist:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f4bb} {woman-technologist} {} {Woman~technologist} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f4bb} {woman-technologist-light-skin-tone} {} {Woman~technologist:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f4bb} {woman-technologist-medium-light-skin-tone} {} {Woman~technologist:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f4bb} {woman-technologist-medium-skin-tone} {} {Woman~technologist:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f4bb} {woman-technologist-medium-dark-skin-tone} {} {Woman~technologist:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f4bb} {woman-technologist-dark-skin-tone} {} {Woman~technologist:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f3a4} {singer} {} {Singer} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f3a4} {singer-light-skin-tone} {} {Singer:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f3a4} {singer-medium-light-skin-tone} {} {Singer:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f3a4} {singer-medium-skin-tone} {} {Singer:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f3a4} {singer-medium-dark-skin-tone} {} {Singer:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f3a4} {singer-dark-skin-tone} {} {Singer:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f3a4} {man-singer} {} {Man~singer} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f3a4} {man-singer-light-skin-tone} {} {Man~singer:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f3a4} {man-singer-medium-light-skin-tone} {} {Man~singer:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f3a4} {man-singer-medium-skin-tone} {} {Man~singer:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f3a4} {man-singer-medium-dark-skin-tone} {} {Man~singer:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f3a4} {man-singer-dark-skin-tone} {} {Man~singer:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f3a4} {woman-singer} {} {Woman~singer} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f3a4} {woman-singer-light-skin-tone} {} {Woman~singer:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f3a4} {woman-singer-medium-light-skin-tone} {} {Woman~singer:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f3a4} {woman-singer-medium-skin-tone} {} {Woman~singer:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f3a4} {woman-singer-medium-dark-skin-tone} {} {Woman~singer:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f3a4} {woman-singer-dark-skin-tone} {} {Woman~singer:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f3a8} {artist} {} {Artist} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f3a8} {artist-light-skin-tone} {} {Artist:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f3a8} {artist-medium-light-skin-tone} {} {Artist:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f3a8} {artist-medium-skin-tone} {} {Artist:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f3a8} {artist-medium-dark-skin-tone} {} {Artist:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f3a8} {artist-dark-skin-tone} {} {Artist:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f3a8} {man-artist} {} {Man~artist} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f3a8} {man-artist-light-skin-tone} {} {Man~artist:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f3a8} {man-artist-medium-light-skin-tone} {} {Man~artist:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f3a8} {man-artist-medium-skin-tone} {} {Man~artist:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f3a8} {man-artist-medium-dark-skin-tone} {} {Man~artist:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f3a8} {man-artist-dark-skin-tone} {} {Man~artist:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f3a8} {woman-artist} {} {Woman~artist} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f3a8} {woman-artist-light-skin-tone} {} {Woman~artist:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f3a8} {woman-artist-medium-light-skin-tone} {} {Woman~artist:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f3a8} {woman-artist-medium-skin-tone} {} {Woman~artist:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f3a8} {woman-artist-medium-dark-skin-tone} {} {Woman~artist:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f3a8} {woman-artist-dark-skin-tone} {} {Woman~artist:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^2708^^^^fe0f} {pilot} {} {Pilot} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2708^^^^fe0f} {pilot-light-skin-tone} {} {Pilot:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2708^^^^fe0f} {pilot-medium-light-skin-tone} {} {Pilot:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2708^^^^fe0f} {pilot-medium-skin-tone} {} {Pilot:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2708^^^^fe0f} {pilot-medium-dark-skin-tone} {} {Pilot:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2708^^^^fe0f} {pilot-dark-skin-tone} {} {Pilot:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^2708^^^^fe0f} {man-pilot} {} {Man~pilot} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2708^^^^fe0f} {man-pilot-light-skin-tone} {} {Man~pilot:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2708^^^^fe0f} {man-pilot-medium-light-skin-tone} {} {Man~pilot:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2708^^^^fe0f} {man-pilot-medium-skin-tone} {} {Man~pilot:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2708^^^^fe0f} {man-pilot-medium-dark-skin-tone} {} {Man~pilot:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2708^^^^fe0f} {man-pilot-dark-skin-tone} {} {Man~pilot:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^2708^^^^fe0f} {woman-pilot} {} {Woman~pilot} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2708^^^^fe0f} {woman-pilot-light-skin-tone} {} {Woman~pilot:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2708^^^^fe0f} {woman-pilot-medium-light-skin-tone} {} {Woman~pilot:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2708^^^^fe0f} {woman-pilot-medium-skin-tone} {} {Woman~pilot:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2708^^^^fe0f} {woman-pilot-medium-dark-skin-tone} {} {Woman~pilot:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2708^^^^fe0f} {woman-pilot-dark-skin-tone} {} {Woman~pilot:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f680} {astronaut} {} {Astronaut} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f680} {astronaut-light-skin-tone} {} {Astronaut:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f680} {astronaut-medium-light-skin-tone} {} {Astronaut:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f680} {astronaut-medium-skin-tone} {} {Astronaut:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f680} {astronaut-medium-dark-skin-tone} {} {Astronaut:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f680} {astronaut-dark-skin-tone} {} {Astronaut:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f680} {man-astronaut} {} {Man~astronaut} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f680} {man-astronaut-light-skin-tone} {} {Man~astronaut:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f680} {man-astronaut-medium-light-skin-tone} {} {Man~astronaut:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f680} {man-astronaut-medium-skin-tone} {} {Man~astronaut:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f680} {man-astronaut-medium-dark-skin-tone} {} {Man~astronaut:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f680} {man-astronaut-dark-skin-tone} {} {Man~astronaut:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f680} {woman-astronaut} {} {Woman~astronaut} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f680} {woman-astronaut-light-skin-tone} {} {Woman~astronaut:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f680} {woman-astronaut-medium-light-skin-tone} {} {Woman~astronaut:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f680} {woman-astronaut-medium-skin-tone} {} {Woman~astronaut:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f680} {woman-astronaut-medium-dark-skin-tone} {} {Woman~astronaut:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f680} {woman-astronaut-dark-skin-tone} {} {Woman~astronaut:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f692} {firefighter} {} {Firefighter} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f692} {firefighter-light-skin-tone} {} {Firefighter:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f692} {firefighter-medium-light-skin-tone} {} {Firefighter:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f692} {firefighter-medium-skin-tone} {} {Firefighter:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f692} {firefighter-medium-dark-skin-tone} {} {Firefighter:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f692} {firefighter-dark-skin-tone} {} {Firefighter:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f692} {man-firefighter} {} {Man~firefighter} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f692} {man-firefighter-light-skin-tone} {} {Man~firefighter:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f692} {man-firefighter-medium-light-skin-tone} {} {Man~firefighter:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f692} {man-firefighter-medium-skin-tone} {} {Man~firefighter:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f692} {man-firefighter-medium-dark-skin-tone} {} {Man~firefighter:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f692} {man-firefighter-dark-skin-tone} {} {Man~firefighter:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f692} {woman-firefighter} {} {Woman~firefighter} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f692} {woman-firefighter-light-skin-tone} {} {Woman~firefighter:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f692} {woman-firefighter-medium-light-skin-tone} {} {Woman~firefighter:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f692} {woman-firefighter-medium-skin-tone} {} {Woman~firefighter:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f692} {woman-firefighter-medium-dark-skin-tone} {} {Woman~firefighter:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f692} {woman-firefighter-dark-skin-tone} {} {Woman~firefighter:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e} {police-officer} {cop} {Police~officer} {0.6}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fb} {police-officer-light-skin-tone} {} {Police~officer:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fc} {police-officer-medium-light-skin-tone} {} {Police~officer:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fd} {police-officer-medium-skin-tone} {} {Police~officer:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fe} {police-officer-medium-dark-skin-tone} {} {Police~officer:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3ff} {police-officer-dark-skin-tone} {} {Police~officer:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^200d^^^^2642^^^^fe0f} {man-police-officer} {policeman} {Man~police~officer} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-police-officer-light-skin-tone} {} {Man~police~officer:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-police-officer-medium-light-skin-tone} {} {Man~police~officer:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-police-officer-medium-skin-tone} {} {Man~police~officer:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-police-officer-medium-dark-skin-tone} {} {Man~police~officer:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-police-officer-dark-skin-tone} {} {Man~police~officer:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^200d^^^^2640^^^^fe0f} {woman-police-officer} {policewoman} {Woman~police~officer} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-police-officer-light-skin-tone} {} {Woman~police~officer:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-police-officer-medium-light-skin-tone} {} {Woman~police~officer:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-police-officer-medium-skin-tone} {} {Woman~police~officer:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-police-officer-medium-dark-skin-tone} {} {Woman~police~officer:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46e^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-police-officer-dark-skin-tone} {} {Woman~police~officer:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^fe0f} {detective} {} {Detective} {0.7}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fb} {detective-light-skin-tone} {} {Detective:~light~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fc} {detective-medium-light-skin-tone} {} {Detective:~medium-light~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fd} {detective-medium-skin-tone} {} {Detective:~medium~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fe} {detective-medium-dark-skin-tone} {} {Detective:~medium-dark~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3ff} {detective-dark-skin-tone} {} {Detective:~dark~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^fe0f^^^^200d^^^^2642^^^^fe0f} {man-detective} {male-detective} {Man~detective} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-detective-light-skin-tone} {} {Man~detective:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-detective-medium-light-skin-tone} {} {Man~detective:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-detective-medium-skin-tone} {} {Man~detective:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-detective-medium-dark-skin-tone} {} {Man~detective:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-detective-dark-skin-tone} {} {Man~detective:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^fe0f^^^^200d^^^^2640^^^^fe0f} {woman-detective} {female-detective} {Woman~detective} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-detective-light-skin-tone} {} {Woman~detective:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-detective-medium-light-skin-tone} {} {Woman~detective:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-detective-medium-skin-tone} {} {Woman~detective:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-detective-medium-dark-skin-tone} {} {Woman~detective:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f575^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-detective-dark-skin-tone} {} {Woman~detective:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482} {guard} {} {Guard} {0.6}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fb} {guard-light-skin-tone} {} {Guard:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fc} {guard-medium-light-skin-tone} {} {Guard:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fd} {guard-medium-skin-tone} {} {Guard:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fe} {guard-medium-dark-skin-tone} {} {Guard:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3ff} {guard-dark-skin-tone} {} {Guard:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^200d^^^^2642^^^^fe0f} {man-guard} {guardsman} {Man~guard} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-guard-light-skin-tone} {} {Man~guard:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-guard-medium-light-skin-tone} {} {Man~guard:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-guard-medium-skin-tone} {} {Man~guard:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-guard-medium-dark-skin-tone} {} {Man~guard:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-guard-dark-skin-tone} {} {Man~guard:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^200d^^^^2640^^^^fe0f} {woman-guard} {guardswoman} {Woman~guard} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-guard-light-skin-tone} {} {Woman~guard:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-guard-medium-light-skin-tone} {} {Woman~guard:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-guard-medium-skin-tone} {} {Woman~guard:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-guard-medium-dark-skin-tone} {} {Woman~guard:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f482^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-guard-dark-skin-tone} {} {Woman~guard:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f977} {ninja} {} {Ninja} {13.0}
\__emoji_def:nnnnn {^^^^^^01f977^^^^^^01f3fb} {ninja-light-skin-tone} {} {Ninja:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f977^^^^^^01f3fc} {ninja-medium-light-skin-tone} {} {Ninja:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f977^^^^^^01f3fd} {ninja-medium-skin-tone} {} {Ninja:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f977^^^^^^01f3fe} {ninja-medium-dark-skin-tone} {} {Ninja:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f977^^^^^^01f3ff} {ninja-dark-skin-tone} {} {Ninja:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f477} {construction-worker} {} {Construction~worker} {0.6}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fb} {construction-worker-light-skin-tone} {} {Construction~worker:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fc} {construction-worker-medium-light-skin-tone} {} {Construction~worker:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fd} {construction-worker-medium-skin-tone} {} {Construction~worker:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fe} {construction-worker-medium-dark-skin-tone} {} {Construction~worker:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3ff} {construction-worker-dark-skin-tone} {} {Construction~worker:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^200d^^^^2642^^^^fe0f} {man-construction-worker} {construction-worker-man} {Man~construction~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-construction-worker-light-skin-tone} {} {Man~construction~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-construction-worker-medium-light-skin-tone} {} {Man~construction~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-construction-worker-medium-skin-tone} {} {Man~construction~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-construction-worker-medium-dark-skin-tone} {} {Man~construction~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-construction-worker-dark-skin-tone} {} {Man~construction~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^200d^^^^2640^^^^fe0f} {woman-construction-worker} {construction-worker-woman} {Woman~construction~worker} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-construction-worker-light-skin-tone} {} {Woman~construction~worker:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-construction-worker-medium-light-skin-tone} {} {Woman~construction~worker:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-construction-worker-medium-skin-tone} {} {Woman~construction~worker:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-construction-worker-medium-dark-skin-tone} {} {Woman~construction~worker:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f477^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-construction-worker-dark-skin-tone} {} {Woman~construction~worker:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f934} {prince} {} {Prince} {3.0}
\__emoji_def:nnnnn {^^^^^^01f934^^^^^^01f3fb} {prince-light-skin-tone} {} {Prince:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f934^^^^^^01f3fc} {prince-medium-light-skin-tone} {} {Prince:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f934^^^^^^01f3fd} {prince-medium-skin-tone} {} {Prince:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f934^^^^^^01f3fe} {prince-medium-dark-skin-tone} {} {Prince:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f934^^^^^^01f3ff} {prince-dark-skin-tone} {} {Prince:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f478} {princess} {} {Princess} {0.6}
\__emoji_def:nnnnn {^^^^^^01f478^^^^^^01f3fb} {princess-light-skin-tone} {} {Princess:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f478^^^^^^01f3fc} {princess-medium-light-skin-tone} {} {Princess:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f478^^^^^^01f3fd} {princess-medium-skin-tone} {} {Princess:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f478^^^^^^01f3fe} {princess-medium-dark-skin-tone} {} {Princess:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f478^^^^^^01f3ff} {princess-dark-skin-tone} {} {Princess:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f473} {person-wearing-turban} {person-with-turban} {Person~wearing~turban} {0.6}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fb} {person-wearing-turban-light-skin-tone} {} {Person~wearing~turban:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fc} {person-wearing-turban-medium-light-skin-tone} {} {Person~wearing~turban:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fd} {person-wearing-turban-medium-skin-tone} {} {Person~wearing~turban:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fe} {person-wearing-turban-medium-dark-skin-tone} {} {Person~wearing~turban:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3ff} {person-wearing-turban-dark-skin-tone} {} {Person~wearing~turban:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^200d^^^^2642^^^^fe0f} {man-wearing-turban} {man-with-turban} {Man~wearing~turban} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-wearing-turban-light-skin-tone} {} {Man~wearing~turban:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-wearing-turban-medium-light-skin-tone} {} {Man~wearing~turban:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-wearing-turban-medium-skin-tone} {} {Man~wearing~turban:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-wearing-turban-medium-dark-skin-tone} {} {Man~wearing~turban:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-wearing-turban-dark-skin-tone} {} {Man~wearing~turban:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^200d^^^^2640^^^^fe0f} {woman-wearing-turban} {woman-with-turban} {Woman~wearing~turban} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-wearing-turban-light-skin-tone} {} {Woman~wearing~turban:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-wearing-turban-medium-light-skin-tone} {} {Woman~wearing~turban:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-wearing-turban-medium-skin-tone} {} {Woman~wearing~turban:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-wearing-turban-medium-dark-skin-tone} {} {Woman~wearing~turban:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f473^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-wearing-turban-dark-skin-tone} {} {Woman~wearing~turban:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f472} {person-with-skullcap} {man-with-gua-pi-mao} {Person~with~skullcap} {0.6}
\__emoji_def:nnnnn {^^^^^^01f472^^^^^^01f3fb} {person-with-skullcap-light-skin-tone} {} {Person~with~skullcap:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f472^^^^^^01f3fc} {person-with-skullcap-medium-light-skin-tone} {} {Person~with~skullcap:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f472^^^^^^01f3fd} {person-with-skullcap-medium-skin-tone} {} {Person~with~skullcap:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f472^^^^^^01f3fe} {person-with-skullcap-medium-dark-skin-tone} {} {Person~with~skullcap:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f472^^^^^^01f3ff} {person-with-skullcap-dark-skin-tone} {} {Person~with~skullcap:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9d5} {woman-with-headscarf} {} {Woman~with~headscarf} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d5^^^^^^01f3fb} {woman-with-headscarf-light-skin-tone} {} {Woman~with~headscarf:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d5^^^^^^01f3fc} {woman-with-headscarf-medium-light-skin-tone} {} {Woman~with~headscarf:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d5^^^^^^01f3fd} {woman-with-headscarf-medium-skin-tone} {} {Woman~with~headscarf:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d5^^^^^^01f3fe} {woman-with-headscarf-medium-dark-skin-tone} {} {Woman~with~headscarf:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d5^^^^^^01f3ff} {woman-with-headscarf-dark-skin-tone} {} {Woman~with~headscarf:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f935} {person-in-tuxedo} {} {Person~in~tuxedo} {3.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fb} {person-in-tuxedo-light-skin-tone} {} {Person~in~tuxedo:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fc} {person-in-tuxedo-medium-light-skin-tone} {} {Person~in~tuxedo:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fd} {person-in-tuxedo-medium-skin-tone} {} {Person~in~tuxedo:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fe} {person-in-tuxedo-medium-dark-skin-tone} {} {Person~in~tuxedo:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3ff} {person-in-tuxedo-dark-skin-tone} {} {Person~in~tuxedo:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^200d^^^^2642^^^^fe0f} {man-in-tuxedo} {} {Man~in~tuxedo} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-in-tuxedo-light-skin-tone} {} {Man~in~tuxedo:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-in-tuxedo-medium-light-skin-tone} {} {Man~in~tuxedo:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-in-tuxedo-medium-skin-tone} {} {Man~in~tuxedo:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-in-tuxedo-medium-dark-skin-tone} {} {Man~in~tuxedo:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-in-tuxedo-dark-skin-tone} {} {Man~in~tuxedo:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^200d^^^^2640^^^^fe0f} {woman-in-tuxedo} {} {Woman~in~tuxedo} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-in-tuxedo-light-skin-tone} {} {Woman~in~tuxedo:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-in-tuxedo-medium-light-skin-tone} {} {Woman~in~tuxedo:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-in-tuxedo-medium-skin-tone} {} {Woman~in~tuxedo:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-in-tuxedo-medium-dark-skin-tone} {} {Woman~in~tuxedo:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f935^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-in-tuxedo-dark-skin-tone} {} {Woman~in~tuxedo:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470} {person-with-veil} {} {Person~with~veil} {0.6}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fb} {person-with-veil-light-skin-tone} {} {Person~with~veil:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fc} {person-with-veil-medium-light-skin-tone} {} {Person~with~veil:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fd} {person-with-veil-medium-skin-tone} {} {Person~with~veil:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fe} {person-with-veil-medium-dark-skin-tone} {} {Person~with~veil:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3ff} {person-with-veil-dark-skin-tone} {} {Person~with~veil:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^200d^^^^2642^^^^fe0f} {man-with-veil} {} {Man~with~veil} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-with-veil-light-skin-tone} {} {Man~with~veil:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-with-veil-medium-light-skin-tone} {} {Man~with~veil:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-with-veil-medium-skin-tone} {} {Man~with~veil:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-with-veil-medium-dark-skin-tone} {} {Man~with~veil:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-with-veil-dark-skin-tone} {} {Man~with~veil:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^200d^^^^2640^^^^fe0f} {woman-with-veil} {bride-with-veil} {Woman~with~veil} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-with-veil-light-skin-tone} {} {Woman~with~veil:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-with-veil-medium-light-skin-tone} {} {Woman~with~veil:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-with-veil-medium-skin-tone} {} {Woman~with~veil:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-with-veil-medium-dark-skin-tone} {} {Woman~with~veil:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f470^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-with-veil-dark-skin-tone} {} {Woman~with~veil:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f930} {pregnant-woman} {} {Pregnant~woman} {3.0}
\__emoji_def:nnnnn {^^^^^^01f930^^^^^^01f3fb} {pregnant-woman-light-skin-tone} {} {Pregnant~woman:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f930^^^^^^01f3fc} {pregnant-woman-medium-light-skin-tone} {} {Pregnant~woman:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f930^^^^^^01f3fd} {pregnant-woman-medium-skin-tone} {} {Pregnant~woman:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f930^^^^^^01f3fe} {pregnant-woman-medium-dark-skin-tone} {} {Pregnant~woman:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f930^^^^^^01f3ff} {pregnant-woman-dark-skin-tone} {} {Pregnant~woman:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f931} {breast-feeding} {} {Breast-feeding} {5.0}
\__emoji_def:nnnnn {^^^^^^01f931^^^^^^01f3fb} {breast-feeding-light-skin-tone} {} {Breast-feeding:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f931^^^^^^01f3fc} {breast-feeding-medium-light-skin-tone} {} {Breast-feeding:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f931^^^^^^01f3fd} {breast-feeding-medium-skin-tone} {} {Breast-feeding:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f931^^^^^^01f3fe} {breast-feeding-medium-dark-skin-tone} {} {Breast-feeding:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f931^^^^^^01f3ff} {breast-feeding-dark-skin-tone} {} {Breast-feeding:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f37c} {woman-feeding-baby} {} {Woman~feeding~baby} {13.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f37c} {woman-feeding-baby-light-skin-tone} {} {Woman~feeding~baby:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f37c} {woman-feeding-baby-medium-light-skin-tone} {} {Woman~feeding~baby:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f37c} {woman-feeding-baby-medium-skin-tone} {} {Woman~feeding~baby:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f37c} {woman-feeding-baby-medium-dark-skin-tone} {} {Woman~feeding~baby:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f37c} {woman-feeding-baby-dark-skin-tone} {} {Woman~feeding~baby:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f37c} {man-feeding-baby} {} {Man~feeding~baby} {13.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f37c} {man-feeding-baby-light-skin-tone} {} {Man~feeding~baby:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f37c} {man-feeding-baby-medium-light-skin-tone} {} {Man~feeding~baby:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f37c} {man-feeding-baby-medium-skin-tone} {} {Man~feeding~baby:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f37c} {man-feeding-baby-medium-dark-skin-tone} {} {Man~feeding~baby:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f37c} {man-feeding-baby-dark-skin-tone} {} {Man~feeding~baby:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f37c} {person-feeding-baby} {} {Person~feeding~baby} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f37c} {person-feeding-baby-light-skin-tone} {} {Person~feeding~baby:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f37c} {person-feeding-baby-medium-light-skin-tone} {} {Person~feeding~baby:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f37c} {person-feeding-baby-medium-skin-tone} {} {Person~feeding~baby:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f37c} {person-feeding-baby-medium-dark-skin-tone} {} {Person~feeding~baby:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f37c} {person-feeding-baby-dark-skin-tone} {} {Person~feeding~baby:~dark~skin~tone} {13.0}
\__emoji_subgroup:n {Person-fantasy}
\__emoji_def:nnnnn {^^^^^^01f47c} {baby-angel} {angel} {Baby~angel} {0.6}
\__emoji_def:nnnnn {^^^^^^01f47c^^^^^^01f3fb} {baby-angel-light-skin-tone} {} {Baby~angel:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f47c^^^^^^01f3fc} {baby-angel-medium-light-skin-tone} {} {Baby~angel:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f47c^^^^^^01f3fd} {baby-angel-medium-skin-tone} {} {Baby~angel:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f47c^^^^^^01f3fe} {baby-angel-medium-dark-skin-tone} {} {Baby~angel:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f47c^^^^^^01f3ff} {baby-angel-dark-skin-tone} {} {Baby~angel:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f385} {santa-claus} {santa} {Santa~Claus} {0.6}
\__emoji_def:nnnnn {^^^^^^01f385^^^^^^01f3fb} {santa-claus-light-skin-tone} {} {Santa~Claus:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f385^^^^^^01f3fc} {santa-claus-medium-light-skin-tone} {} {Santa~Claus:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f385^^^^^^01f3fd} {santa-claus-medium-skin-tone} {} {Santa~Claus:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f385^^^^^^01f3fe} {santa-claus-medium-dark-skin-tone} {} {Santa~Claus:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f385^^^^^^01f3ff} {santa-claus-dark-skin-tone} {} {Santa~Claus:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f936} {mrs-claus} {} {Mrs.~Claus} {3.0}
\__emoji_def:nnnnn {^^^^^^01f936^^^^^^01f3fb} {mrs-claus-light-skin-tone} {} {Mrs.~Claus:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f936^^^^^^01f3fc} {mrs-claus-medium-light-skin-tone} {} {Mrs.~Claus:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f936^^^^^^01f3fd} {mrs-claus-medium-skin-tone} {} {Mrs.~Claus:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f936^^^^^^01f3fe} {mrs-claus-medium-dark-skin-tone} {} {Mrs.~Claus:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f936^^^^^^01f3ff} {mrs-claus-dark-skin-tone} {} {Mrs.~Claus:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f384} {mx-claus} {} {Mx~claus} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f384} {mx-claus-light-skin-tone} {} {Mx~claus:~light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f384} {mx-claus-medium-light-skin-tone} {} {Mx~claus:~medium-light~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f384} {mx-claus-medium-skin-tone} {} {Mx~claus:~medium~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f384} {mx-claus-medium-dark-skin-tone} {} {Mx~claus:~medium-dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f384} {mx-claus-dark-skin-tone} {} {Mx~claus:~dark~skin~tone} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9b8} {superhero} {} {Superhero} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fb} {superhero-light-skin-tone} {} {Superhero:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fc} {superhero-medium-light-skin-tone} {} {Superhero:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fd} {superhero-medium-skin-tone} {} {Superhero:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fe} {superhero-medium-dark-skin-tone} {} {Superhero:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3ff} {superhero-dark-skin-tone} {} {Superhero:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^200d^^^^2642^^^^fe0f} {man-superhero} {superhero-man} {Man~superhero} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-superhero-light-skin-tone} {} {Man~superhero:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-superhero-medium-light-skin-tone} {} {Man~superhero:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-superhero-medium-skin-tone} {} {Man~superhero:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-superhero-medium-dark-skin-tone} {} {Man~superhero:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-superhero-dark-skin-tone} {} {Man~superhero:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^200d^^^^2640^^^^fe0f} {woman-superhero} {superhero-woman} {Woman~superhero} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-superhero-light-skin-tone} {} {Woman~superhero:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-superhero-medium-light-skin-tone} {} {Woman~superhero:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-superhero-medium-skin-tone} {} {Woman~superhero:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-superhero-medium-dark-skin-tone} {} {Woman~superhero:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b8^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-superhero-dark-skin-tone} {} {Woman~superhero:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9} {supervillain} {} {Supervillain} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fb} {supervillain-light-skin-tone} {} {Supervillain:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fc} {supervillain-medium-light-skin-tone} {} {Supervillain:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fd} {supervillain-medium-skin-tone} {} {Supervillain:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fe} {supervillain-medium-dark-skin-tone} {} {Supervillain:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3ff} {supervillain-dark-skin-tone} {} {Supervillain:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^200d^^^^2642^^^^fe0f} {man-supervillain} {supervillain-man} {Man~supervillain} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-supervillain-light-skin-tone} {} {Man~supervillain:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-supervillain-medium-light-skin-tone} {} {Man~supervillain:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-supervillain-medium-skin-tone} {} {Man~supervillain:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-supervillain-medium-dark-skin-tone} {} {Man~supervillain:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-supervillain-dark-skin-tone} {} {Man~supervillain:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^200d^^^^2640^^^^fe0f} {woman-supervillain} {supervillain-woman} {Woman~supervillain} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-supervillain-light-skin-tone} {} {Woman~supervillain:~light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-supervillain-medium-light-skin-tone} {} {Woman~supervillain:~medium-light~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-supervillain-medium-skin-tone} {} {Woman~supervillain:~medium~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-supervillain-medium-dark-skin-tone} {} {Woman~supervillain:~medium-dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b9^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-supervillain-dark-skin-tone} {} {Woman~supervillain:~dark~skin~tone} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9d9} {mage} {} {Mage} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fb} {mage-light-skin-tone} {} {Mage:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fc} {mage-medium-light-skin-tone} {} {Mage:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fd} {mage-medium-skin-tone} {} {Mage:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fe} {mage-medium-dark-skin-tone} {} {Mage:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3ff} {mage-dark-skin-tone} {} {Mage:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^200d^^^^2642^^^^fe0f} {man-mage} {mage-man} {Man~mage} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-mage-light-skin-tone} {} {Man~mage:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-mage-medium-light-skin-tone} {} {Man~mage:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-mage-medium-skin-tone} {} {Man~mage:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-mage-medium-dark-skin-tone} {} {Man~mage:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-mage-dark-skin-tone} {} {Man~mage:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^200d^^^^2640^^^^fe0f} {woman-mage} {mage-woman} {Woman~mage} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-mage-light-skin-tone} {} {Woman~mage:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-mage-medium-light-skin-tone} {} {Woman~mage:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-mage-medium-skin-tone} {} {Woman~mage:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-mage-medium-dark-skin-tone} {} {Woman~mage:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d9^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-mage-dark-skin-tone} {} {Woman~mage:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da} {fairy} {} {Fairy} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fb} {fairy-light-skin-tone} {} {Fairy:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fc} {fairy-medium-light-skin-tone} {} {Fairy:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fd} {fairy-medium-skin-tone} {} {Fairy:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fe} {fairy-medium-dark-skin-tone} {} {Fairy:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3ff} {fairy-dark-skin-tone} {} {Fairy:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^200d^^^^2642^^^^fe0f} {man-fairy} {fairy-man} {Man~fairy} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-fairy-light-skin-tone} {} {Man~fairy:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-fairy-medium-light-skin-tone} {} {Man~fairy:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-fairy-medium-skin-tone} {} {Man~fairy:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-fairy-medium-dark-skin-tone} {} {Man~fairy:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-fairy-dark-skin-tone} {} {Man~fairy:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^200d^^^^2640^^^^fe0f} {woman-fairy} {fairy-woman} {Woman~fairy} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-fairy-light-skin-tone} {} {Woman~fairy:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-fairy-medium-light-skin-tone} {} {Woman~fairy:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-fairy-medium-skin-tone} {} {Woman~fairy:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-fairy-medium-dark-skin-tone} {} {Woman~fairy:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9da^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-fairy-dark-skin-tone} {} {Woman~fairy:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db} {vampire} {} {Vampire} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fb} {vampire-light-skin-tone} {} {Vampire:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fc} {vampire-medium-light-skin-tone} {} {Vampire:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fd} {vampire-medium-skin-tone} {} {Vampire:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fe} {vampire-medium-dark-skin-tone} {} {Vampire:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3ff} {vampire-dark-skin-tone} {} {Vampire:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^200d^^^^2642^^^^fe0f} {man-vampire} {vampire-man} {Man~vampire} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-vampire-light-skin-tone} {} {Man~vampire:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-vampire-medium-light-skin-tone} {} {Man~vampire:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-vampire-medium-skin-tone} {} {Man~vampire:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-vampire-medium-dark-skin-tone} {} {Man~vampire:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-vampire-dark-skin-tone} {} {Man~vampire:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^200d^^^^2640^^^^fe0f} {woman-vampire} {vampire-woman} {Woman~vampire} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-vampire-light-skin-tone} {} {Woman~vampire:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-vampire-medium-light-skin-tone} {} {Woman~vampire:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-vampire-medium-skin-tone} {} {Woman~vampire:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-vampire-medium-dark-skin-tone} {} {Woman~vampire:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9db^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-vampire-dark-skin-tone} {} {Woman~vampire:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc} {merperson} {} {Merperson} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fb} {merperson-light-skin-tone} {} {Merperson:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fc} {merperson-medium-light-skin-tone} {} {Merperson:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fd} {merperson-medium-skin-tone} {} {Merperson:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fe} {merperson-medium-dark-skin-tone} {} {Merperson:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3ff} {merperson-dark-skin-tone} {} {Merperson:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^200d^^^^2642^^^^fe0f} {merman} {} {Merman} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {merman-light-skin-tone} {} {Merman:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {merman-medium-light-skin-tone} {} {Merman:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {merman-medium-skin-tone} {} {Merman:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {merman-medium-dark-skin-tone} {} {Merman:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {merman-dark-skin-tone} {} {Merman:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^200d^^^^2640^^^^fe0f} {mermaid} {} {Mermaid} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {mermaid-light-skin-tone} {} {Mermaid:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {mermaid-medium-light-skin-tone} {} {Mermaid:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {mermaid-medium-skin-tone} {} {Mermaid:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {mermaid-medium-dark-skin-tone} {} {Mermaid:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dc^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {mermaid-dark-skin-tone} {} {Mermaid:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd} {elf} {} {Elf} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fb} {elf-light-skin-tone} {} {Elf:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fc} {elf-medium-light-skin-tone} {} {Elf:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fd} {elf-medium-skin-tone} {} {Elf:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fe} {elf-medium-dark-skin-tone} {} {Elf:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3ff} {elf-dark-skin-tone} {} {Elf:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^200d^^^^2642^^^^fe0f} {man-elf} {elf-man} {Man~elf} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-elf-light-skin-tone} {} {Man~elf:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-elf-medium-light-skin-tone} {} {Man~elf:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-elf-medium-skin-tone} {} {Man~elf:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-elf-medium-dark-skin-tone} {} {Man~elf:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-elf-dark-skin-tone} {} {Man~elf:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^200d^^^^2640^^^^fe0f} {woman-elf} {elf-woman} {Woman~elf} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-elf-light-skin-tone} {} {Woman~elf:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-elf-medium-light-skin-tone} {} {Woman~elf:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-elf-medium-skin-tone} {} {Woman~elf:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-elf-medium-dark-skin-tone} {} {Woman~elf:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9dd^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-elf-dark-skin-tone} {} {Woman~elf:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9de} {genie} {} {Genie} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9de^^^^200d^^^^2642^^^^fe0f} {man-genie} {genie-man} {Man~genie} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9de^^^^200d^^^^2640^^^^fe0f} {woman-genie} {genie-woman} {Woman~genie} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9df} {zombie} {} {Zombie} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9df^^^^200d^^^^2642^^^^fe0f} {man-zombie} {zombie-man} {Man~zombie} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9df^^^^200d^^^^2640^^^^fe0f} {woman-zombie} {zombie-woman} {Woman~zombie} {5.0}
\__emoji_subgroup:n {Person-activity}
\__emoji_def:nnnnn {^^^^^^01f486} {person-getting-massage} {massage} {Person~getting~massage} {0.6}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fb} {person-getting-massage-light-skin-tone} {} {Person~getting~massage:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fc} {person-getting-massage-medium-light-skin-tone} {} {Person~getting~massage:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fd} {person-getting-massage-medium-skin-tone} {} {Person~getting~massage:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fe} {person-getting-massage-medium-dark-skin-tone} {} {Person~getting~massage:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3ff} {person-getting-massage-dark-skin-tone} {} {Person~getting~massage:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^200d^^^^2642^^^^fe0f} {man-getting-massage} {massage-man} {Man~getting~massage} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-getting-massage-light-skin-tone} {} {Man~getting~massage:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-getting-massage-medium-light-skin-tone} {} {Man~getting~massage:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-getting-massage-medium-skin-tone} {} {Man~getting~massage:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-getting-massage-medium-dark-skin-tone} {} {Man~getting~massage:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-getting-massage-dark-skin-tone} {} {Man~getting~massage:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^200d^^^^2640^^^^fe0f} {woman-getting-massage} {massage-woman} {Woman~getting~massage} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-getting-massage-light-skin-tone} {} {Woman~getting~massage:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-getting-massage-medium-light-skin-tone} {} {Woman~getting~massage:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-getting-massage-medium-skin-tone} {} {Woman~getting~massage:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-getting-massage-medium-dark-skin-tone} {} {Woman~getting~massage:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f486^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-getting-massage-dark-skin-tone} {} {Woman~getting~massage:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487} {person-getting-haircut} {haircut} {Person~getting~haircut} {0.6}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fb} {person-getting-haircut-light-skin-tone} {} {Person~getting~haircut:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fc} {person-getting-haircut-medium-light-skin-tone} {} {Person~getting~haircut:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fd} {person-getting-haircut-medium-skin-tone} {} {Person~getting~haircut:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fe} {person-getting-haircut-medium-dark-skin-tone} {} {Person~getting~haircut:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3ff} {person-getting-haircut-dark-skin-tone} {} {Person~getting~haircut:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^200d^^^^2642^^^^fe0f} {man-getting-haircut} {haircut-man} {Man~getting~haircut} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-getting-haircut-light-skin-tone} {} {Man~getting~haircut:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-getting-haircut-medium-light-skin-tone} {} {Man~getting~haircut:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-getting-haircut-medium-skin-tone} {} {Man~getting~haircut:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-getting-haircut-medium-dark-skin-tone} {} {Man~getting~haircut:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-getting-haircut-dark-skin-tone} {} {Man~getting~haircut:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^200d^^^^2640^^^^fe0f} {woman-getting-haircut} {haircut-woman} {Woman~getting~haircut} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-getting-haircut-light-skin-tone} {} {Woman~getting~haircut:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-getting-haircut-medium-light-skin-tone} {} {Woman~getting~haircut:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-getting-haircut-medium-skin-tone} {} {Woman~getting~haircut:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-getting-haircut-medium-dark-skin-tone} {} {Woman~getting~haircut:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f487^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-getting-haircut-dark-skin-tone} {} {Woman~getting~haircut:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6} {person-walking} {walking} {Person~walking} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fb} {person-walking-light-skin-tone} {} {Person~walking:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fc} {person-walking-medium-light-skin-tone} {} {Person~walking:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fd} {person-walking-medium-skin-tone} {} {Person~walking:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fe} {person-walking-medium-dark-skin-tone} {} {Person~walking:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3ff} {person-walking-dark-skin-tone} {} {Person~walking:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^200d^^^^2642^^^^fe0f} {man-walking} {walking-man} {Man~walking} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-walking-light-skin-tone} {} {Man~walking:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-walking-medium-light-skin-tone} {} {Man~walking:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-walking-medium-skin-tone} {} {Man~walking:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-walking-medium-dark-skin-tone} {} {Man~walking:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-walking-dark-skin-tone} {} {Man~walking:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^200d^^^^2640^^^^fe0f} {woman-walking} {walking-woman} {Woman~walking} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-walking-light-skin-tone} {} {Woman~walking:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-walking-medium-light-skin-tone} {} {Woman~walking:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-walking-medium-skin-tone} {} {Woman~walking:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-walking-medium-dark-skin-tone} {} {Woman~walking:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b6^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-walking-dark-skin-tone} {} {Woman~walking:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9cd} {person-standing} {standing-person} {Person~standing} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fb} {person-standing-light-skin-tone} {} {Person~standing:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fc} {person-standing-medium-light-skin-tone} {} {Person~standing:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fd} {person-standing-medium-skin-tone} {} {Person~standing:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fe} {person-standing-medium-dark-skin-tone} {} {Person~standing:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3ff} {person-standing-dark-skin-tone} {} {Person~standing:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^200d^^^^2642^^^^fe0f} {man-standing} {standing-man} {Man~standing} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-standing-light-skin-tone} {} {Man~standing:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-standing-medium-light-skin-tone} {} {Man~standing:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-standing-medium-skin-tone} {} {Man~standing:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-standing-medium-dark-skin-tone} {} {Man~standing:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-standing-dark-skin-tone} {} {Man~standing:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^200d^^^^2640^^^^fe0f} {woman-standing} {standing-woman} {Woman~standing} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-standing-light-skin-tone} {} {Woman~standing:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-standing-medium-light-skin-tone} {} {Woman~standing:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-standing-medium-skin-tone} {} {Woman~standing:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-standing-medium-dark-skin-tone} {} {Woman~standing:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9cd^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-standing-dark-skin-tone} {} {Woman~standing:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce} {person-kneeling} {kneeling-person} {Person~kneeling} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fb} {person-kneeling-light-skin-tone} {} {Person~kneeling:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fc} {person-kneeling-medium-light-skin-tone} {} {Person~kneeling:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fd} {person-kneeling-medium-skin-tone} {} {Person~kneeling:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fe} {person-kneeling-medium-dark-skin-tone} {} {Person~kneeling:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3ff} {person-kneeling-dark-skin-tone} {} {Person~kneeling:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^200d^^^^2642^^^^fe0f} {man-kneeling} {kneeling-man} {Man~kneeling} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-kneeling-light-skin-tone} {} {Man~kneeling:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-kneeling-medium-light-skin-tone} {} {Man~kneeling:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-kneeling-medium-skin-tone} {} {Man~kneeling:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-kneeling-medium-dark-skin-tone} {} {Man~kneeling:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-kneeling-dark-skin-tone} {} {Man~kneeling:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^200d^^^^2640^^^^fe0f} {woman-kneeling} {kneeling-woman} {Woman~kneeling} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-kneeling-light-skin-tone} {} {Woman~kneeling:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-kneeling-medium-light-skin-tone} {} {Woman~kneeling:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-kneeling-medium-skin-tone} {} {Woman~kneeling:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-kneeling-medium-dark-skin-tone} {} {Woman~kneeling:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ce^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-kneeling-dark-skin-tone} {} {Woman~kneeling:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f9af} {person-with-white-cane} {person-with-probing-cane} {Person~with~white~cane} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f9af} {person-with-white-cane-light-skin-tone} {} {Person~with~white~cane:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f9af} {person-with-white-cane-medium-light-skin-tone} {} {Person~with~white~cane:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f9af} {person-with-white-cane-medium-skin-tone} {} {Person~with~white~cane:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f9af} {person-with-white-cane-medium-dark-skin-tone} {} {Person~with~white~cane:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f9af} {person-with-white-cane-dark-skin-tone} {} {Person~with~white~cane:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f9af} {man-with-white-cane} {man-with-probing-cane} {Man~with~white~cane} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f9af} {man-with-white-cane-light-skin-tone} {} {Man~with~white~cane:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f9af} {man-with-white-cane-medium-light-skin-tone} {} {Man~with~white~cane:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f9af} {man-with-white-cane-medium-skin-tone} {} {Man~with~white~cane:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f9af} {man-with-white-cane-medium-dark-skin-tone} {} {Man~with~white~cane:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f9af} {man-with-white-cane-dark-skin-tone} {} {Man~with~white~cane:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f9af} {woman-with-white-cane} {woman-with-probing-cane} {Woman~with~white~cane} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f9af} {woman-with-white-cane-light-skin-tone} {} {Woman~with~white~cane:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f9af} {woman-with-white-cane-medium-light-skin-tone} {} {Woman~with~white~cane:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f9af} {woman-with-white-cane-medium-skin-tone} {} {Woman~with~white~cane:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f9af} {woman-with-white-cane-medium-dark-skin-tone} {} {Woman~with~white~cane:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f9af} {woman-with-white-cane-dark-skin-tone} {} {Woman~with~white~cane:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f9bc} {person-in-motorized-wheelchair} {} {Person~in~motorized~wheelchair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f9bc} {person-in-motorized-wheelchair-light-skin-tone} {} {Person~in~motorized~wheelchair:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f9bc} {person-in-motorized-wheelchair-medium-light-skin-tone} {} {Person~in~motorized~wheelchair:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f9bc} {person-in-motorized-wheelchair-medium-skin-tone} {} {Person~in~motorized~wheelchair:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f9bc} {person-in-motorized-wheelchair-medium-dark-skin-tone} {} {Person~in~motorized~wheelchair:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f9bc} {person-in-motorized-wheelchair-dark-skin-tone} {} {Person~in~motorized~wheelchair:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f9bc} {man-in-motorized-wheelchair} {} {Man~in~motorized~wheelchair} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f9bc} {man-in-motorized-wheelchair-light-skin-tone} {} {Man~in~motorized~wheelchair:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f9bc} {man-in-motorized-wheelchair-medium-light-skin-tone} {} {Man~in~motorized~wheelchair:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f9bc} {man-in-motorized-wheelchair-medium-skin-tone} {} {Man~in~motorized~wheelchair:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f9bc} {man-in-motorized-wheelchair-medium-dark-skin-tone} {} {Man~in~motorized~wheelchair:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f9bc} {man-in-motorized-wheelchair-dark-skin-tone} {} {Man~in~motorized~wheelchair:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f9bc} {woman-in-motorized-wheelchair} {} {Woman~in~motorized~wheelchair} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f9bc} {woman-in-motorized-wheelchair-light-skin-tone} {} {Woman~in~motorized~wheelchair:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f9bc} {woman-in-motorized-wheelchair-medium-light-skin-tone} {} {Woman~in~motorized~wheelchair:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f9bc} {woman-in-motorized-wheelchair-medium-skin-tone} {} {Woman~in~motorized~wheelchair:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f9bc} {woman-in-motorized-wheelchair-medium-dark-skin-tone} {} {Woman~in~motorized~wheelchair:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f9bc} {woman-in-motorized-wheelchair-dark-skin-tone} {} {Woman~in~motorized~wheelchair:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f9bd} {person-in-manual-wheelchair} {} {Person~in~manual~wheelchair} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f9bd} {person-in-manual-wheelchair-light-skin-tone} {} {Person~in~manual~wheelchair:~light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f9bd} {person-in-manual-wheelchair-medium-light-skin-tone} {} {Person~in~manual~wheelchair:~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f9bd} {person-in-manual-wheelchair-medium-skin-tone} {} {Person~in~manual~wheelchair:~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f9bd} {person-in-manual-wheelchair-medium-dark-skin-tone} {} {Person~in~manual~wheelchair:~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f9bd} {person-in-manual-wheelchair-dark-skin-tone} {} {Person~in~manual~wheelchair:~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f9bd} {man-in-manual-wheelchair} {} {Man~in~manual~wheelchair} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f9bd} {man-in-manual-wheelchair-light-skin-tone} {} {Man~in~manual~wheelchair:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f9bd} {man-in-manual-wheelchair-medium-light-skin-tone} {} {Man~in~manual~wheelchair:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f9bd} {man-in-manual-wheelchair-medium-skin-tone} {} {Man~in~manual~wheelchair:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f9bd} {man-in-manual-wheelchair-medium-dark-skin-tone} {} {Man~in~manual~wheelchair:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f9bd} {man-in-manual-wheelchair-dark-skin-tone} {} {Man~in~manual~wheelchair:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f9bd} {woman-in-manual-wheelchair} {} {Woman~in~manual~wheelchair} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f9bd} {woman-in-manual-wheelchair-light-skin-tone} {} {Woman~in~manual~wheelchair:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f9bd} {woman-in-manual-wheelchair-medium-light-skin-tone} {} {Woman~in~manual~wheelchair:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f9bd} {woman-in-manual-wheelchair-medium-skin-tone} {} {Woman~in~manual~wheelchair:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f9bd} {woman-in-manual-wheelchair-medium-dark-skin-tone} {} {Woman~in~manual~wheelchair:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f9bd} {woman-in-manual-wheelchair-dark-skin-tone} {} {Woman~in~manual~wheelchair:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f3c3} {person-running} {runner, running} {Person~running} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fb} {person-running-light-skin-tone} {} {Person~running:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fc} {person-running-medium-light-skin-tone} {} {Person~running:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fd} {person-running-medium-skin-tone} {} {Person~running:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fe} {person-running-medium-dark-skin-tone} {} {Person~running:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3ff} {person-running-dark-skin-tone} {} {Person~running:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^200d^^^^2642^^^^fe0f} {man-running} {running-man} {Man~running} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-running-light-skin-tone} {} {Man~running:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-running-medium-light-skin-tone} {} {Man~running:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-running-medium-skin-tone} {} {Man~running:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-running-medium-dark-skin-tone} {} {Man~running:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-running-dark-skin-tone} {} {Man~running:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^200d^^^^2640^^^^fe0f} {woman-running} {running-woman} {Woman~running} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-running-light-skin-tone} {} {Woman~running:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-running-medium-light-skin-tone} {} {Woman~running:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-running-medium-skin-tone} {} {Woman~running:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-running-medium-dark-skin-tone} {} {Woman~running:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c3^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-running-dark-skin-tone} {} {Woman~running:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f483} {woman-dancing} {dancer} {Woman~dancing} {0.6}
\__emoji_def:nnnnn {^^^^^^01f483^^^^^^01f3fb} {woman-dancing-light-skin-tone} {} {Woman~dancing:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f483^^^^^^01f3fc} {woman-dancing-medium-light-skin-tone} {} {Woman~dancing:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f483^^^^^^01f3fd} {woman-dancing-medium-skin-tone} {} {Woman~dancing:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f483^^^^^^01f3fe} {woman-dancing-medium-dark-skin-tone} {} {Woman~dancing:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f483^^^^^^01f3ff} {woman-dancing-dark-skin-tone} {} {Woman~dancing:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f57a} {man-dancing} {} {Man~dancing} {3.0}
\__emoji_def:nnnnn {^^^^^^01f57a^^^^^^01f3fb} {man-dancing-light-skin-tone} {} {Man~dancing:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f57a^^^^^^01f3fc} {man-dancing-medium-light-skin-tone} {} {Man~dancing:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f57a^^^^^^01f3fd} {man-dancing-medium-skin-tone} {} {Man~dancing:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f57a^^^^^^01f3fe} {man-dancing-medium-dark-skin-tone} {} {Man~dancing:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f57a^^^^^^01f3ff} {man-dancing-dark-skin-tone} {} {Man~dancing:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f574^^^^fe0f} {person-in-suit-levitating} {business-suit-levitating} {Person~in~suit~levitating} {0.7}
\__emoji_def:nnnnn {^^^^^^01f574^^^^^^01f3fb} {person-in-suit-levitating-light-skin-tone} {} {Person~in~suit~levitating:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f574^^^^^^01f3fc} {person-in-suit-levitating-medium-light-skin-tone} {} {Person~in~suit~levitating:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f574^^^^^^01f3fd} {person-in-suit-levitating-medium-skin-tone} {} {Person~in~suit~levitating:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f574^^^^^^01f3fe} {person-in-suit-levitating-medium-dark-skin-tone} {} {Person~in~suit~levitating:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f574^^^^^^01f3ff} {person-in-suit-levitating-dark-skin-tone} {} {Person~in~suit~levitating:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46f} {people-with-bunny-ears} {dancers} {People~with~bunny~ears} {0.6}
\__emoji_def:nnnnn {^^^^^^01f46f^^^^200d^^^^2642^^^^fe0f} {men-with-bunny-ears} {dancing-men} {Men~with~bunny~ears} {4.0}
\__emoji_def:nnnnn {^^^^^^01f46f^^^^200d^^^^2640^^^^fe0f} {women-with-bunny-ears} {dancing-women} {Women~with~bunny~ears} {4.0}
\__emoji_def:nnnnn {^^^^^^01f9d6} {person-in-steamy-room} {sauna-person} {Person~in~steamy~room} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fb} {person-in-steamy-room-light-skin-tone} {} {Person~in~steamy~room:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fc} {person-in-steamy-room-medium-light-skin-tone} {} {Person~in~steamy~room:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fd} {person-in-steamy-room-medium-skin-tone} {} {Person~in~steamy~room:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fe} {person-in-steamy-room-medium-dark-skin-tone} {} {Person~in~steamy~room:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3ff} {person-in-steamy-room-dark-skin-tone} {} {Person~in~steamy~room:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^200d^^^^2642^^^^fe0f} {man-in-steamy-room} {sauna-man} {Man~in~steamy~room} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-in-steamy-room-light-skin-tone} {} {Man~in~steamy~room:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-in-steamy-room-medium-light-skin-tone} {} {Man~in~steamy~room:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-in-steamy-room-medium-skin-tone} {} {Man~in~steamy~room:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-in-steamy-room-medium-dark-skin-tone} {} {Man~in~steamy~room:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-in-steamy-room-dark-skin-tone} {} {Man~in~steamy~room:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^200d^^^^2640^^^^fe0f} {woman-in-steamy-room} {sauna-woman} {Woman~in~steamy~room} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-in-steamy-room-light-skin-tone} {} {Woman~in~steamy~room:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-in-steamy-room-medium-light-skin-tone} {} {Woman~in~steamy~room:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-in-steamy-room-medium-skin-tone} {} {Woman~in~steamy~room:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-in-steamy-room-medium-dark-skin-tone} {} {Woman~in~steamy~room:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d6^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-in-steamy-room-dark-skin-tone} {} {Woman~in~steamy~room:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7} {person-climbing} {climbing} {Person~climbing} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fb} {person-climbing-light-skin-tone} {} {Person~climbing:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fc} {person-climbing-medium-light-skin-tone} {} {Person~climbing:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fd} {person-climbing-medium-skin-tone} {} {Person~climbing:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fe} {person-climbing-medium-dark-skin-tone} {} {Person~climbing:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3ff} {person-climbing-dark-skin-tone} {} {Person~climbing:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^200d^^^^2642^^^^fe0f} {man-climbing} {climbing-man} {Man~climbing} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-climbing-light-skin-tone} {} {Man~climbing:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-climbing-medium-light-skin-tone} {} {Man~climbing:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-climbing-medium-skin-tone} {} {Man~climbing:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-climbing-medium-dark-skin-tone} {} {Man~climbing:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-climbing-dark-skin-tone} {} {Man~climbing:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^200d^^^^2640^^^^fe0f} {woman-climbing} {climbing-woman} {Woman~climbing} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-climbing-light-skin-tone} {} {Woman~climbing:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-climbing-medium-light-skin-tone} {} {Woman~climbing:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-climbing-medium-skin-tone} {} {Woman~climbing:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-climbing-medium-dark-skin-tone} {} {Woman~climbing:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d7^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-climbing-dark-skin-tone} {} {Woman~climbing:~dark~skin~tone} {5.0}
\__emoji_subgroup:n {Person-sport}
\__emoji_def:nnnnn {^^^^^^01f93a} {person-fencing} {} {Person~fencing} {3.0}
\__emoji_def:nnnnn {^^^^^^01f3c7} {horse-racing} {} {Horse~racing} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c7^^^^^^01f3fb} {horse-racing-light-skin-tone} {} {Horse~racing:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c7^^^^^^01f3fc} {horse-racing-medium-light-skin-tone} {} {Horse~racing:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c7^^^^^^01f3fd} {horse-racing-medium-skin-tone} {} {Horse~racing:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c7^^^^^^01f3fe} {horse-racing-medium-dark-skin-tone} {} {Horse~racing:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c7^^^^^^01f3ff} {horse-racing-dark-skin-tone} {} {Horse~racing:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^26f7^^^^fe0f} {skier} {} {Skier} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3c2} {snowboarder} {} {Snowboarder} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3c2^^^^^^01f3fb} {snowboarder-light-skin-tone} {} {Snowboarder:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c2^^^^^^01f3fc} {snowboarder-medium-light-skin-tone} {} {Snowboarder:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c2^^^^^^01f3fd} {snowboarder-medium-skin-tone} {} {Snowboarder:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c2^^^^^^01f3fe} {snowboarder-medium-dark-skin-tone} {} {Snowboarder:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c2^^^^^^01f3ff} {snowboarder-dark-skin-tone} {} {Snowboarder:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^fe0f} {person-golfing} {golfing} {Person~golfing} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fb} {person-golfing-light-skin-tone} {} {Person~golfing:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fc} {person-golfing-medium-light-skin-tone} {} {Person~golfing:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fd} {person-golfing-medium-skin-tone} {} {Person~golfing:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fe} {person-golfing-medium-dark-skin-tone} {} {Person~golfing:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3ff} {person-golfing-dark-skin-tone} {} {Person~golfing:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^fe0f^^^^200d^^^^2642^^^^fe0f} {man-golfing} {golfing-man} {Man~golfing} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-golfing-light-skin-tone} {} {Man~golfing:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-golfing-medium-light-skin-tone} {} {Man~golfing:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-golfing-medium-skin-tone} {} {Man~golfing:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-golfing-medium-dark-skin-tone} {} {Man~golfing:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-golfing-dark-skin-tone} {} {Man~golfing:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^fe0f^^^^200d^^^^2640^^^^fe0f} {woman-golfing} {golfing-woman} {Woman~golfing} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-golfing-light-skin-tone} {} {Woman~golfing:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-golfing-medium-light-skin-tone} {} {Woman~golfing:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-golfing-medium-skin-tone} {} {Woman~golfing:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-golfing-medium-dark-skin-tone} {} {Woman~golfing:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cc^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-golfing-dark-skin-tone} {} {Woman~golfing:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4} {person-surfing} {surfer} {Person~surfing} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fb} {person-surfing-light-skin-tone} {} {Person~surfing:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fc} {person-surfing-medium-light-skin-tone} {} {Person~surfing:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fd} {person-surfing-medium-skin-tone} {} {Person~surfing:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fe} {person-surfing-medium-dark-skin-tone} {} {Person~surfing:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3ff} {person-surfing-dark-skin-tone} {} {Person~surfing:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^200d^^^^2642^^^^fe0f} {man-surfing} {surfing-man} {Man~surfing} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-surfing-light-skin-tone} {} {Man~surfing:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-surfing-medium-light-skin-tone} {} {Man~surfing:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-surfing-medium-skin-tone} {} {Man~surfing:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-surfing-medium-dark-skin-tone} {} {Man~surfing:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-surfing-dark-skin-tone} {} {Man~surfing:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^200d^^^^2640^^^^fe0f} {woman-surfing} {surfing-woman} {Woman~surfing} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-surfing-light-skin-tone} {} {Woman~surfing:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-surfing-medium-light-skin-tone} {} {Woman~surfing:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-surfing-medium-skin-tone} {} {Woman~surfing:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-surfing-medium-dark-skin-tone} {} {Woman~surfing:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3c4^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-surfing-dark-skin-tone} {} {Woman~surfing:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3} {person-rowing-boat} {rowboat} {Person~rowing~boat} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fb} {person-rowing-boat-light-skin-tone} {} {Person~rowing~boat:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fc} {person-rowing-boat-medium-light-skin-tone} {} {Person~rowing~boat:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fd} {person-rowing-boat-medium-skin-tone} {} {Person~rowing~boat:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fe} {person-rowing-boat-medium-dark-skin-tone} {} {Person~rowing~boat:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3ff} {person-rowing-boat-dark-skin-tone} {} {Person~rowing~boat:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^200d^^^^2642^^^^fe0f} {man-rowing-boat} {rowing-man} {Man~rowing~boat} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-rowing-boat-light-skin-tone} {} {Man~rowing~boat:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-rowing-boat-medium-light-skin-tone} {} {Man~rowing~boat:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-rowing-boat-medium-skin-tone} {} {Man~rowing~boat:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-rowing-boat-medium-dark-skin-tone} {} {Man~rowing~boat:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-rowing-boat-dark-skin-tone} {} {Man~rowing~boat:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^200d^^^^2640^^^^fe0f} {woman-rowing-boat} {rowing-woman} {Woman~rowing~boat} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-rowing-boat-light-skin-tone} {} {Woman~rowing~boat:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-rowing-boat-medium-light-skin-tone} {} {Woman~rowing~boat:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-rowing-boat-medium-skin-tone} {} {Woman~rowing~boat:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-rowing-boat-medium-dark-skin-tone} {} {Woman~rowing~boat:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6a3^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-rowing-boat-dark-skin-tone} {} {Woman~rowing~boat:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca} {person-swimming} {swimmer} {Person~swimming} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fb} {person-swimming-light-skin-tone} {} {Person~swimming:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fc} {person-swimming-medium-light-skin-tone} {} {Person~swimming:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fd} {person-swimming-medium-skin-tone} {} {Person~swimming:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fe} {person-swimming-medium-dark-skin-tone} {} {Person~swimming:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3ff} {person-swimming-dark-skin-tone} {} {Person~swimming:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^200d^^^^2642^^^^fe0f} {man-swimming} {swimming-man} {Man~swimming} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-swimming-light-skin-tone} {} {Man~swimming:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-swimming-medium-light-skin-tone} {} {Man~swimming:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-swimming-medium-skin-tone} {} {Man~swimming:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-swimming-medium-dark-skin-tone} {} {Man~swimming:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-swimming-dark-skin-tone} {} {Man~swimming:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^200d^^^^2640^^^^fe0f} {woman-swimming} {swimming-woman} {Woman~swimming} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-swimming-light-skin-tone} {} {Woman~swimming:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-swimming-medium-light-skin-tone} {} {Woman~swimming:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-swimming-medium-skin-tone} {} {Woman~swimming:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-swimming-medium-dark-skin-tone} {} {Woman~swimming:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3ca^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-swimming-dark-skin-tone} {} {Woman~swimming:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^fe0f} {person-bouncing-ball} {bouncing-ball-person} {Person~bouncing~ball} {0.7}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fb} {person-bouncing-ball-light-skin-tone} {} {Person~bouncing~ball:~light~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fc} {person-bouncing-ball-medium-light-skin-tone} {} {Person~bouncing~ball:~medium-light~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fd} {person-bouncing-ball-medium-skin-tone} {} {Person~bouncing~ball:~medium~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fe} {person-bouncing-ball-medium-dark-skin-tone} {} {Person~bouncing~ball:~medium-dark~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3ff} {person-bouncing-ball-dark-skin-tone} {} {Person~bouncing~ball:~dark~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^26f9^^^^fe0f^^^^200d^^^^2642^^^^fe0f} {man-bouncing-ball} {basketball-man, bouncing-ball-man} {Man~bouncing~ball} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-bouncing-ball-light-skin-tone} {} {Man~bouncing~ball:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-bouncing-ball-medium-light-skin-tone} {} {Man~bouncing~ball:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-bouncing-ball-medium-skin-tone} {} {Man~bouncing~ball:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-bouncing-ball-medium-dark-skin-tone} {} {Man~bouncing~ball:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-bouncing-ball-dark-skin-tone} {} {Man~bouncing~ball:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^fe0f^^^^200d^^^^2640^^^^fe0f} {woman-bouncing-ball} {basketball-woman, bouncing-ball-woman} {Woman~bouncing~ball} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-bouncing-ball-light-skin-tone} {} {Woman~bouncing~ball:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-bouncing-ball-medium-light-skin-tone} {} {Woman~bouncing~ball:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-bouncing-ball-medium-skin-tone} {} {Woman~bouncing~ball:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-bouncing-ball-medium-dark-skin-tone} {} {Woman~bouncing~ball:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^26f9^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-bouncing-ball-dark-skin-tone} {} {Woman~bouncing~ball:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^fe0f} {person-lifting-weights} {weight-lifting} {Person~lifting~weights} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fb} {person-lifting-weights-light-skin-tone} {} {Person~lifting~weights:~light~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fc} {person-lifting-weights-medium-light-skin-tone} {} {Person~lifting~weights:~medium-light~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fd} {person-lifting-weights-medium-skin-tone} {} {Person~lifting~weights:~medium~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fe} {person-lifting-weights-medium-dark-skin-tone} {} {Person~lifting~weights:~medium-dark~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3ff} {person-lifting-weights-dark-skin-tone} {} {Person~lifting~weights:~dark~skin~tone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^fe0f^^^^200d^^^^2642^^^^fe0f} {man-lifting-weights} {weight-lifting-man} {Man~lifting~weights} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-lifting-weights-light-skin-tone} {} {Man~lifting~weights:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-lifting-weights-medium-light-skin-tone} {} {Man~lifting~weights:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-lifting-weights-medium-skin-tone} {} {Man~lifting~weights:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-lifting-weights-medium-dark-skin-tone} {} {Man~lifting~weights:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-lifting-weights-dark-skin-tone} {} {Man~lifting~weights:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^fe0f^^^^200d^^^^2640^^^^fe0f} {woman-lifting-weights} {weight-lifting-woman} {Woman~lifting~weights} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-lifting-weights-light-skin-tone} {} {Woman~lifting~weights:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-lifting-weights-medium-light-skin-tone} {} {Woman~lifting~weights:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-lifting-weights-medium-skin-tone} {} {Woman~lifting~weights:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-lifting-weights-medium-dark-skin-tone} {} {Woman~lifting~weights:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3cb^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-lifting-weights-dark-skin-tone} {} {Woman~lifting~weights:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4} {person-biking} {bicyclist} {Person~biking} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fb} {person-biking-light-skin-tone} {} {Person~biking:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fc} {person-biking-medium-light-skin-tone} {} {Person~biking:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fd} {person-biking-medium-skin-tone} {} {Person~biking:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fe} {person-biking-medium-dark-skin-tone} {} {Person~biking:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3ff} {person-biking-dark-skin-tone} {} {Person~biking:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^200d^^^^2642^^^^fe0f} {man-biking} {biking-man} {Man~biking} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-biking-light-skin-tone} {} {Man~biking:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-biking-medium-light-skin-tone} {} {Man~biking:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-biking-medium-skin-tone} {} {Man~biking:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-biking-medium-dark-skin-tone} {} {Man~biking:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-biking-dark-skin-tone} {} {Man~biking:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^200d^^^^2640^^^^fe0f} {woman-biking} {biking-woman} {Woman~biking} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-biking-light-skin-tone} {} {Woman~biking:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-biking-medium-light-skin-tone} {} {Woman~biking:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-biking-medium-skin-tone} {} {Woman~biking:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-biking-medium-dark-skin-tone} {} {Woman~biking:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b4^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-biking-dark-skin-tone} {} {Woman~biking:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5} {person-mountain-biking} {mountain-bicyclist} {Person~mountain~biking} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fb} {person-mountain-biking-light-skin-tone} {} {Person~mountain~biking:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fc} {person-mountain-biking-medium-light-skin-tone} {} {Person~mountain~biking:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fd} {person-mountain-biking-medium-skin-tone} {} {Person~mountain~biking:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fe} {person-mountain-biking-medium-dark-skin-tone} {} {Person~mountain~biking:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3ff} {person-mountain-biking-dark-skin-tone} {} {Person~mountain~biking:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^200d^^^^2642^^^^fe0f} {man-mountain-biking} {mountain-biking-man} {Man~mountain~biking} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-mountain-biking-light-skin-tone} {} {Man~mountain~biking:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-mountain-biking-medium-light-skin-tone} {} {Man~mountain~biking:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-mountain-biking-medium-skin-tone} {} {Man~mountain~biking:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-mountain-biking-medium-dark-skin-tone} {} {Man~mountain~biking:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-mountain-biking-dark-skin-tone} {} {Man~mountain~biking:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^200d^^^^2640^^^^fe0f} {woman-mountain-biking} {mountain-biking-woman} {Woman~mountain~biking} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-mountain-biking-light-skin-tone} {} {Woman~mountain~biking:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-mountain-biking-medium-light-skin-tone} {} {Woman~mountain~biking:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-mountain-biking-medium-skin-tone} {} {Woman~mountain~biking:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-mountain-biking-medium-dark-skin-tone} {} {Woman~mountain~biking:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6b5^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-mountain-biking-dark-skin-tone} {} {Woman~mountain~biking:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938} {person-cartwheeling} {cartwheeling} {Person~cartwheeling} {3.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fb} {person-cartwheeling-light-skin-tone} {} {Person~cartwheeling:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fc} {person-cartwheeling-medium-light-skin-tone} {} {Person~cartwheeling:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fd} {person-cartwheeling-medium-skin-tone} {} {Person~cartwheeling:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fe} {person-cartwheeling-medium-dark-skin-tone} {} {Person~cartwheeling:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3ff} {person-cartwheeling-dark-skin-tone} {} {Person~cartwheeling:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^200d^^^^2642^^^^fe0f} {man-cartwheeling} {} {Man~cartwheeling} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-cartwheeling-light-skin-tone} {} {Man~cartwheeling:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-cartwheeling-medium-light-skin-tone} {} {Man~cartwheeling:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-cartwheeling-medium-skin-tone} {} {Man~cartwheeling:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-cartwheeling-medium-dark-skin-tone} {} {Man~cartwheeling:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-cartwheeling-dark-skin-tone} {} {Man~cartwheeling:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^200d^^^^2640^^^^fe0f} {woman-cartwheeling} {} {Woman~cartwheeling} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-cartwheeling-light-skin-tone} {} {Woman~cartwheeling:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-cartwheeling-medium-light-skin-tone} {} {Woman~cartwheeling:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-cartwheeling-medium-skin-tone} {} {Woman~cartwheeling:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-cartwheeling-medium-dark-skin-tone} {} {Woman~cartwheeling:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f938^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-cartwheeling-dark-skin-tone} {} {Woman~cartwheeling:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93c} {people-wrestling} {wrestling} {People~wrestling} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93c^^^^200d^^^^2642^^^^fe0f} {men-wrestling} {} {Men~wrestling} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93c^^^^200d^^^^2640^^^^fe0f} {women-wrestling} {} {Women~wrestling} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d} {person-playing-water-polo} {water-polo} {Person~playing~water~polo} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fb} {person-playing-water-polo-light-skin-tone} {} {Person~playing~water~polo:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fc} {person-playing-water-polo-medium-light-skin-tone} {} {Person~playing~water~polo:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fd} {person-playing-water-polo-medium-skin-tone} {} {Person~playing~water~polo:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fe} {person-playing-water-polo-medium-dark-skin-tone} {} {Person~playing~water~polo:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3ff} {person-playing-water-polo-dark-skin-tone} {} {Person~playing~water~polo:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^200d^^^^2642^^^^fe0f} {man-playing-water-polo} {} {Man~playing~water~polo} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-playing-water-polo-light-skin-tone} {} {Man~playing~water~polo:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-playing-water-polo-medium-light-skin-tone} {} {Man~playing~water~polo:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-playing-water-polo-medium-skin-tone} {} {Man~playing~water~polo:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-playing-water-polo-medium-dark-skin-tone} {} {Man~playing~water~polo:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-playing-water-polo-dark-skin-tone} {} {Man~playing~water~polo:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^200d^^^^2640^^^^fe0f} {woman-playing-water-polo} {} {Woman~playing~water~polo} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-playing-water-polo-light-skin-tone} {} {Woman~playing~water~polo:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-playing-water-polo-medium-light-skin-tone} {} {Woman~playing~water~polo:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-playing-water-polo-medium-skin-tone} {} {Woman~playing~water~polo:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-playing-water-polo-medium-dark-skin-tone} {} {Woman~playing~water~polo:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93d^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-playing-water-polo-dark-skin-tone} {} {Woman~playing~water~polo:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e} {person-playing-handball} {handball-person} {Person~playing~handball} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fb} {person-playing-handball-light-skin-tone} {} {Person~playing~handball:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fc} {person-playing-handball-medium-light-skin-tone} {} {Person~playing~handball:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fd} {person-playing-handball-medium-skin-tone} {} {Person~playing~handball:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fe} {person-playing-handball-medium-dark-skin-tone} {} {Person~playing~handball:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3ff} {person-playing-handball-dark-skin-tone} {} {Person~playing~handball:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^200d^^^^2642^^^^fe0f} {man-playing-handball} {} {Man~playing~handball} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-playing-handball-light-skin-tone} {} {Man~playing~handball:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-playing-handball-medium-light-skin-tone} {} {Man~playing~handball:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-playing-handball-medium-skin-tone} {} {Man~playing~handball:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-playing-handball-medium-dark-skin-tone} {} {Man~playing~handball:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-playing-handball-dark-skin-tone} {} {Man~playing~handball:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^200d^^^^2640^^^^fe0f} {woman-playing-handball} {} {Woman~playing~handball} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-playing-handball-light-skin-tone} {} {Woman~playing~handball:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-playing-handball-medium-light-skin-tone} {} {Woman~playing~handball:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-playing-handball-medium-skin-tone} {} {Woman~playing~handball:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-playing-handball-medium-dark-skin-tone} {} {Woman~playing~handball:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f93e^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-playing-handball-dark-skin-tone} {} {Woman~playing~handball:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939} {person-juggling} {juggling-person} {Person~juggling} {3.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fb} {person-juggling-light-skin-tone} {} {Person~juggling:~light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fc} {person-juggling-medium-light-skin-tone} {} {Person~juggling:~medium-light~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fd} {person-juggling-medium-skin-tone} {} {Person~juggling:~medium~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fe} {person-juggling-medium-dark-skin-tone} {} {Person~juggling:~medium-dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3ff} {person-juggling-dark-skin-tone} {} {Person~juggling:~dark~skin~tone} {3.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^200d^^^^2642^^^^fe0f} {man-juggling} {} {Man~juggling} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-juggling-light-skin-tone} {} {Man~juggling:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-juggling-medium-light-skin-tone} {} {Man~juggling:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-juggling-medium-skin-tone} {} {Man~juggling:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-juggling-medium-dark-skin-tone} {} {Man~juggling:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-juggling-dark-skin-tone} {} {Man~juggling:~dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^200d^^^^2640^^^^fe0f} {woman-juggling} {} {Woman~juggling} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-juggling-light-skin-tone} {} {Woman~juggling:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-juggling-medium-light-skin-tone} {} {Woman~juggling:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-juggling-medium-skin-tone} {} {Woman~juggling:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-juggling-medium-dark-skin-tone} {} {Woman~juggling:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f939^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-juggling-dark-skin-tone} {} {Woman~juggling:~dark~skin~tone} {4.0}
\__emoji_subgroup:n {Person-resting}
\__emoji_def:nnnnn {^^^^^^01f9d8} {person-in-lotus-position} {lotus-position} {Person~in~lotus~position} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fb} {person-in-lotus-position-light-skin-tone} {} {Person~in~lotus~position:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fc} {person-in-lotus-position-medium-light-skin-tone} {} {Person~in~lotus~position:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fd} {person-in-lotus-position-medium-skin-tone} {} {Person~in~lotus~position:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fe} {person-in-lotus-position-medium-dark-skin-tone} {} {Person~in~lotus~position:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3ff} {person-in-lotus-position-dark-skin-tone} {} {Person~in~lotus~position:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^200d^^^^2642^^^^fe0f} {man-in-lotus-position} {lotus-position-man} {Man~in~lotus~position} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fb^^^^200d^^^^2642^^^^fe0f} {man-in-lotus-position-light-skin-tone} {} {Man~in~lotus~position:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fc^^^^200d^^^^2642^^^^fe0f} {man-in-lotus-position-medium-light-skin-tone} {} {Man~in~lotus~position:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fd^^^^200d^^^^2642^^^^fe0f} {man-in-lotus-position-medium-skin-tone} {} {Man~in~lotus~position:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fe^^^^200d^^^^2642^^^^fe0f} {man-in-lotus-position-medium-dark-skin-tone} {} {Man~in~lotus~position:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3ff^^^^200d^^^^2642^^^^fe0f} {man-in-lotus-position-dark-skin-tone} {} {Man~in~lotus~position:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^200d^^^^2640^^^^fe0f} {woman-in-lotus-position} {lotus-position-woman} {Woman~in~lotus~position} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fb^^^^200d^^^^2640^^^^fe0f} {woman-in-lotus-position-light-skin-tone} {} {Woman~in~lotus~position:~light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fc^^^^200d^^^^2640^^^^fe0f} {woman-in-lotus-position-medium-light-skin-tone} {} {Woman~in~lotus~position:~medium-light~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fd^^^^200d^^^^2640^^^^fe0f} {woman-in-lotus-position-medium-skin-tone} {} {Woman~in~lotus~position:~medium~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3fe^^^^200d^^^^2640^^^^fe0f} {woman-in-lotus-position-medium-dark-skin-tone} {} {Woman~in~lotus~position:~medium-dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9d8^^^^^^01f3ff^^^^200d^^^^2640^^^^fe0f} {woman-in-lotus-position-dark-skin-tone} {} {Woman~in~lotus~position:~dark~skin~tone} {5.0}
\__emoji_def:nnnnn {^^^^^^01f6c0} {person-taking-bath} {bath} {Person~taking~bath} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6c0^^^^^^01f3fb} {person-taking-bath-light-skin-tone} {} {Person~taking~bath:~light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c0^^^^^^01f3fc} {person-taking-bath-medium-light-skin-tone} {} {Person~taking~bath:~medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c0^^^^^^01f3fd} {person-taking-bath-medium-skin-tone} {} {Person~taking~bath:~medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c0^^^^^^01f3fe} {person-taking-bath-medium-dark-skin-tone} {} {Person~taking~bath:~medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c0^^^^^^01f3ff} {person-taking-bath-dark-skin-tone} {} {Person~taking~bath:~dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6cc} {person-in-bed} {sleeping-bed} {Person~in~bed} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6cc^^^^^^01f3fb} {person-in-bed-light-skin-tone} {} {Person~in~bed:~light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6cc^^^^^^01f3fc} {person-in-bed-medium-light-skin-tone} {} {Person~in~bed:~medium-light~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6cc^^^^^^01f3fd} {person-in-bed-medium-skin-tone} {} {Person~in~bed:~medium~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6cc^^^^^^01f3fe} {person-in-bed-medium-dark-skin-tone} {} {Person~in~bed:~medium-dark~skin~tone} {4.0}
\__emoji_def:nnnnn {^^^^^^01f6cc^^^^^^01f3ff} {person-in-bed-dark-skin-tone} {} {Person~in~bed:~dark~skin~tone} {4.0}
\__emoji_subgroup:n {Family}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1} {people-holding-hands} {} {People~holding~hands} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {people-holding-hands-light-skin-tone} {} {People~holding~hands:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {people-holding-hands-light-skin-tone-medium-light-skin-tone} {} {People~holding~hands:~light~skin~tone,~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {people-holding-hands-light-skin-tone-medium-skin-tone} {} {People~holding~hands:~light~skin~tone,~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {people-holding-hands-light-skin-tone-medium-dark-skin-tone} {} {People~holding~hands:~light~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {people-holding-hands-light-skin-tone-dark-skin-tone} {} {People~holding~hands:~light~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {people-holding-hands-medium-light-skin-tone-light-skin-tone} {} {People~holding~hands:~medium-light~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {people-holding-hands-medium-light-skin-tone} {} {People~holding~hands:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {people-holding-hands-medium-light-skin-tone-medium-skin-tone} {} {People~holding~hands:~medium-light~skin~tone,~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {people-holding-hands-medium-light-skin-tone-medium-dark-skin-tone} {} {People~holding~hands:~medium-light~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {people-holding-hands-medium-light-skin-tone-dark-skin-tone} {} {People~holding~hands:~medium-light~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {people-holding-hands-medium-skin-tone-light-skin-tone} {} {People~holding~hands:~medium~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {people-holding-hands-medium-skin-tone-medium-light-skin-tone} {} {People~holding~hands:~medium~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {people-holding-hands-medium-skin-tone} {} {People~holding~hands:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {people-holding-hands-medium-skin-tone-medium-dark-skin-tone} {} {People~holding~hands:~medium~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {people-holding-hands-medium-skin-tone-dark-skin-tone} {} {People~holding~hands:~medium~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {people-holding-hands-medium-dark-skin-tone-light-skin-tone} {} {People~holding~hands:~medium-dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {people-holding-hands-medium-dark-skin-tone-medium-light-skin-tone} {} {People~holding~hands:~medium-dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {people-holding-hands-medium-dark-skin-tone-medium-skin-tone} {} {People~holding~hands:~medium-dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {people-holding-hands-medium-dark-skin-tone} {} {People~holding~hands:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {people-holding-hands-medium-dark-skin-tone-dark-skin-tone} {} {People~holding~hands:~medium-dark~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {people-holding-hands-dark-skin-tone-light-skin-tone} {} {People~holding~hands:~dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {people-holding-hands-dark-skin-tone-medium-light-skin-tone} {} {People~holding~hands:~dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {people-holding-hands-dark-skin-tone-medium-skin-tone} {} {People~holding~hands:~dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {people-holding-hands-dark-skin-tone-medium-dark-skin-tone} {} {People~holding~hands:~dark~skin~tone,~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {people-holding-hands-dark-skin-tone} {} {People~holding~hands:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46d} {women-holding-hands} {two-women-holding-hands} {Women~holding~hands} {1.0}
\__emoji_def:nnnnn {^^^^^^01f46d^^^^^^01f3fb} {women-holding-hands-light-skin-tone} {} {Women~holding~hands:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fc} {women-holding-hands-light-skin-tone-medium-light-skin-tone} {} {Women~holding~hands:~light~skin~tone,~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fd} {women-holding-hands-light-skin-tone-medium-skin-tone} {} {Women~holding~hands:~light~skin~tone,~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fe} {women-holding-hands-light-skin-tone-medium-dark-skin-tone} {} {Women~holding~hands:~light~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3ff} {women-holding-hands-light-skin-tone-dark-skin-tone} {} {Women~holding~hands:~light~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fb} {women-holding-hands-medium-light-skin-tone-light-skin-tone} {} {Women~holding~hands:~medium-light~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46d^^^^^^01f3fc} {women-holding-hands-medium-light-skin-tone} {} {Women~holding~hands:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fd} {women-holding-hands-medium-light-skin-tone-medium-skin-tone} {} {Women~holding~hands:~medium-light~skin~tone,~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fe} {women-holding-hands-medium-light-skin-tone-medium-dark-skin-tone} {} {Women~holding~hands:~medium-light~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3ff} {women-holding-hands-medium-light-skin-tone-dark-skin-tone} {} {Women~holding~hands:~medium-light~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fb} {women-holding-hands-medium-skin-tone-light-skin-tone} {} {Women~holding~hands:~medium~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fc} {women-holding-hands-medium-skin-tone-medium-light-skin-tone} {} {Women~holding~hands:~medium~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46d^^^^^^01f3fd} {women-holding-hands-medium-skin-tone} {} {Women~holding~hands:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fe} {women-holding-hands-medium-skin-tone-medium-dark-skin-tone} {} {Women~holding~hands:~medium~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3ff} {women-holding-hands-medium-skin-tone-dark-skin-tone} {} {Women~holding~hands:~medium~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fb} {women-holding-hands-medium-dark-skin-tone-light-skin-tone} {} {Women~holding~hands:~medium-dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fc} {women-holding-hands-medium-dark-skin-tone-medium-light-skin-tone} {} {Women~holding~hands:~medium-dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fd} {women-holding-hands-medium-dark-skin-tone-medium-skin-tone} {} {Women~holding~hands:~medium-dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46d^^^^^^01f3fe} {women-holding-hands-medium-dark-skin-tone} {} {Women~holding~hands:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3ff} {women-holding-hands-medium-dark-skin-tone-dark-skin-tone} {} {Women~holding~hands:~medium-dark~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fb} {women-holding-hands-dark-skin-tone-light-skin-tone} {} {Women~holding~hands:~dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fc} {women-holding-hands-dark-skin-tone-medium-light-skin-tone} {} {Women~holding~hands:~dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fd} {women-holding-hands-dark-skin-tone-medium-skin-tone} {} {Women~holding~hands:~dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f469^^^^^^01f3fe} {women-holding-hands-dark-skin-tone-medium-dark-skin-tone} {} {Women~holding~hands:~dark~skin~tone,~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46d^^^^^^01f3ff} {women-holding-hands-dark-skin-tone} {} {Women~holding~hands:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46b} {woman-and-man-holding-hands} {couple} {Woman~and~man~holding~hands} {0.6}
\__emoji_def:nnnnn {^^^^^^01f46b^^^^^^01f3fb} {woman-and-man-holding-hands-light-skin-tone} {} {Woman~and~man~holding~hands:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {woman-and-man-holding-hands-light-skin-tone-medium-light-skin-tone} {} {Woman~and~man~holding~hands:~light~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {woman-and-man-holding-hands-light-skin-tone-medium-skin-tone} {} {Woman~and~man~holding~hands:~light~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {woman-and-man-holding-hands-light-skin-tone-medium-dark-skin-tone} {} {Woman~and~man~holding~hands:~light~skin~tone,~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {woman-and-man-holding-hands-light-skin-tone-dark-skin-tone} {} {Woman~and~man~holding~hands:~light~skin~tone,~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {woman-and-man-holding-hands-medium-light-skin-tone-light-skin-tone} {} {Woman~and~man~holding~hands:~medium-light~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46b^^^^^^01f3fc} {woman-and-man-holding-hands-medium-light-skin-tone} {} {Woman~and~man~holding~hands:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {woman-and-man-holding-hands-medium-light-skin-tone-medium-skin-tone} {} {Woman~and~man~holding~hands:~medium-light~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {woman-and-man-holding-hands-medium-light-skin-tone-medium-dark-skin-tone} {} {Woman~and~man~holding~hands:~medium-light~skin~tone,~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {woman-and-man-holding-hands-medium-light-skin-tone-dark-skin-tone} {} {Woman~and~man~holding~hands:~medium-light~skin~tone,~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {woman-and-man-holding-hands-medium-skin-tone-light-skin-tone} {} {Woman~and~man~holding~hands:~medium~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {woman-and-man-holding-hands-medium-skin-tone-medium-light-skin-tone} {} {Woman~and~man~holding~hands:~medium~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46b^^^^^^01f3fd} {woman-and-man-holding-hands-medium-skin-tone} {} {Woman~and~man~holding~hands:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {woman-and-man-holding-hands-medium-skin-tone-medium-dark-skin-tone} {} {Woman~and~man~holding~hands:~medium~skin~tone,~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {woman-and-man-holding-hands-medium-skin-tone-dark-skin-tone} {} {Woman~and~man~holding~hands:~medium~skin~tone,~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {woman-and-man-holding-hands-medium-dark-skin-tone-light-skin-tone} {} {Woman~and~man~holding~hands:~medium-dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {woman-and-man-holding-hands-medium-dark-skin-tone-medium-light-skin-tone} {} {Woman~and~man~holding~hands:~medium-dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {woman-and-man-holding-hands-medium-dark-skin-tone-medium-skin-tone} {} {Woman~and~man~holding~hands:~medium-dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46b^^^^^^01f3fe} {woman-and-man-holding-hands-medium-dark-skin-tone} {} {Woman~and~man~holding~hands:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {woman-and-man-holding-hands-medium-dark-skin-tone-dark-skin-tone} {} {Woman~and~man~holding~hands:~medium-dark~skin~tone,~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {woman-and-man-holding-hands-dark-skin-tone-light-skin-tone} {} {Woman~and~man~holding~hands:~dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {woman-and-man-holding-hands-dark-skin-tone-medium-light-skin-tone} {} {Woman~and~man~holding~hands:~dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {woman-and-man-holding-hands-dark-skin-tone-medium-skin-tone} {} {Woman~and~man~holding~hands:~dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {woman-and-man-holding-hands-dark-skin-tone-medium-dark-skin-tone} {} {Woman~and~man~holding~hands:~dark~skin~tone,~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46b^^^^^^01f3ff} {woman-and-man-holding-hands-dark-skin-tone} {} {Woman~and~man~holding~hands:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46c} {men-holding-hands} {two-men-holding-hands} {Men~holding~hands} {1.0}
\__emoji_def:nnnnn {^^^^^^01f46c^^^^^^01f3fb} {men-holding-hands-light-skin-tone} {} {Men~holding~hands:~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {men-holding-hands-light-skin-tone-medium-light-skin-tone} {} {Men~holding~hands:~light~skin~tone,~medium-light~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {men-holding-hands-light-skin-tone-medium-skin-tone} {} {Men~holding~hands:~light~skin~tone,~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {men-holding-hands-light-skin-tone-medium-dark-skin-tone} {} {Men~holding~hands:~light~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {men-holding-hands-light-skin-tone-dark-skin-tone} {} {Men~holding~hands:~light~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {men-holding-hands-medium-light-skin-tone-light-skin-tone} {} {Men~holding~hands:~medium-light~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46c^^^^^^01f3fc} {men-holding-hands-medium-light-skin-tone} {} {Men~holding~hands:~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {men-holding-hands-medium-light-skin-tone-medium-skin-tone} {} {Men~holding~hands:~medium-light~skin~tone,~medium~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {men-holding-hands-medium-light-skin-tone-medium-dark-skin-tone} {} {Men~holding~hands:~medium-light~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {men-holding-hands-medium-light-skin-tone-dark-skin-tone} {} {Men~holding~hands:~medium-light~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {men-holding-hands-medium-skin-tone-light-skin-tone} {} {Men~holding~hands:~medium~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {men-holding-hands-medium-skin-tone-medium-light-skin-tone} {} {Men~holding~hands:~medium~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46c^^^^^^01f3fd} {men-holding-hands-medium-skin-tone} {} {Men~holding~hands:~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {men-holding-hands-medium-skin-tone-medium-dark-skin-tone} {} {Men~holding~hands:~medium~skin~tone,~medium-dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {men-holding-hands-medium-skin-tone-dark-skin-tone} {} {Men~holding~hands:~medium~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {men-holding-hands-medium-dark-skin-tone-light-skin-tone} {} {Men~holding~hands:~medium-dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {men-holding-hands-medium-dark-skin-tone-medium-light-skin-tone} {} {Men~holding~hands:~medium-dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {men-holding-hands-medium-dark-skin-tone-medium-skin-tone} {} {Men~holding~hands:~medium-dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46c^^^^^^01f3fe} {men-holding-hands-medium-dark-skin-tone} {} {Men~holding~hands:~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3ff} {men-holding-hands-medium-dark-skin-tone-dark-skin-tone} {} {Men~holding~hands:~medium-dark~skin~tone,~dark~skin~tone} {12.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fb} {men-holding-hands-dark-skin-tone-light-skin-tone} {} {Men~holding~hands:~dark~skin~tone,~light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fc} {men-holding-hands-dark-skin-tone-medium-light-skin-tone} {} {Men~holding~hands:~dark~skin~tone,~medium-light~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fd} {men-holding-hands-dark-skin-tone-medium-skin-tone} {} {Men~holding~hands:~dark~skin~tone,~medium~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^^^01f91d^^^^200d^^^^^^01f468^^^^^^01f3fe} {men-holding-hands-dark-skin-tone-medium-dark-skin-tone} {} {Men~holding~hands:~dark~skin~tone,~medium-dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f46c^^^^^^01f3ff} {men-holding-hands-dark-skin-tone} {} {Men~holding~hands:~dark~skin~tone} {12.0}
\__emoji_def:nnnnn {^^^^^^01f48f} {kiss} {couplekiss} {Kiss} {0.6}
\__emoji_def:nnnnn {^^^^^^01f48f^^^^^^01f3fb} {kiss-light-skin-tone} {} {Kiss:~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f48f^^^^^^01f3fc} {kiss-medium-light-skin-tone} {} {Kiss:~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f48f^^^^^^01f3fd} {kiss-medium-skin-tone} {} {Kiss:~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f48f^^^^^^01f3fe} {kiss-medium-dark-skin-tone} {} {Kiss:~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f48f^^^^^^01f3ff} {kiss-dark-skin-tone} {} {Kiss:~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {kiss-person-person-light-skin-tone-medium-light-skin-tone} {} {Kiss:~person,~person,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {kiss-person-person-light-skin-tone-medium-skin-tone} {} {Kiss:~person,~person,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {kiss-person-person-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~person,~person,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {kiss-person-person-light-skin-tone-dark-skin-tone} {} {Kiss:~person,~person,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {kiss-person-person-medium-light-skin-tone-light-skin-tone} {} {Kiss:~person,~person,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {kiss-person-person-medium-light-skin-tone-medium-skin-tone} {} {Kiss:~person,~person,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {kiss-person-person-medium-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~person,~person,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {kiss-person-person-medium-light-skin-tone-dark-skin-tone} {} {Kiss:~person,~person,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {kiss-person-person-medium-skin-tone-light-skin-tone} {} {Kiss:~person,~person,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {kiss-person-person-medium-skin-tone-medium-light-skin-tone} {} {Kiss:~person,~person,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {kiss-person-person-medium-skin-tone-medium-dark-skin-tone} {} {Kiss:~person,~person,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {kiss-person-person-medium-skin-tone-dark-skin-tone} {} {Kiss:~person,~person,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {kiss-person-person-medium-dark-skin-tone-light-skin-tone} {} {Kiss:~person,~person,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {kiss-person-person-medium-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~person,~person,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {kiss-person-person-medium-dark-skin-tone-medium-skin-tone} {} {Kiss:~person,~person,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {kiss-person-person-medium-dark-skin-tone-dark-skin-tone} {} {Kiss:~person,~person,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {kiss-person-person-dark-skin-tone-light-skin-tone} {} {Kiss:~person,~person,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {kiss-person-person-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~person,~person,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {kiss-person-person-dark-skin-tone-medium-skin-tone} {} {Kiss:~person,~person,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {kiss-person-person-dark-skin-tone-medium-dark-skin-tone} {} {Kiss:~person,~person,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468} {kiss-woman-man} {couplekiss-man-woman} {Kiss:~woman,~man} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-woman-man-light-skin-tone} {} {Kiss:~woman,~man,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-woman-man-light-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~man,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-woman-man-light-skin-tone-medium-skin-tone} {} {Kiss:~woman,~man,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-woman-man-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~man,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-woman-man-light-skin-tone-dark-skin-tone} {} {Kiss:~woman,~man,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-woman-man-medium-light-skin-tone-light-skin-tone} {} {Kiss:~woman,~man,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-woman-man-medium-light-skin-tone} {} {Kiss:~woman,~man,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-woman-man-medium-light-skin-tone-medium-skin-tone} {} {Kiss:~woman,~man,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-woman-man-medium-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~man,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-woman-man-medium-light-skin-tone-dark-skin-tone} {} {Kiss:~woman,~man,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-woman-man-medium-skin-tone-light-skin-tone} {} {Kiss:~woman,~man,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-woman-man-medium-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~man,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-woman-man-medium-skin-tone} {} {Kiss:~woman,~man,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-woman-man-medium-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~man,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-woman-man-medium-skin-tone-dark-skin-tone} {} {Kiss:~woman,~man,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-woman-man-medium-dark-skin-tone-light-skin-tone} {} {Kiss:~woman,~man,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-woman-man-medium-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~man,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-woman-man-medium-dark-skin-tone-medium-skin-tone} {} {Kiss:~woman,~man,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-woman-man-medium-dark-skin-tone} {} {Kiss:~woman,~man,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-woman-man-medium-dark-skin-tone-dark-skin-tone} {} {Kiss:~woman,~man,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-woman-man-dark-skin-tone-light-skin-tone} {} {Kiss:~woman,~man,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-woman-man-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~man,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-woman-man-dark-skin-tone-medium-skin-tone} {} {Kiss:~woman,~man,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-woman-man-dark-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~man,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-woman-man-dark-skin-tone} {} {Kiss:~woman,~man,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468} {kiss-man-man} {couplekiss-man-man} {Kiss:~man,~man} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-man-man-light-skin-tone} {} {Kiss:~man,~man,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-man-man-light-skin-tone-medium-light-skin-tone} {} {Kiss:~man,~man,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-man-man-light-skin-tone-medium-skin-tone} {} {Kiss:~man,~man,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-man-man-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~man,~man,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-man-man-light-skin-tone-dark-skin-tone} {} {Kiss:~man,~man,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-man-man-medium-light-skin-tone-light-skin-tone} {} {Kiss:~man,~man,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-man-man-medium-light-skin-tone} {} {Kiss:~man,~man,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-man-man-medium-light-skin-tone-medium-skin-tone} {} {Kiss:~man,~man,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-man-man-medium-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~man,~man,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-man-man-medium-light-skin-tone-dark-skin-tone} {} {Kiss:~man,~man,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-man-man-medium-skin-tone-light-skin-tone} {} {Kiss:~man,~man,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-man-man-medium-skin-tone-medium-light-skin-tone} {} {Kiss:~man,~man,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-man-man-medium-skin-tone} {} {Kiss:~man,~man,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-man-man-medium-skin-tone-medium-dark-skin-tone} {} {Kiss:~man,~man,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-man-man-medium-skin-tone-dark-skin-tone} {} {Kiss:~man,~man,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-man-man-medium-dark-skin-tone-light-skin-tone} {} {Kiss:~man,~man,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-man-man-medium-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~man,~man,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-man-man-medium-dark-skin-tone-medium-skin-tone} {} {Kiss:~man,~man,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-man-man-medium-dark-skin-tone} {} {Kiss:~man,~man,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-man-man-medium-dark-skin-tone-dark-skin-tone} {} {Kiss:~man,~man,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fb} {kiss-man-man-dark-skin-tone-light-skin-tone} {} {Kiss:~man,~man,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fc} {kiss-man-man-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~man,~man,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fd} {kiss-man-man-dark-skin-tone-medium-skin-tone} {} {Kiss:~man,~man,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3fe} {kiss-man-man-dark-skin-tone-medium-dark-skin-tone} {} {Kiss:~man,~man,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f468^^^^^^01f3ff} {kiss-man-man-dark-skin-tone} {} {Kiss:~man,~man,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469} {kiss-woman-woman} {couplekiss-woman-woman} {Kiss:~woman,~woman} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fb} {kiss-woman-woman-light-skin-tone} {} {Kiss:~woman,~woman,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fc} {kiss-woman-woman-light-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~woman,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fd} {kiss-woman-woman-light-skin-tone-medium-skin-tone} {} {Kiss:~woman,~woman,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fe} {kiss-woman-woman-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~woman,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3ff} {kiss-woman-woman-light-skin-tone-dark-skin-tone} {} {Kiss:~woman,~woman,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fb} {kiss-woman-woman-medium-light-skin-tone-light-skin-tone} {} {Kiss:~woman,~woman,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fc} {kiss-woman-woman-medium-light-skin-tone} {} {Kiss:~woman,~woman,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fd} {kiss-woman-woman-medium-light-skin-tone-medium-skin-tone} {} {Kiss:~woman,~woman,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fe} {kiss-woman-woman-medium-light-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~woman,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3ff} {kiss-woman-woman-medium-light-skin-tone-dark-skin-tone} {} {Kiss:~woman,~woman,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fb} {kiss-woman-woman-medium-skin-tone-light-skin-tone} {} {Kiss:~woman,~woman,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fc} {kiss-woman-woman-medium-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~woman,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fd} {kiss-woman-woman-medium-skin-tone} {} {Kiss:~woman,~woman,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fe} {kiss-woman-woman-medium-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~woman,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3ff} {kiss-woman-woman-medium-skin-tone-dark-skin-tone} {} {Kiss:~woman,~woman,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fb} {kiss-woman-woman-medium-dark-skin-tone-light-skin-tone} {} {Kiss:~woman,~woman,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fc} {kiss-woman-woman-medium-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~woman,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fd} {kiss-woman-woman-medium-dark-skin-tone-medium-skin-tone} {} {Kiss:~woman,~woman,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fe} {kiss-woman-woman-medium-dark-skin-tone} {} {Kiss:~woman,~woman,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3ff} {kiss-woman-woman-medium-dark-skin-tone-dark-skin-tone} {} {Kiss:~woman,~woman,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fb} {kiss-woman-woman-dark-skin-tone-light-skin-tone} {} {Kiss:~woman,~woman,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fc} {kiss-woman-woman-dark-skin-tone-medium-light-skin-tone} {} {Kiss:~woman,~woman,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fd} {kiss-woman-woman-dark-skin-tone-medium-skin-tone} {} {Kiss:~woman,~woman,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3fe} {kiss-woman-woman-dark-skin-tone-medium-dark-skin-tone} {} {Kiss:~woman,~woman,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f48b^^^^200d^^^^^^01f469^^^^^^01f3ff} {kiss-woman-woman-dark-skin-tone} {} {Kiss:~woman,~woman,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f491} {couple-with-heart} {} {Couple~with~heart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f491^^^^^^01f3fb} {couple-with-heart-light-skin-tone} {} {Couple~with~heart:~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f491^^^^^^01f3fc} {couple-with-heart-medium-light-skin-tone} {} {Couple~with~heart:~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f491^^^^^^01f3fd} {couple-with-heart-medium-skin-tone} {} {Couple~with~heart:~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f491^^^^^^01f3fe} {couple-with-heart-medium-dark-skin-tone} {} {Couple~with~heart:~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f491^^^^^^01f3ff} {couple-with-heart-dark-skin-tone} {} {Couple~with~heart:~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {couple-with-heart-person-person-light-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~person,~person,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {couple-with-heart-person-person-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~person,~person,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {couple-with-heart-person-person-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~person,~person,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {couple-with-heart-person-person-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~person,~person,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {couple-with-heart-person-person-medium-light-skin-tone-light-skin-tone} {} {Couple~with~heart:~person,~person,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {couple-with-heart-person-person-medium-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~person,~person,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {couple-with-heart-person-person-medium-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~person,~person,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {couple-with-heart-person-person-medium-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~person,~person,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {couple-with-heart-person-person-medium-skin-tone-light-skin-tone} {} {Couple~with~heart:~person,~person,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {couple-with-heart-person-person-medium-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~person,~person,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {couple-with-heart-person-person-medium-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~person,~person,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {couple-with-heart-person-person-medium-skin-tone-dark-skin-tone} {} {Couple~with~heart:~person,~person,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {couple-with-heart-person-person-medium-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~person,~person,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {couple-with-heart-person-person-medium-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~person,~person,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {couple-with-heart-person-person-medium-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~person,~person,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3ff} {couple-with-heart-person-person-medium-dark-skin-tone-dark-skin-tone} {} {Couple~with~heart:~person,~person,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fb} {couple-with-heart-person-person-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~person,~person,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fc} {couple-with-heart-person-person-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~person,~person,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fd} {couple-with-heart-person-person-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~person,~person,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f9d1^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f9d1^^^^^^01f3fe} {couple-with-heart-person-person-dark-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~person,~person,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468} {couple-with-heart-woman-man} {} {Couple~with~heart:~woman,~man} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-woman-man-light-skin-tone} {} {Couple~with~heart:~woman,~man,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-woman-man-light-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~man,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-woman-man-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~man,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-woman-man-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-woman-man-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-woman-man-medium-light-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-woman-man-medium-light-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-woman-man-medium-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-woman-man-medium-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-woman-man-medium-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-woman-man-medium-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~man,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-woman-man-medium-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~man,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-woman-man-medium-skin-tone} {} {Couple~with~heart:~woman,~man,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-woman-man-medium-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-woman-man-medium-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-woman-man-medium-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-woman-man-medium-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-woman-man-medium-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-woman-man-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-woman-man-medium-dark-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-woman-man-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~man,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-woman-man-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~man,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-woman-man-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~man,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-woman-man-dark-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-woman-man-dark-skin-tone} {} {Couple~with~heart:~woman,~man,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468} {couple-with-heart-man-man} {} {Couple~with~heart:~man,~man} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-man-man-light-skin-tone} {} {Couple~with~heart:~man,~man,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-man-man-light-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~man,~man,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-man-man-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~man,~man,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-man-man-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~man,~man,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-man-man-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~man,~man,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-man-man-medium-light-skin-tone-light-skin-tone} {} {Couple~with~heart:~man,~man,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-man-man-medium-light-skin-tone} {} {Couple~with~heart:~man,~man,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-man-man-medium-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~man,~man,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-man-man-medium-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~man,~man,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-man-man-medium-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~man,~man,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-man-man-medium-skin-tone-light-skin-tone} {} {Couple~with~heart:~man,~man,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-man-man-medium-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~man,~man,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-man-man-medium-skin-tone} {} {Couple~with~heart:~man,~man,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-man-man-medium-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~man,~man,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-man-man-medium-skin-tone-dark-skin-tone} {} {Couple~with~heart:~man,~man,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-man-man-medium-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~man,~man,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-man-man-medium-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~man,~man,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-man-man-medium-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~man,~man,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-man-man-medium-dark-skin-tone} {} {Couple~with~heart:~man,~man,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-man-man-medium-dark-skin-tone-dark-skin-tone} {} {Couple~with~heart:~man,~man,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fb} {couple-with-heart-man-man-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~man,~man,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fc} {couple-with-heart-man-man-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~man,~man,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fd} {couple-with-heart-man-man-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~man,~man,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3fe} {couple-with-heart-man-man-dark-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~man,~man,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f468^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f468^^^^^^01f3ff} {couple-with-heart-man-man-dark-skin-tone} {} {Couple~with~heart:~man,~man,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469} {couple-with-heart-woman-woman} {} {Couple~with~heart:~woman,~woman} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fb} {couple-with-heart-woman-woman-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fc} {couple-with-heart-woman-woman-light-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~light~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fd} {couple-with-heart-woman-woman-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~woman,~light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fe} {couple-with-heart-woman-woman-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fb^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3ff} {couple-with-heart-woman-woman-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fb} {couple-with-heart-woman-woman-medium-light-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-light~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fc} {couple-with-heart-woman-woman-medium-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fd} {couple-with-heart-woman-woman-medium-light-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-light~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fe} {couple-with-heart-woman-woman-medium-light-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-light~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fc^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3ff} {couple-with-heart-woman-woman-medium-light-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-light~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fb} {couple-with-heart-woman-woman-medium-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fc} {couple-with-heart-woman-woman-medium-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fd} {couple-with-heart-woman-woman-medium-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fe} {couple-with-heart-woman-woman-medium-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fd^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3ff} {couple-with-heart-woman-woman-medium-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fb} {couple-with-heart-woman-woman-medium-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fc} {couple-with-heart-woman-woman-medium-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fd} {couple-with-heart-woman-woman-medium-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fe} {couple-with-heart-woman-woman-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3fe^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3ff} {couple-with-heart-woman-woman-medium-dark-skin-tone-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~medium-dark~skin~tone,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fb} {couple-with-heart-woman-woman-dark-skin-tone-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~dark~skin~tone,~light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fc} {couple-with-heart-woman-woman-dark-skin-tone-medium-light-skin-tone} {} {Couple~with~heart:~woman,~woman,~dark~skin~tone,~medium-light~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fd} {couple-with-heart-woman-woman-dark-skin-tone-medium-skin-tone} {} {Couple~with~heart:~woman,~woman,~dark~skin~tone,~medium~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3fe} {couple-with-heart-woman-woman-dark-skin-tone-medium-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~dark~skin~tone,~medium-dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f469^^^^^^01f3ff^^^^200d^^^^2764^^^^fe0f^^^^200d^^^^^^01f469^^^^^^01f3ff} {couple-with-heart-woman-woman-dark-skin-tone} {} {Couple~with~heart:~woman,~woman,~dark~skin~tone} {13.1}
\__emoji_def:nnnnn {^^^^^^01f46a} {family} {} {Family} {0.6}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f469^^^^200d^^^^^^01f466} {family-man-woman-boy} {} {Family:~man,~woman,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f469^^^^200d^^^^^^01f467} {family-man-woman-girl} {} {Family:~man,~woman,~girl} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f469^^^^200d^^^^^^01f467^^^^200d^^^^^^01f466} {family-man-woman-girl-boy} {} {Family:~man,~woman,~girl,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f469^^^^200d^^^^^^01f466^^^^200d^^^^^^01f466} {family-man-woman-boy-boy} {} {Family:~man,~woman,~boy,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f469^^^^200d^^^^^^01f467^^^^200d^^^^^^01f467} {family-man-woman-girl-girl} {} {Family:~man,~woman,~girl,~girl} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f468^^^^200d^^^^^^01f466} {family-man-man-boy} {} {Family:~man,~man,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f468^^^^200d^^^^^^01f467} {family-man-man-girl} {} {Family:~man,~man,~girl} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f468^^^^200d^^^^^^01f467^^^^200d^^^^^^01f466} {family-man-man-girl-boy} {} {Family:~man,~man,~girl,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f468^^^^200d^^^^^^01f466^^^^200d^^^^^^01f466} {family-man-man-boy-boy} {} {Family:~man,~man,~boy,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f468^^^^200d^^^^^^01f467^^^^200d^^^^^^01f467} {family-man-man-girl-girl} {} {Family:~man,~man,~girl,~girl} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f469^^^^200d^^^^^^01f466} {family-woman-woman-boy} {} {Family:~woman,~woman,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f469^^^^200d^^^^^^01f467} {family-woman-woman-girl} {} {Family:~woman,~woman,~girl} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f469^^^^200d^^^^^^01f467^^^^200d^^^^^^01f466} {family-woman-woman-girl-boy} {} {Family:~woman,~woman,~girl,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f469^^^^200d^^^^^^01f466^^^^200d^^^^^^01f466} {family-woman-woman-boy-boy} {} {Family:~woman,~woman,~boy,~boy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f469^^^^200d^^^^^^01f467^^^^200d^^^^^^01f467} {family-woman-woman-girl-girl} {} {Family:~woman,~woman,~girl,~girl} {2.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f466} {family-man-boy} {} {Family:~man,~boy} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f466^^^^200d^^^^^^01f466} {family-man-boy-boy} {} {Family:~man,~boy,~boy} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f467} {family-man-girl} {} {Family:~man,~girl} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f467^^^^200d^^^^^^01f466} {family-man-girl-boy} {} {Family:~man,~girl,~boy} {4.0}
\__emoji_def:nnnnn {^^^^^^01f468^^^^200d^^^^^^01f467^^^^200d^^^^^^01f467} {family-man-girl-girl} {} {Family:~man,~girl,~girl} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f466} {family-woman-boy} {} {Family:~woman,~boy} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f466^^^^200d^^^^^^01f466} {family-woman-boy-boy} {} {Family:~woman,~boy,~boy} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f467} {family-woman-girl} {} {Family:~woman,~girl} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f467^^^^200d^^^^^^01f466} {family-woman-girl-boy} {} {Family:~woman,~girl,~boy} {4.0}
\__emoji_def:nnnnn {^^^^^^01f469^^^^200d^^^^^^01f467^^^^200d^^^^^^01f467} {family-woman-girl-girl} {} {Family:~woman,~girl,~girl} {4.0}
\__emoji_subgroup:n {Person-symbol}
\__emoji_def:nnnnn {^^^^^^01f5e3^^^^fe0f} {speaking-head} {} {Speaking~head} {0.7}
\__emoji_def:nnnnn {^^^^^^01f464} {bust-in-silhouette} {} {Bust~in~silhouette} {0.6}
\__emoji_def:nnnnn {^^^^^^01f465} {busts-in-silhouette} {} {Busts~in~silhouette} {1.0}
\__emoji_def:nnnnn {^^^^^^01fac2} {people-hugging} {} {People~hugging} {13.0}
\__emoji_def:nnnnn {^^^^^^01f463} {footprints} {} {Footprints} {0.6}
\__emoji_group:n {Component}
\__emoji_subgroup:n {Skin-tone}
\__emoji_def:nnnnn {^^^^^^01f3fb} {light-skin-tone} {} {Light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3fc} {medium-light-skin-tone} {} {Medium-light~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3fd} {medium-skin-tone} {} {Medium~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3fe} {medium-dark-skin-tone} {} {Medium-dark~skin~tone} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3ff} {dark-skin-tone} {} {Dark~skin~tone} {1.0}
\__emoji_subgroup:n {Hair-style}
\__emoji_def:nnnnn {^^^^^^01f9b0} {red-hair} {} {Red~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b1} {curly-hair} {} {Curly~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b3} {white-hair} {} {White~hair} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9b2} {bald} {} {Bald} {11.0}
\__emoji_group:n {Animals~\&~Nature}
\__emoji_subgroup:n {Animal-mammal}
\__emoji_def:nnnnn {^^^^^^01f435} {monkey-face} {} {Monkey~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f412} {monkey} {} {Monkey} {0.6}
\__emoji_def:nnnnn {^^^^^^01f98d} {gorilla} {} {Gorilla} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9a7} {orangutan} {} {Orangutan} {12.0}
\__emoji_def:nnnnn {^^^^^^01f436} {dog-face} {} {Dog~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f415} {dog} {dog2} {Dog} {0.7}
\__emoji_def:nnnnn {^^^^^^01f9ae} {guide-dog} {} {Guide~dog} {12.0}
\__emoji_def:nnnnn {^^^^^^01f415^^^^200d^^^^^^01f9ba} {service-dog} {} {Service~dog} {12.0}
\__emoji_def:nnnnn {^^^^^^01f429} {poodle} {} {Poodle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f43a} {wolf} {} {Wolf} {0.6}
\__emoji_def:nnnnn {^^^^^^01f98a} {fox} {fox-face} {Fox} {3.0}
\__emoji_def:nnnnn {^^^^^^01f99d} {raccoon} {} {Raccoon} {11.0}
\__emoji_def:nnnnn {^^^^^^01f431} {cat-face} {} {Cat~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f408} {cat} {cat2} {Cat} {0.7}
\__emoji_def:nnnnn {^^^^^^01f408^^^^200d^^^^2b1b} {black-cat} {} {Black~cat} {13.0}
\__emoji_def:nnnnn {^^^^^^01f981} {lion} {} {Lion} {1.0}
\__emoji_def:nnnnn {^^^^^^01f42f} {tiger-face} {} {Tiger~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f405} {tiger} {tiger2} {Tiger} {1.0}
\__emoji_def:nnnnn {^^^^^^01f406} {leopard} {} {Leopard} {1.0}
\__emoji_def:nnnnn {^^^^^^01f434} {horse-face} {} {Horse~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f40e} {horse} {racehorse} {Horse} {0.6}
\__emoji_def:nnnnn {^^^^^^01f984} {unicorn} {} {Unicorn} {1.0}
\__emoji_def:nnnnn {^^^^^^01f993} {zebra} {} {Zebra} {5.0}
\__emoji_def:nnnnn {^^^^^^01f98c} {deer} {} {Deer} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9ac} {bison} {} {Bison} {13.0}
\__emoji_def:nnnnn {^^^^^^01f42e} {cow-face} {} {Cow~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f402} {ox} {} {Ox} {1.0}
\__emoji_def:nnnnn {^^^^^^01f403} {water-buffalo} {} {Water~buffalo} {1.0}
\__emoji_def:nnnnn {^^^^^^01f404} {cow} {cow2} {Cow} {1.0}
\__emoji_def:nnnnn {^^^^^^01f437} {pig-face} {} {Pig~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f416} {pig} {pig2} {Pig} {1.0}
\__emoji_def:nnnnn {^^^^^^01f417} {boar} {} {Boar} {0.6}
\__emoji_def:nnnnn {^^^^^^01f43d} {pig-nose} {} {Pig~nose} {0.6}
\__emoji_def:nnnnn {^^^^^^01f40f} {ram} {} {Ram} {1.0}
\__emoji_def:nnnnn {^^^^^^01f411} {ewe} {sheep} {Ewe} {0.6}
\__emoji_def:nnnnn {^^^^^^01f410} {goat} {} {Goat} {1.0}
\__emoji_def:nnnnn {^^^^^^01f42a} {camel} {dromedary-camel} {Camel} {1.0}
\__emoji_def:nnnnn {^^^^^^01f42b} {two-hump-camel} {} {Two-hump~camel} {0.6}
\__emoji_def:nnnnn {^^^^^^01f999} {llama} {} {Llama} {11.0}
\__emoji_def:nnnnn {^^^^^^01f992} {giraffe} {} {Giraffe} {5.0}
\__emoji_def:nnnnn {^^^^^^01f418} {elephant} {} {Elephant} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9a3} {mammoth} {} {Mammoth} {13.0}
\__emoji_def:nnnnn {^^^^^^01f98f} {rhinoceros} {} {Rhinoceros} {3.0}
\__emoji_def:nnnnn {^^^^^^01f99b} {hippopotamus} {} {Hippopotamus} {11.0}
\__emoji_def:nnnnn {^^^^^^01f42d} {mouse-face} {} {Mouse~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f401} {mouse} {mouse2} {Mouse} {1.0}
\__emoji_def:nnnnn {^^^^^^01f400} {rat} {} {Rat} {1.0}
\__emoji_def:nnnnn {^^^^^^01f439} {hamster} {} {Hamster} {0.6}
\__emoji_def:nnnnn {^^^^^^01f430} {rabbit-face} {} {Rabbit~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f407} {rabbit} {rabbit2} {Rabbit} {1.0}
\__emoji_def:nnnnn {^^^^^^01f43f^^^^fe0f} {chipmunk} {} {Chipmunk} {0.7}
\__emoji_def:nnnnn {^^^^^^01f9ab} {beaver} {} {Beaver} {13.0}
\__emoji_def:nnnnn {^^^^^^01f994} {hedgehog} {} {Hedgehog} {5.0}
\__emoji_def:nnnnn {^^^^^^01f987} {bat} {} {Bat} {3.0}
\__emoji_def:nnnnn {^^^^^^01f43b} {bear} {} {Bear} {0.6}
\__emoji_def:nnnnn {^^^^^^01f43b^^^^200d^^^^2744^^^^fe0f} {polar-bear} {} {Polar~bear} {13.0}
\__emoji_def:nnnnn {^^^^^^01f428} {koala} {} {Koala} {0.6}
\__emoji_def:nnnnn {^^^^^^01f43c} {panda} {panda-face} {Panda} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9a5} {sloth} {} {Sloth} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9a6} {otter} {} {Otter} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9a8} {skunk} {} {Skunk} {12.0}
\__emoji_def:nnnnn {^^^^^^01f998} {kangaroo} {} {Kangaroo} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9a1} {badger} {} {Badger} {11.0}
\__emoji_def:nnnnn {^^^^^^01f43e} {paw-prints} {feet} {Paw~prints} {0.6}
\__emoji_subgroup:n {Animal-bird}
\__emoji_def:nnnnn {^^^^^^01f983} {turkey} {} {Turkey} {1.0}
\__emoji_def:nnnnn {^^^^^^01f414} {chicken} {} {Chicken} {0.6}
\__emoji_def:nnnnn {^^^^^^01f413} {rooster} {} {Rooster} {1.0}
\__emoji_def:nnnnn {^^^^^^01f423} {hatching-chick} {} {Hatching~chick} {0.6}
\__emoji_def:nnnnn {^^^^^^01f424} {baby-chick} {} {Baby~chick} {0.6}
\__emoji_def:nnnnn {^^^^^^01f425} {front-facing-baby-chick} {hatched-chick} {Front-facing~baby~chick} {0.6}
\__emoji_def:nnnnn {^^^^^^01f426} {bird} {} {Bird} {0.6}
\__emoji_def:nnnnn {^^^^^^01f427} {penguin} {} {Penguin} {0.6}
\__emoji_def:nnnnn {^^^^^^01f54a^^^^fe0f} {dove} {} {Dove} {0.7}
\__emoji_def:nnnnn {^^^^^^01f985} {eagle} {} {Eagle} {3.0}
\__emoji_def:nnnnn {^^^^^^01f986} {duck} {} {Duck} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9a2} {swan} {} {Swan} {11.0}
\__emoji_def:nnnnn {^^^^^^01f989} {owl} {} {Owl} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9a4} {dodo} {} {Dodo} {13.0}
\__emoji_def:nnnnn {^^^^^^01fab6} {feather} {} {Feather} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9a9} {flamingo} {} {Flamingo} {12.0}
\__emoji_def:nnnnn {^^^^^^01f99a} {peacock} {} {Peacock} {11.0}
\__emoji_def:nnnnn {^^^^^^01f99c} {parrot} {} {Parrot} {11.0}
\__emoji_subgroup:n {Animal-amphibian}
\__emoji_def:nnnnn {^^^^^^01f438} {frog} {} {Frog} {0.6}
\__emoji_subgroup:n {Animal-reptile}
\__emoji_def:nnnnn {^^^^^^01f40a} {crocodile} {} {Crocodile} {1.0}
\__emoji_def:nnnnn {^^^^^^01f422} {turtle} {} {Turtle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f98e} {lizard} {} {Lizard} {3.0}
\__emoji_def:nnnnn {^^^^^^01f40d} {snake} {} {Snake} {0.6}
\__emoji_def:nnnnn {^^^^^^01f432} {dragon-face} {} {Dragon~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f409} {dragon} {} {Dragon} {1.0}
\__emoji_def:nnnnn {^^^^^^01f995} {sauropod} {} {Sauropod} {5.0}
\__emoji_def:nnnnn {^^^^^^01f996} {t-rex} {} {T-Rex} {5.0}
\__emoji_subgroup:n {Animal-marine}
\__emoji_def:nnnnn {^^^^^^01f433} {spouting-whale} {} {Spouting~whale} {0.6}
\__emoji_def:nnnnn {^^^^^^01f40b} {whale} {whale2} {Whale} {1.0}
\__emoji_def:nnnnn {^^^^^^01f42c} {dolphin} {flipper} {Dolphin} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9ad} {seal} {} {Seal} {13.0}
\__emoji_def:nnnnn {^^^^^^01f41f} {fish} {} {Fish} {0.6}
\__emoji_def:nnnnn {^^^^^^01f420} {tropical-fish} {} {Tropical~fish} {0.6}
\__emoji_def:nnnnn {^^^^^^01f421} {blowfish} {} {Blowfish} {0.6}
\__emoji_def:nnnnn {^^^^^^01f988} {shark} {} {Shark} {3.0}
\__emoji_def:nnnnn {^^^^^^01f419} {octopus} {} {Octopus} {0.6}
\__emoji_def:nnnnn {^^^^^^01f41a} {spiral-shell} {shell} {Spiral~shell} {0.6}
\__emoji_subgroup:n {Animal-bug}
\__emoji_def:nnnnn {^^^^^^01f40c} {snail} {} {Snail} {0.6}
\__emoji_def:nnnnn {^^^^^^01f98b} {butterfly} {} {Butterfly} {3.0}
\__emoji_def:nnnnn {^^^^^^01f41b} {bug} {} {Bug} {0.6}
\__emoji_def:nnnnn {^^^^^^01f41c} {ant} {} {Ant} {0.6}
\__emoji_def:nnnnn {^^^^^^01f41d} {honeybee} {bee} {Honeybee} {0.6}
\__emoji_def:nnnnn {^^^^^^01fab2} {beetle} {} {Beetle} {13.0}
\__emoji_def:nnnnn {^^^^^^01f41e} {lady-beetle} {} {Lady~beetle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f997} {cricket} {} {Cricket} {5.0}
\__emoji_def:nnnnn {^^^^^^01fab3} {cockroach} {} {Cockroach} {13.0}
\__emoji_def:nnnnn {^^^^^^01f577^^^^fe0f} {spider} {} {Spider} {0.7}
\__emoji_def:nnnnn {^^^^^^01f578^^^^fe0f} {spider-web} {} {Spider~web} {0.7}
\__emoji_def:nnnnn {^^^^^^01f982} {scorpion} {} {Scorpion} {1.0}
\__emoji_def:nnnnn {^^^^^^01f99f} {mosquito} {} {Mosquito} {11.0}
\__emoji_def:nnnnn {^^^^^^01fab0} {fly} {} {Fly} {13.0}
\__emoji_def:nnnnn {^^^^^^01fab1} {worm} {} {Worm} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9a0} {microbe} {} {Microbe} {11.0}
\__emoji_subgroup:n {Plant-flower}
\__emoji_def:nnnnn {^^^^^^01f490} {bouquet} {} {Bouquet} {0.6}
\__emoji_def:nnnnn {^^^^^^01f338} {cherry-blossom} {} {Cherry~blossom} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ae} {white-flower} {} {White~flower} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3f5^^^^fe0f} {rosette} {} {Rosette} {0.7}
\__emoji_def:nnnnn {^^^^^^01f339} {rose} {} {Rose} {0.6}
\__emoji_def:nnnnn {^^^^^^01f940} {wilted-flower} {} {Wilted~flower} {3.0}
\__emoji_def:nnnnn {^^^^^^01f33a} {hibiscus} {} {Hibiscus} {0.6}
\__emoji_def:nnnnn {^^^^^^01f33b} {sunflower} {} {Sunflower} {0.6}
\__emoji_def:nnnnn {^^^^^^01f33c} {blossom} {} {Blossom} {0.6}
\__emoji_def:nnnnn {^^^^^^01f337} {tulip} {} {Tulip} {0.6}
\__emoji_subgroup:n {Plant-other}
\__emoji_def:nnnnn {^^^^^^01f331} {seedling} {} {Seedling} {0.6}
\__emoji_def:nnnnn {^^^^^^01fab4} {potted-plant} {} {Potted~plant} {13.0}
\__emoji_def:nnnnn {^^^^^^01f332} {evergreen-tree} {} {Evergreen~tree} {1.0}
\__emoji_def:nnnnn {^^^^^^01f333} {deciduous-tree} {} {Deciduous~tree} {1.0}
\__emoji_def:nnnnn {^^^^^^01f334} {palm-tree} {} {Palm~tree} {0.6}
\__emoji_def:nnnnn {^^^^^^01f335} {cactus} {} {Cactus} {0.6}
\__emoji_def:nnnnn {^^^^^^01f33e} {sheaf-of-rice} {ear-of-rice} {Sheaf~of~rice} {0.6}
\__emoji_def:nnnnn {^^^^^^01f33f} {herb} {} {Herb} {0.6}
\__emoji_def:nnnnn {^^^^2618^^^^fe0f} {shamrock} {} {Shamrock} {1.0}
\__emoji_def:nnnnn {^^^^^^01f340} {four-leaf-clover} {} {Four~leaf~clover} {0.6}
\__emoji_def:nnnnn {^^^^^^01f341} {maple-leaf} {} {Maple~leaf} {0.6}
\__emoji_def:nnnnn {^^^^^^01f342} {fallen-leaf} {} {Fallen~leaf} {0.6}
\__emoji_def:nnnnn {^^^^^^01f343} {leaf-fluttering-in-wind} {leaves} {Leaf~fluttering~in~wind} {0.6}
\__emoji_group:n {Food~\&~Drink}
\__emoji_subgroup:n {Food-fruit}
\__emoji_def:nnnnn {^^^^^^01f347} {grapes} {} {Grapes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f348} {melon} {} {Melon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f349} {watermelon} {} {Watermelon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f34a} {tangerine} {mandarin, orange} {Tangerine} {0.6}
\__emoji_def:nnnnn {^^^^^^01f34b} {lemon} {} {Lemon} {1.0}
\__emoji_def:nnnnn {^^^^^^01f34c} {banana} {} {Banana} {0.6}
\__emoji_def:nnnnn {^^^^^^01f34d} {pineapple} {} {Pineapple} {0.6}
\__emoji_def:nnnnn {^^^^^^01f96d} {mango} {} {Mango} {11.0}
\__emoji_def:nnnnn {^^^^^^01f34e} {red-apple} {apple} {Red~apple} {0.6}
\__emoji_def:nnnnn {^^^^^^01f34f} {green-apple} {} {Green~apple} {0.6}
\__emoji_def:nnnnn {^^^^^^01f350} {pear} {} {Pear} {1.0}
\__emoji_def:nnnnn {^^^^^^01f351} {peach} {} {Peach} {0.6}
\__emoji_def:nnnnn {^^^^^^01f352} {cherries} {} {Cherries} {0.6}
\__emoji_def:nnnnn {^^^^^^01f353} {strawberry} {} {Strawberry} {0.6}
\__emoji_def:nnnnn {^^^^^^01fad0} {blueberries} {} {Blueberries} {13.0}
\__emoji_def:nnnnn {^^^^^^01f95d} {kiwi-fruit} {} {Kiwi~fruit} {3.0}
\__emoji_def:nnnnn {^^^^^^01f345} {tomato} {} {Tomato} {0.6}
\__emoji_def:nnnnn {^^^^^^01fad2} {olive} {} {Olive} {13.0}
\__emoji_def:nnnnn {^^^^^^01f965} {coconut} {} {Coconut} {5.0}
\__emoji_subgroup:n {Food-vegetable}
\__emoji_def:nnnnn {^^^^^^01f951} {avocado} {} {Avocado} {3.0}
\__emoji_def:nnnnn {^^^^^^01f346} {eggplant} {} {Eggplant} {0.6}
\__emoji_def:nnnnn {^^^^^^01f954} {potato} {} {Potato} {3.0}
\__emoji_def:nnnnn {^^^^^^01f955} {carrot} {} {Carrot} {3.0}
\__emoji_def:nnnnn {^^^^^^01f33d} {ear-of-corn} {corn} {Ear~of~corn} {0.6}
\__emoji_def:nnnnn {^^^^^^01f336^^^^fe0f} {hot-pepper} {} {Hot~pepper} {0.7}
\__emoji_def:nnnnn {^^^^^^01fad1} {bell-pepper} {} {Bell~pepper} {13.0}
\__emoji_def:nnnnn {^^^^^^01f952} {cucumber} {} {Cucumber} {3.0}
\__emoji_def:nnnnn {^^^^^^01f96c} {leafy-green} {} {Leafy~green} {11.0}
\__emoji_def:nnnnn {^^^^^^01f966} {broccoli} {} {Broccoli} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9c4} {garlic} {} {Garlic} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9c5} {onion} {} {Onion} {12.0}
\__emoji_def:nnnnn {^^^^^^01f344} {mushroom} {} {Mushroom} {0.6}
\__emoji_def:nnnnn {^^^^^^01f95c} {peanuts} {} {Peanuts} {3.0}
\__emoji_def:nnnnn {^^^^^^01f330} {chestnut} {} {Chestnut} {0.6}
\__emoji_subgroup:n {Food-prepared}
\__emoji_def:nnnnn {^^^^^^01f35e} {bread} {} {Bread} {0.6}
\__emoji_def:nnnnn {^^^^^^01f950} {croissant} {} {Croissant} {3.0}
\__emoji_def:nnnnn {^^^^^^01f956} {baguette-bread} {} {Baguette~bread} {3.0}
\__emoji_def:nnnnn {^^^^^^01fad3} {flatbread} {} {Flatbread} {13.0}
\__emoji_def:nnnnn {^^^^^^01f968} {pretzel} {} {Pretzel} {5.0}
\__emoji_def:nnnnn {^^^^^^01f96f} {bagel} {} {Bagel} {11.0}
\__emoji_def:nnnnn {^^^^^^01f95e} {pancakes} {} {Pancakes} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9c7} {waffle} {} {Waffle} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9c0} {cheese-wedge} {cheese} {Cheese~wedge} {1.0}
\__emoji_def:nnnnn {^^^^^^01f356} {meat-on-bone} {} {Meat~on~bone} {0.6}
\__emoji_def:nnnnn {^^^^^^01f357} {poultry-leg} {} {Poultry~leg} {0.6}
\__emoji_def:nnnnn {^^^^^^01f969} {cut-of-meat} {} {Cut~of~meat} {5.0}
\__emoji_def:nnnnn {^^^^^^01f953} {bacon} {} {Bacon} {3.0}
\__emoji_def:nnnnn {^^^^^^01f354} {hamburger} {} {Hamburger} {0.6}
\__emoji_def:nnnnn {^^^^^^01f35f} {french-fries} {fries} {French~fries} {0.6}
\__emoji_def:nnnnn {^^^^^^01f355} {pizza} {} {Pizza} {0.6}
\__emoji_def:nnnnn {^^^^^^01f32d} {hot-dog} {hotdog} {Hot~dog} {1.0}
\__emoji_def:nnnnn {^^^^^^01f96a} {sandwich} {} {Sandwich} {5.0}
\__emoji_def:nnnnn {^^^^^^01f32e} {taco} {} {Taco} {1.0}
\__emoji_def:nnnnn {^^^^^^01f32f} {burrito} {} {Burrito} {1.0}
\__emoji_def:nnnnn {^^^^^^01fad4} {tamale} {} {Tamale} {13.0}
\__emoji_def:nnnnn {^^^^^^01f959} {stuffed-flatbread} {} {Stuffed~flatbread} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9c6} {falafel} {} {Falafel} {12.0}
\__emoji_def:nnnnn {^^^^^^01f95a} {egg} {} {Egg} {3.0}
\__emoji_def:nnnnn {^^^^^^01f373} {cooking} {fried-egg} {Cooking} {0.6}
\__emoji_def:nnnnn {^^^^^^01f958} {shallow-pan-of-food} {} {Shallow~pan~of~food} {3.0}
\__emoji_def:nnnnn {^^^^^^01f372} {pot-of-food} {stew} {Pot~of~food} {0.6}
\__emoji_def:nnnnn {^^^^^^01fad5} {fondue} {} {Fondue} {13.0}
\__emoji_def:nnnnn {^^^^^^01f963} {bowl-with-spoon} {} {Bowl~with~spoon} {5.0}
\__emoji_def:nnnnn {^^^^^^01f957} {green-salad} {} {Green~salad} {3.0}
\__emoji_def:nnnnn {^^^^^^01f37f} {popcorn} {} {Popcorn} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9c8} {butter} {} {Butter} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9c2} {salt} {} {Salt} {11.0}
\__emoji_def:nnnnn {^^^^^^01f96b} {canned-food} {} {Canned~food} {5.0}
\__emoji_subgroup:n {Food-asian}
\__emoji_def:nnnnn {^^^^^^01f371} {bento-box} {bento} {Bento~box} {0.6}
\__emoji_def:nnnnn {^^^^^^01f358} {rice-cracker} {} {Rice~cracker} {0.6}
\__emoji_def:nnnnn {^^^^^^01f359} {rice-ball} {} {Rice~ball} {0.6}
\__emoji_def:nnnnn {^^^^^^01f35a} {cooked-rice} {rice} {Cooked~rice} {0.6}
\__emoji_def:nnnnn {^^^^^^01f35b} {curry-rice} {curry} {Curry~rice} {0.6}
\__emoji_def:nnnnn {^^^^^^01f35c} {steaming-bowl} {ramen} {Steaming~bowl} {0.6}
\__emoji_def:nnnnn {^^^^^^01f35d} {spaghetti} {} {Spaghetti} {0.6}
\__emoji_def:nnnnn {^^^^^^01f360} {roasted-sweet-potato} {sweet-potato} {Roasted~sweet~potato} {0.6}
\__emoji_def:nnnnn {^^^^^^01f362} {oden} {} {Oden} {0.6}
\__emoji_def:nnnnn {^^^^^^01f363} {sushi} {} {Sushi} {0.6}
\__emoji_def:nnnnn {^^^^^^01f364} {fried-shrimp} {} {Fried~shrimp} {0.6}
\__emoji_def:nnnnn {^^^^^^01f365} {fish-cake-with-swirl} {fish-cake} {Fish~cake~with~swirl} {0.6}
\__emoji_def:nnnnn {^^^^^^01f96e} {moon-cake} {} {Moon~cake} {11.0}
\__emoji_def:nnnnn {^^^^^^01f361} {dango} {} {Dango} {0.6}
\__emoji_def:nnnnn {^^^^^^01f95f} {dumpling} {} {Dumpling} {5.0}
\__emoji_def:nnnnn {^^^^^^01f960} {fortune-cookie} {} {Fortune~cookie} {5.0}
\__emoji_def:nnnnn {^^^^^^01f961} {takeout-box} {} {Takeout~box} {5.0}
\__emoji_subgroup:n {Food-marine}
\__emoji_def:nnnnn {^^^^^^01f980} {crab} {} {Crab} {1.0}
\__emoji_def:nnnnn {^^^^^^01f99e} {lobster} {} {Lobster} {11.0}
\__emoji_def:nnnnn {^^^^^^01f990} {shrimp} {} {Shrimp} {3.0}
\__emoji_def:nnnnn {^^^^^^01f991} {squid} {} {Squid} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9aa} {oyster} {} {Oyster} {12.0}
\__emoji_subgroup:n {Food-sweet}
\__emoji_def:nnnnn {^^^^^^01f366} {soft-ice-cream} {icecream} {Soft~ice~cream} {0.6}
\__emoji_def:nnnnn {^^^^^^01f367} {shaved-ice} {} {Shaved~ice} {0.6}
\__emoji_def:nnnnn {^^^^^^01f368} {ice-cream} {} {Ice~cream} {0.6}
\__emoji_def:nnnnn {^^^^^^01f369} {doughnut} {} {Doughnut} {0.6}
\__emoji_def:nnnnn {^^^^^^01f36a} {cookie} {} {Cookie} {0.6}
\__emoji_def:nnnnn {^^^^^^01f382} {birthday-cake} {birthday} {Birthday~cake} {0.6}
\__emoji_def:nnnnn {^^^^^^01f370} {shortcake} {cake} {Shortcake} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9c1} {cupcake} {} {Cupcake} {11.0}
\__emoji_def:nnnnn {^^^^^^01f967} {pie} {} {Pie} {5.0}
\__emoji_def:nnnnn {^^^^^^01f36b} {chocolate-bar} {} {Chocolate~bar} {0.6}
\__emoji_def:nnnnn {^^^^^^01f36c} {candy} {} {Candy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f36d} {lollipop} {} {Lollipop} {0.6}
\__emoji_def:nnnnn {^^^^^^01f36e} {custard} {} {Custard} {0.6}
\__emoji_def:nnnnn {^^^^^^01f36f} {honey-pot} {} {Honey~pot} {0.6}
\__emoji_subgroup:n {Drink}
\__emoji_def:nnnnn {^^^^^^01f37c} {baby-bottle} {} {Baby~bottle} {1.0}
\__emoji_def:nnnnn {^^^^^^01f95b} {glass-of-milk} {milk-glass} {Glass~of~milk} {3.0}
\__emoji_def:nnnnn {^^^^2615} {hot-beverage} {coffee} {Hot~beverage} {0.6}
\__emoji_def:nnnnn {^^^^^^01fad6} {teapot} {} {Teapot} {13.0}
\__emoji_def:nnnnn {^^^^^^01f375} {teacup-without-handle} {tea} {Teacup~without~handle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f376} {sake} {} {Sake} {0.6}
\__emoji_def:nnnnn {^^^^^^01f37e} {bottle-with-popping-cork} {champagne} {Bottle~with~popping~cork} {1.0}
\__emoji_def:nnnnn {^^^^^^01f377} {wine-glass} {} {Wine~glass} {0.6}
\__emoji_def:nnnnn {^^^^^^01f378} {cocktail-glass} {cocktail} {Cocktail~glass} {0.6}
\__emoji_def:nnnnn {^^^^^^01f379} {tropical-drink} {} {Tropical~drink} {0.6}
\__emoji_def:nnnnn {^^^^^^01f37a} {beer-mug} {beer} {Beer~mug} {0.6}
\__emoji_def:nnnnn {^^^^^^01f37b} {clinking-beer-mugs} {beers} {Clinking~beer~mugs} {0.6}
\__emoji_def:nnnnn {^^^^^^01f942} {clinking-glasses} {} {Clinking~glasses} {3.0}
\__emoji_def:nnnnn {^^^^^^01f943} {tumbler-glass} {} {Tumbler~glass} {3.0}
\__emoji_def:nnnnn {^^^^^^01f964} {cup-with-straw} {} {Cup~with~straw} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9cb} {bubble-tea} {} {Bubble~tea} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9c3} {beverage-box} {} {Beverage~box} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9c9} {mate} {} {Mate} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9ca} {ice} {ice-cube} {Ice} {12.0}
\__emoji_subgroup:n {Dishware}
\__emoji_def:nnnnn {^^^^^^01f962} {chopsticks} {} {Chopsticks} {5.0}
\__emoji_def:nnnnn {^^^^^^01f37d^^^^fe0f} {fork-and-knife-with-plate} {plate-with-cutlery} {Fork~and~knife~with~plate} {0.7}
\__emoji_def:nnnnn {^^^^^^01f374} {fork-and-knife} {} {Fork~and~knife} {0.6}
\__emoji_def:nnnnn {^^^^^^01f944} {spoon} {} {Spoon} {3.0}
\__emoji_def:nnnnn {^^^^^^01f52a} {kitchen-knife} {hocho, knife} {Kitchen~knife} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3fa} {amphora} {} {Amphora} {1.0}
\__emoji_group:n {Travel~\&~Places}
\__emoji_subgroup:n {Place-map}
\__emoji_def:nnnnn {^^^^^^01f30d} {globe-showing-europe-africa} {earth-africa} {Globe~showing~Europe-Africa} {0.7}
\__emoji_def:nnnnn {^^^^^^01f30e} {globe-showing-americas} {earth-americas} {Globe~showing~Americas} {0.7}
\__emoji_def:nnnnn {^^^^^^01f30f} {globe-showing-asia-australia} {earth-asia} {Globe~showing~Asia-Australia} {0.6}
\__emoji_def:nnnnn {^^^^^^01f310} {globe-with-meridians} {} {Globe~with~meridians} {1.0}
\__emoji_def:nnnnn {^^^^^^01f5fa^^^^fe0f} {world-map} {} {World~map} {0.7}
\__emoji_def:nnnnn {^^^^^^01f5fe} {map-of-japan} {japan} {Map~of~Japan} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9ed} {compass} {} {Compass} {11.0}
\__emoji_subgroup:n {Place-geographic}
\__emoji_def:nnnnn {^^^^^^01f3d4^^^^fe0f} {snow-capped-mountain} {mountain-snow} {Snow-capped~mountain} {0.7}
\__emoji_def:nnnnn {^^^^26f0^^^^fe0f} {mountain} {} {Mountain} {0.7}
\__emoji_def:nnnnn {^^^^^^01f30b} {volcano} {} {Volcano} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5fb} {mount-fuji} {} {Mount~fuji} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3d5^^^^fe0f} {camping} {} {Camping} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3d6^^^^fe0f} {beach-with-umbrella} {beach-umbrella} {Beach~with~umbrella} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3dc^^^^fe0f} {desert} {} {Desert} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3dd^^^^fe0f} {desert-island} {} {Desert~island} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3de^^^^fe0f} {national-park} {} {National~park} {0.7}
\__emoji_subgroup:n {Place-building}
\__emoji_def:nnnnn {^^^^^^01f3df^^^^fe0f} {stadium} {} {Stadium} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3db^^^^fe0f} {classical-building} {} {Classical~building} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3d7^^^^fe0f} {building-construction} {} {Building~construction} {0.7}
\__emoji_def:nnnnn {^^^^^^01f9f1} {brick} {bricks} {Brick} {11.0}
\__emoji_def:nnnnn {^^^^^^01faa8} {rock} {} {Rock} {13.0}
\__emoji_def:nnnnn {^^^^^^01fab5} {wood} {} {Wood} {13.0}
\__emoji_def:nnnnn {^^^^^^01f6d6} {hut} {} {Hut} {13.0}
\__emoji_def:nnnnn {^^^^^^01f3d8^^^^fe0f} {houses} {} {Houses} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3da^^^^fe0f} {derelict-house} {} {Derelict~house} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3e0} {house} {} {House} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3e1} {house-with-garden} {} {House~with~garden} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3e2} {office-building} {office} {Office~building} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3e3} {japanese-post-office} {} {Japanese~post~office} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3e4} {post-office} {european-post-office} {Post~office} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3e5} {hospital} {} {Hospital} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3e6} {bank} {} {Bank} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3e8} {hotel} {} {Hotel} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3e9} {love-hotel} {} {Love~hotel} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3ea} {convenience-store} {} {Convenience~store} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3eb} {school} {} {School} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3ec} {department-store} {} {Department~store} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3ed} {factory} {} {Factory} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3ef} {japanese-castle} {} {Japanese~castle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3f0} {castle} {european-castle} {Castle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f492} {wedding} {} {Wedding} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5fc} {tokyo-tower} {} {Tokyo~tower} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5fd} {statue-of-liberty} {} {Statue~of~Liberty} {0.6}
\__emoji_subgroup:n {Place-religious}
\__emoji_def:nnnnn {^^^^26ea} {church} {} {Church} {0.6}
\__emoji_def:nnnnn {^^^^^^01f54c} {mosque} {} {Mosque} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6d5} {hindu-temple} {} {Hindu~temple} {12.0}
\__emoji_def:nnnnn {^^^^^^01f54d} {synagogue} {} {Synagogue} {1.0}
\__emoji_def:nnnnn {^^^^26e9^^^^fe0f} {shinto-shrine} {} {Shinto~shrine} {0.7}
\__emoji_def:nnnnn {^^^^^^01f54b} {kaaba} {} {Kaaba} {1.0}
\__emoji_subgroup:n {Place-other}
\__emoji_def:nnnnn {^^^^26f2} {fountain} {} {Fountain} {0.6}
\__emoji_def:nnnnn {^^^^26fa} {tent} {} {Tent} {0.6}
\__emoji_def:nnnnn {^^^^^^01f301} {foggy} {} {Foggy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f303} {night-with-stars} {} {Night~with~stars} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3d9^^^^fe0f} {cityscape} {} {Cityscape} {0.7}
\__emoji_def:nnnnn {^^^^^^01f304} {sunrise-over-mountains} {} {Sunrise~over~mountains} {0.6}
\__emoji_def:nnnnn {^^^^^^01f305} {sunrise} {} {Sunrise} {0.6}
\__emoji_def:nnnnn {^^^^^^01f306} {cityscape-at-dusk} {city-sunset} {Cityscape~at~dusk} {0.6}
\__emoji_def:nnnnn {^^^^^^01f307} {sunset} {city-sunrise} {Sunset} {0.6}
\__emoji_def:nnnnn {^^^^^^01f309} {bridge-at-night} {} {Bridge~at~night} {0.6}
\__emoji_def:nnnnn {^^^^2668^^^^fe0f} {hot-springs} {hotsprings} {Hot~springs} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3a0} {carousel-horse} {} {Carousel~horse} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3a1} {ferris-wheel} {} {Ferris~wheel} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3a2} {roller-coaster} {} {Roller~coaster} {0.6}
\__emoji_def:nnnnn {^^^^^^01f488} {barber-pole} {barber} {Barber~pole} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3aa} {circus-tent} {} {Circus~tent} {0.6}
\__emoji_subgroup:n {Transport-ground}
\__emoji_def:nnnnn {^^^^^^01f682} {locomotive} {steam-locomotive} {Locomotive} {1.0}
\__emoji_def:nnnnn {^^^^^^01f683} {railway-car} {} {Railway~car} {0.6}
\__emoji_def:nnnnn {^^^^^^01f684} {high-speed-train} {bullettrain-side} {High-speed~train} {0.6}
\__emoji_def:nnnnn {^^^^^^01f685} {bullet-train} {bullettrain-front} {Bullet~train} {0.6}
\__emoji_def:nnnnn {^^^^^^01f686} {train} {train2} {Train} {1.0}
\__emoji_def:nnnnn {^^^^^^01f687} {metro} {} {Metro} {0.6}
\__emoji_def:nnnnn {^^^^^^01f688} {light-rail} {} {Light~rail} {1.0}
\__emoji_def:nnnnn {^^^^^^01f689} {station} {} {Station} {0.6}
\__emoji_def:nnnnn {^^^^^^01f68a} {tram} {} {Tram} {1.0}
\__emoji_def:nnnnn {^^^^^^01f69d} {monorail} {} {Monorail} {1.0}
\__emoji_def:nnnnn {^^^^^^01f69e} {mountain-railway} {} {Mountain~railway} {1.0}
\__emoji_def:nnnnn {^^^^^^01f68b} {tram-car} {} {Tram~car} {1.0}
\__emoji_def:nnnnn {^^^^^^01f68c} {bus} {} {Bus} {0.6}
\__emoji_def:nnnnn {^^^^^^01f68d} {oncoming-bus} {} {Oncoming~bus} {0.7}
\__emoji_def:nnnnn {^^^^^^01f68e} {trolleybus} {} {Trolleybus} {1.0}
\__emoji_def:nnnnn {^^^^^^01f690} {minibus} {} {Minibus} {1.0}
\__emoji_def:nnnnn {^^^^^^01f691} {ambulance} {} {Ambulance} {0.6}
\__emoji_def:nnnnn {^^^^^^01f692} {fire-engine} {} {Fire~engine} {0.6}
\__emoji_def:nnnnn {^^^^^^01f693} {police-car} {} {Police~car} {0.6}
\__emoji_def:nnnnn {^^^^^^01f694} {oncoming-police-car} {} {Oncoming~police~car} {0.7}
\__emoji_def:nnnnn {^^^^^^01f695} {taxi} {} {Taxi} {0.6}
\__emoji_def:nnnnn {^^^^^^01f696} {oncoming-taxi} {} {Oncoming~taxi} {1.0}
\__emoji_def:nnnnn {^^^^^^01f697} {automobile} {car, red-car} {Automobile} {0.6}
\__emoji_def:nnnnn {^^^^^^01f698} {oncoming-automobile} {} {Oncoming~automobile} {0.7}
\__emoji_def:nnnnn {^^^^^^01f699} {sport-utility-vehicle} {blue-car} {Sport~utility~vehicle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6fb} {pickup-truck} {} {Pickup~truck} {13.0}
\__emoji_def:nnnnn {^^^^^^01f69a} {delivery-truck} {truck} {Delivery~truck} {0.6}
\__emoji_def:nnnnn {^^^^^^01f69b} {articulated-lorry} {} {Articulated~lorry} {1.0}
\__emoji_def:nnnnn {^^^^^^01f69c} {tractor} {} {Tractor} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3ce^^^^fe0f} {racing-car} {} {Racing~car} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3cd^^^^fe0f} {motorcycle} {} {Motorcycle} {0.7}
\__emoji_def:nnnnn {^^^^^^01f6f5} {motor-scooter} {} {Motor~scooter} {3.0}
\__emoji_def:nnnnn {^^^^^^01f9bd} {manual-wheelchair} {} {Manual~wheelchair} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9bc} {motorized-wheelchair} {} {Motorized~wheelchair} {12.0}
\__emoji_def:nnnnn {^^^^^^01f6fa} {auto-rickshaw} {} {Auto~rickshaw} {12.0}
\__emoji_def:nnnnn {^^^^^^01f6b2} {bicycle} {bike} {Bicycle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6f4} {kick-scooter} {} {Kick~scooter} {3.0}
\__emoji_def:nnnnn {^^^^^^01f6f9} {skateboard} {} {Skateboard} {11.0}
\__emoji_def:nnnnn {^^^^^^01f6fc} {roller-skate} {} {Roller~skate} {13.0}
\__emoji_def:nnnnn {^^^^^^01f68f} {bus-stop} {busstop} {Bus~stop} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6e3^^^^fe0f} {motorway} {} {Motorway} {0.7}
\__emoji_def:nnnnn {^^^^^^01f6e4^^^^fe0f} {railway-track} {} {Railway~track} {0.7}
\__emoji_def:nnnnn {^^^^^^01f6e2^^^^fe0f} {oil-drum} {} {Oil~drum} {0.7}
\__emoji_def:nnnnn {^^^^26fd} {fuel-pump} {fuelpump} {Fuel~pump} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6a8} {police-car-light} {rotating-light} {Police~car~light} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6a5} {horizontal-traffic-light} {traffic-light} {Horizontal~traffic~light} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6a6} {vertical-traffic-light} {} {Vertical~traffic~light} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6d1} {stop-sign} {} {Stop~sign} {3.0}
\__emoji_def:nnnnn {^^^^^^01f6a7} {construction} {} {Construction} {0.6}
\__emoji_subgroup:n {Transport-water}
\__emoji_def:nnnnn {^^^^2693} {anchor} {} {Anchor} {0.6}
\__emoji_def:nnnnn {^^^^26f5} {sailboat} {boat} {Sailboat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6f6} {canoe} {} {Canoe} {3.0}
\__emoji_def:nnnnn {^^^^^^01f6a4} {speedboat} {} {Speedboat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6f3^^^^fe0f} {passenger-ship} {} {Passenger~ship} {0.7}
\__emoji_def:nnnnn {^^^^26f4^^^^fe0f} {ferry} {} {Ferry} {0.7}
\__emoji_def:nnnnn {^^^^^^01f6e5^^^^fe0f} {motor-boat} {} {Motor~boat} {0.7}
\__emoji_def:nnnnn {^^^^^^01f6a2} {ship} {} {Ship} {0.6}
\__emoji_subgroup:n {Transport-air}
\__emoji_def:nnnnn {^^^^2708^^^^fe0f} {airplane} {} {Airplane} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6e9^^^^fe0f} {small-airplane} {} {Small~airplane} {0.7}
\__emoji_def:nnnnn {^^^^^^01f6eb} {airplane-departure} {flight-departure} {Airplane~departure} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6ec} {airplane-arrival} {flight-arrival} {Airplane~arrival} {1.0}
\__emoji_def:nnnnn {^^^^^^01fa82} {parachute} {} {Parachute} {12.0}
\__emoji_def:nnnnn {^^^^^^01f4ba} {seat} {} {Seat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f681} {helicopter} {} {Helicopter} {1.0}
\__emoji_def:nnnnn {^^^^^^01f69f} {suspension-railway} {} {Suspension~railway} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a0} {mountain-cableway} {} {Mountain~cableway} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6a1} {aerial-tramway} {} {Aerial~tramway} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6f0^^^^fe0f} {satellite} {artificial-satellite} {Satellite} {0.7}
\__emoji_def:nnnnn {^^^^^^01f680} {rocket} {} {Rocket} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6f8} {flying-saucer} {} {Flying~saucer} {5.0}
\__emoji_subgroup:n {Hotel}
\__emoji_def:nnnnn {^^^^^^01f6ce^^^^fe0f} {bellhop-bell} {} {Bellhop~bell} {0.7}
\__emoji_def:nnnnn {^^^^^^01f9f3} {luggage} {} {Luggage} {11.0}
\__emoji_subgroup:n {Time}
\__emoji_def:nnnnn {^^^^231b} {hourglass-done} {hourglass} {Hourglass~done} {0.6}
\__emoji_def:nnnnn {^^^^23f3} {hourglass-not-done} {hourglass-flowing-sand} {Hourglass~not~done} {0.6}
\__emoji_def:nnnnn {^^^^231a} {watch} {} {Watch} {0.6}
\__emoji_def:nnnnn {^^^^23f0} {alarm-clock} {} {Alarm~clock} {0.6}
\__emoji_def:nnnnn {^^^^23f1^^^^fe0f} {stopwatch} {} {Stopwatch} {1.0}
\__emoji_def:nnnnn {^^^^23f2^^^^fe0f} {timer-clock} {} {Timer~clock} {1.0}
\__emoji_def:nnnnn {^^^^^^01f570^^^^fe0f} {mantelpiece-clock} {} {Mantelpiece~clock} {0.7}
\__emoji_def:nnnnn {^^^^^^01f55b} {twelve-oclock} {clock12} {Twelve~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f567} {twelve-thirty} {clock1230} {Twelve-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f550} {one-oclock} {clock1} {One~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f55c} {one-thirty} {clock130} {One-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f551} {two-oclock} {clock2} {Two~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f55d} {two-thirty} {clock230} {Two-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f552} {three-oclock} {clock3} {Three~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f55e} {three-thirty} {clock330} {Three-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f553} {four-oclock} {clock4} {Four~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f55f} {four-thirty} {clock430} {Four-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f554} {five-oclock} {clock5} {Five~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f560} {five-thirty} {clock530} {Five-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f555} {six-oclock} {clock6} {Six~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f561} {six-thirty} {clock630} {Six-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f556} {seven-oclock} {clock7} {Seven~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f562} {seven-thirty} {clock730} {Seven-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f557} {eight-oclock} {clock8} {Eight~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f563} {eight-thirty} {clock830} {Eight-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f558} {nine-oclock} {clock9} {Nine~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f564} {nine-thirty} {clock930} {Nine-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f559} {ten-oclock} {clock10} {Ten~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f565} {ten-thirty} {clock1030} {Ten-thirty} {0.7}
\__emoji_def:nnnnn {^^^^^^01f55a} {eleven-oclock} {clock11} {Eleven~o'clock} {0.6}
\__emoji_def:nnnnn {^^^^^^01f566} {eleven-thirty} {clock1130} {Eleven-thirty} {0.7}
\__emoji_subgroup:n {Sky~\&~weather}
\__emoji_def:nnnnn {^^^^^^01f311} {new-moon} {} {New~moon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f312} {waxing-crescent-moon} {} {Waxing~crescent~moon} {1.0}
\__emoji_def:nnnnn {^^^^^^01f313} {first-quarter-moon} {} {First~quarter~moon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f314} {waxing-gibbous-moon} {moon} {Waxing~gibbous~moon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f315} {full-moon} {} {Full~moon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f316} {waning-gibbous-moon} {} {Waning~gibbous~moon} {1.0}
\__emoji_def:nnnnn {^^^^^^01f317} {last-quarter-moon} {} {Last~quarter~moon} {1.0}
\__emoji_def:nnnnn {^^^^^^01f318} {waning-crescent-moon} {} {Waning~crescent~moon} {1.0}
\__emoji_def:nnnnn {^^^^^^01f319} {crescent-moon} {} {Crescent~moon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f31a} {new-moon-face} {new-moon-with-face} {New~moon~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f31b} {first-quarter-moon-face} {first-quarter-moon-with-face} {First~quarter~moon~face} {0.6}
\__emoji_def:nnnnn {^^^^^^01f31c} {last-quarter-moon-face} {last-quarter-moon-with-face} {Last~quarter~moon~face} {0.7}
\__emoji_def:nnnnn {^^^^^^01f321^^^^fe0f} {thermometer} {} {Thermometer} {0.7}
\__emoji_def:nnnnn {^^^^2600^^^^fe0f} {sun} {sunny} {Sun} {0.6}
\__emoji_def:nnnnn {^^^^^^01f31d} {full-moon-face} {full-moon-with-face} {Full~moon~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01f31e} {sun-with-face} {} {Sun~with~face} {1.0}
\__emoji_def:nnnnn {^^^^^^01fa90} {ringed-planet} {} {Ringed~planet} {12.0}
\__emoji_def:nnnnn {^^^^2b50} {star} {} {Star} {0.6}
\__emoji_def:nnnnn {^^^^^^01f31f} {glowing-star} {star2} {Glowing~star} {0.6}
\__emoji_def:nnnnn {^^^^^^01f320} {shooting-star} {stars} {Shooting~star} {0.6}
\__emoji_def:nnnnn {^^^^^^01f30c} {milky-way} {} {Milky~way} {0.6}
\__emoji_def:nnnnn {^^^^2601^^^^fe0f} {cloud} {} {Cloud} {0.6}
\__emoji_def:nnnnn {^^^^26c5} {sun-behind-cloud} {partly-sunny} {Sun~behind~cloud} {0.6}
\__emoji_def:nnnnn {^^^^26c8^^^^fe0f} {cloud-with-lightning-and-rain} {} {Cloud~with~lightning~and~rain} {0.7}
\__emoji_def:nnnnn {^^^^^^01f324^^^^fe0f} {sun-behind-small-cloud} {} {Sun~behind~small~cloud} {0.7}
\__emoji_def:nnnnn {^^^^^^01f325^^^^fe0f} {sun-behind-large-cloud} {} {Sun~behind~large~cloud} {0.7}
\__emoji_def:nnnnn {^^^^^^01f326^^^^fe0f} {sun-behind-rain-cloud} {} {Sun~behind~rain~cloud} {0.7}
\__emoji_def:nnnnn {^^^^^^01f327^^^^fe0f} {cloud-with-rain} {} {Cloud~with~rain} {0.7}
\__emoji_def:nnnnn {^^^^^^01f328^^^^fe0f} {cloud-with-snow} {} {Cloud~with~snow} {0.7}
\__emoji_def:nnnnn {^^^^^^01f329^^^^fe0f} {cloud-with-lightning} {} {Cloud~with~lightning} {0.7}
\__emoji_def:nnnnn {^^^^^^01f32a^^^^fe0f} {tornado} {} {Tornado} {0.7}
\__emoji_def:nnnnn {^^^^^^01f32b^^^^fe0f} {fog} {} {Fog} {0.7}
\__emoji_def:nnnnn {^^^^^^01f32c^^^^fe0f} {wind-face} {} {Wind~face} {0.7}
\__emoji_def:nnnnn {^^^^^^01f300} {cyclone} {} {Cyclone} {0.6}
\__emoji_def:nnnnn {^^^^^^01f308} {rainbow} {} {Rainbow} {0.6}
\__emoji_def:nnnnn {^^^^^^01f302} {closed-umbrella} {} {Closed~umbrella} {0.6}
\__emoji_def:nnnnn {^^^^2602^^^^fe0f} {umbrella} {open-umbrella} {Umbrella} {0.7}
\__emoji_def:nnnnn {^^^^2614} {umbrella-with-rain-drops} {} {Umbrella~with~rain~drops} {0.6}
\__emoji_def:nnnnn {^^^^26f1^^^^fe0f} {umbrella-on-ground} {parasol-on-ground} {Umbrella~on~ground} {0.7}
\__emoji_def:nnnnn {^^^^26a1} {high-voltage} {zap} {High~voltage} {0.6}
\__emoji_def:nnnnn {^^^^2744^^^^fe0f} {snowflake} {} {Snowflake} {0.6}
\__emoji_def:nnnnn {^^^^2603^^^^fe0f} {snowman} {snowman-with-snow} {Snowman} {0.7}
\__emoji_def:nnnnn {^^^^26c4} {snowman-without-snow} {} {Snowman~without~snow} {0.6}
\__emoji_def:nnnnn {^^^^2604^^^^fe0f} {comet} {} {Comet} {1.0}
\__emoji_def:nnnnn {^^^^^^01f525} {fire} {} {Fire} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4a7} {droplet} {} {Droplet} {0.6}
\__emoji_def:nnnnn {^^^^^^01f30a} {water-wave} {ocean} {Water~wave} {0.6}
\__emoji_group:n {Activities}
\__emoji_subgroup:n {Event}
\__emoji_def:nnnnn {^^^^^^01f383} {jack-o-lantern} {} {Jack-o-lantern} {0.6}
\__emoji_def:nnnnn {^^^^^^01f384} {christmas-tree} {} {Christmas~tree} {0.6}
\__emoji_def:nnnnn {^^^^^^01f386} {fireworks} {} {Fireworks} {0.6}
\__emoji_def:nnnnn {^^^^^^01f387} {sparkler} {} {Sparkler} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9e8} {firecracker} {} {Firecracker} {11.0}
\__emoji_def:nnnnn {^^^^2728} {sparkles} {} {Sparkles} {0.6}
\__emoji_def:nnnnn {^^^^^^01f388} {balloon} {} {Balloon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f389} {party-popper} {tada} {Party~popper} {0.6}
\__emoji_def:nnnnn {^^^^^^01f38a} {confetti-ball} {} {Confetti~ball} {0.6}
\__emoji_def:nnnnn {^^^^^^01f38b} {tanabata-tree} {} {Tanabata~tree} {0.6}
\__emoji_def:nnnnn {^^^^^^01f38d} {pine-decoration} {bamboo} {Pine~decoration} {0.6}
\__emoji_def:nnnnn {^^^^^^01f38e} {japanese-dolls} {dolls} {Japanese~dolls} {0.6}
\__emoji_def:nnnnn {^^^^^^01f38f} {carp-streamer} {flags} {Carp~streamer} {0.6}
\__emoji_def:nnnnn {^^^^^^01f390} {wind-chime} {} {Wind~chime} {0.6}
\__emoji_def:nnnnn {^^^^^^01f391} {moon-viewing-ceremony} {rice-scene} {Moon~viewing~ceremony} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9e7} {red-envelope} {} {Red~envelope} {11.0}
\__emoji_def:nnnnn {^^^^^^01f380} {ribbon} {} {Ribbon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f381} {wrapped-gift} {gift} {Wrapped~gift} {0.6}
\__emoji_def:nnnnn {^^^^^^01f397^^^^fe0f} {reminder-ribbon} {} {Reminder~ribbon} {0.7}
\__emoji_def:nnnnn {^^^^^^01f39f^^^^fe0f} {admission-tickets} {tickets} {Admission~tickets} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3ab} {ticket} {} {Ticket} {0.6}
\__emoji_subgroup:n {Award-medal}
\__emoji_def:nnnnn {^^^^^^01f396^^^^fe0f} {military-medal} {medal-military} {Military~medal} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3c6} {trophy} {} {Trophy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3c5} {sports-medal} {medal-sports} {Sports~medal} {1.0}
\__emoji_def:nnnnn {^^^^^^01f947} {1st-place-medal} {} {1st~place~medal} {3.0}
\__emoji_def:nnnnn {^^^^^^01f948} {2nd-place-medal} {} {2nd~place~medal} {3.0}
\__emoji_def:nnnnn {^^^^^^01f949} {3rd-place-medal} {} {3rd~place~medal} {3.0}
\__emoji_subgroup:n {Sport}
\__emoji_def:nnnnn {^^^^26bd} {soccer-ball} {soccer} {Soccer~ball} {0.6}
\__emoji_def:nnnnn {^^^^26be} {baseball} {} {Baseball} {0.6}
\__emoji_def:nnnnn {^^^^^^01f94e} {softball} {} {Softball} {11.0}
\__emoji_def:nnnnn {^^^^^^01f3c0} {basketball} {} {Basketball} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3d0} {volleyball} {} {Volleyball} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3c8} {american-football} {football} {American~football} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3c9} {rugby-football} {} {Rugby~football} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3be} {tennis} {} {Tennis} {0.6}
\__emoji_def:nnnnn {^^^^^^01f94f} {flying-disc} {} {Flying~disc} {11.0}
\__emoji_def:nnnnn {^^^^^^01f3b3} {bowling} {} {Bowling} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3cf} {cricket-game} {} {Cricket~game} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3d1} {field-hockey} {} {Field~hockey} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3d2} {ice-hockey} {} {Ice~hockey} {1.0}
\__emoji_def:nnnnn {^^^^^^01f94d} {lacrosse} {} {Lacrosse} {11.0}
\__emoji_def:nnnnn {^^^^^^01f3d3} {ping-pong} {} {Ping~pong} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3f8} {badminton} {} {Badminton} {1.0}
\__emoji_def:nnnnn {^^^^^^01f94a} {boxing-glove} {} {Boxing~glove} {3.0}
\__emoji_def:nnnnn {^^^^^^01f94b} {martial-arts-uniform} {} {Martial~arts~uniform} {3.0}
\__emoji_def:nnnnn {^^^^^^01f945} {goal-net} {} {Goal~net} {3.0}
\__emoji_def:nnnnn {^^^^26f3} {flag-in-hole} {golf} {Flag~in~hole} {0.6}
\__emoji_def:nnnnn {^^^^26f8^^^^fe0f} {ice-skate} {} {Ice~skate} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3a3} {fishing-pole} {fishing-pole-and-fish} {Fishing~pole} {0.6}
\__emoji_def:nnnnn {^^^^^^01f93f} {diving-mask} {} {Diving~mask} {12.0}
\__emoji_def:nnnnn {^^^^^^01f3bd} {running-shirt} {running-shirt-with-sash} {Running~shirt} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3bf} {skis} {ski} {Skis} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6f7} {sled} {} {Sled} {5.0}
\__emoji_def:nnnnn {^^^^^^01f94c} {curling-stone} {} {Curling~stone} {5.0}
\__emoji_subgroup:n {Game}
\__emoji_def:nnnnn {^^^^^^01f3af} {bullseye} {dart} {Bullseye} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa80} {yo-yo} {} {Yo-yo} {12.0}
\__emoji_def:nnnnn {^^^^^^01fa81} {kite} {} {Kite} {12.0}
\__emoji_def:nnnnn {^^^^^^01f3b1} {pool-8-ball} {8ball} {Pool~8~ball} {0.6}
\__emoji_def:nnnnn {^^^^^^01f52e} {crystal-ball} {} {Crystal~ball} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa84} {magic-wand} {} {Magic~wand} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9ff} {nazar-amulet} {} {Nazar~amulet} {11.0}
\__emoji_def:nnnnn {^^^^^^01f3ae} {video-game} {} {Video~game} {0.6}
\__emoji_def:nnnnn {^^^^^^01f579^^^^fe0f} {joystick} {} {Joystick} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3b0} {slot-machine} {} {Slot~machine} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3b2} {game-die} {} {Game~die} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9e9} {puzzle-piece} {jigsaw} {Puzzle~piece} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9f8} {teddy-bear} {} {Teddy~bear} {11.0}
\__emoji_def:nnnnn {^^^^^^01fa85} {pinata} {} {Piñata} {13.0}
\__emoji_def:nnnnn {^^^^^^01fa86} {nesting-dolls} {} {Nesting~dolls} {13.0}
\__emoji_def:nnnnn {^^^^2660^^^^fe0f} {spade-suit} {spades} {Spade~suit} {0.6}
\__emoji_def:nnnnn {^^^^2665^^^^fe0f} {heart-suit} {hearts} {Heart~suit} {0.6}
\__emoji_def:nnnnn {^^^^2666^^^^fe0f} {diamond-suit} {diamonds} {Diamond~suit} {0.6}
\__emoji_def:nnnnn {^^^^2663^^^^fe0f} {club-suit} {clubs} {Club~suit} {0.6}
\__emoji_def:nnnnn {^^^^265f^^^^fe0f} {chess-pawn} {} {Chess~pawn} {11.0}
\__emoji_def:nnnnn {^^^^^^01f0cf} {joker} {black-joker} {Joker} {0.6}
\__emoji_def:nnnnn {^^^^^^01f004} {mahjong-red-dragon} {mahjong} {Mahjong~red~dragon} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3b4} {flower-playing-cards} {} {Flower~playing~cards} {0.6}
\__emoji_subgroup:n {Arts~\&~crafts}
\__emoji_def:nnnnn {^^^^^^01f3ad} {performing-arts} {} {Performing~arts} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5bc^^^^fe0f} {framed-picture} {} {Framed~picture} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3a8} {artist-palette} {art} {Artist~palette} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9f5} {thread} {} {Thread} {11.0}
\__emoji_def:nnnnn {^^^^^^01faa1} {sewing-needle} {} {Sewing~needle} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9f6} {yarn} {} {Yarn} {11.0}
\__emoji_def:nnnnn {^^^^^^01faa2} {knot} {} {Knot} {13.0}
\__emoji_group:n {Objects}
\__emoji_subgroup:n {Clothing}
\__emoji_def:nnnnn {^^^^^^01f453} {glasses} {eyeglasses} {Glasses} {0.6}
\__emoji_def:nnnnn {^^^^^^01f576^^^^fe0f} {sunglasses} {dark-sunglasses} {Sunglasses} {0.7}
\__emoji_def:nnnnn {^^^^^^01f97d} {goggles} {} {Goggles} {11.0}
\__emoji_def:nnnnn {^^^^^^01f97c} {lab-coat} {} {Lab~coat} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9ba} {safety-vest} {} {Safety~vest} {12.0}
\__emoji_def:nnnnn {^^^^^^01f454} {necktie} {} {Necktie} {0.6}
\__emoji_def:nnnnn {^^^^^^01f455} {t-shirt} {shirt, tshirt} {T-shirt} {0.6}
\__emoji_def:nnnnn {^^^^^^01f456} {jeans} {} {Jeans} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9e3} {scarf} {} {Scarf} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9e4} {gloves} {} {Gloves} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9e5} {coat} {} {Coat} {5.0}
\__emoji_def:nnnnn {^^^^^^01f9e6} {socks} {} {Socks} {5.0}
\__emoji_def:nnnnn {^^^^^^01f457} {dress} {} {Dress} {0.6}
\__emoji_def:nnnnn {^^^^^^01f458} {kimono} {} {Kimono} {0.6}
\__emoji_def:nnnnn {^^^^^^01f97b} {sari} {} {Sari} {12.0}
\__emoji_def:nnnnn {^^^^^^01fa71} {one-piece-swimsuit} {} {One-piece~swimsuit} {12.0}
\__emoji_def:nnnnn {^^^^^^01fa72} {briefs} {swim-brief} {Briefs} {12.0}
\__emoji_def:nnnnn {^^^^^^01fa73} {shorts} {} {Shorts} {12.0}
\__emoji_def:nnnnn {^^^^^^01f459} {bikini} {} {Bikini} {0.6}
\__emoji_def:nnnnn {^^^^^^01f45a} {womans-clothes} {} {Woman's~clothes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f45b} {purse} {} {Purse} {0.6}
\__emoji_def:nnnnn {^^^^^^01f45c} {handbag} {} {Handbag} {0.6}
\__emoji_def:nnnnn {^^^^^^01f45d} {clutch-bag} {pouch} {Clutch~bag} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6cd^^^^fe0f} {shopping-bags} {shopping} {Shopping~bags} {0.7}
\__emoji_def:nnnnn {^^^^^^01f392} {backpack} {school-satchel} {Backpack} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa74} {thong-sandal} {} {Thong~sandal} {13.0}
\__emoji_def:nnnnn {^^^^^^01f45e} {mans-shoe} {shoe} {Man's~shoe} {0.6}
\__emoji_def:nnnnn {^^^^^^01f45f} {running-shoe} {athletic-shoe} {Running~shoe} {0.6}
\__emoji_def:nnnnn {^^^^^^01f97e} {hiking-boot} {} {Hiking~boot} {11.0}
\__emoji_def:nnnnn {^^^^^^01f97f} {flat-shoe} {} {Flat~shoe} {11.0}
\__emoji_def:nnnnn {^^^^^^01f460} {high-heeled-shoe} {high-heel} {High-heeled~shoe} {0.6}
\__emoji_def:nnnnn {^^^^^^01f461} {womans-sandal} {sandal} {Woman's~sandal} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa70} {ballet-shoes} {} {Ballet~shoes} {12.0}
\__emoji_def:nnnnn {^^^^^^01f462} {womans-boot} {boot} {Woman's~boot} {0.6}
\__emoji_def:nnnnn {^^^^^^01f451} {crown} {} {Crown} {0.6}
\__emoji_def:nnnnn {^^^^^^01f452} {womans-hat} {} {Woman's~hat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3a9} {top-hat} {tophat} {Top~hat} {0.6}
\__emoji_def:nnnnn {^^^^^^01f393} {graduation-cap} {mortar-board} {Graduation~cap} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9e2} {billed-cap} {} {Billed~cap} {5.0}
\__emoji_def:nnnnn {^^^^^^01fa96} {military-helmet} {} {Military~helmet} {13.0}
\__emoji_def:nnnnn {^^^^26d1^^^^fe0f} {rescue-workers-helmet} {rescue-worker-helmet} {Rescue~worker's~helmet} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4ff} {prayer-beads} {} {Prayer~beads} {1.0}
\__emoji_def:nnnnn {^^^^^^01f484} {lipstick} {} {Lipstick} {0.6}
\__emoji_def:nnnnn {^^^^^^01f48d} {ring} {} {Ring} {0.6}
\__emoji_def:nnnnn {^^^^^^01f48e} {gem-stone} {gem} {Gem~stone} {0.6}
\__emoji_subgroup:n {Sound}
\__emoji_def:nnnnn {^^^^^^01f507} {muted-speaker} {mute} {Muted~speaker} {1.0}
\__emoji_def:nnnnn {^^^^^^01f508} {speaker-low-volume} {speaker} {Speaker~low~volume} {0.7}
\__emoji_def:nnnnn {^^^^^^01f509} {speaker-medium-volume} {sound} {Speaker~medium~volume} {1.0}
\__emoji_def:nnnnn {^^^^^^01f50a} {speaker-high-volume} {loud-sound} {Speaker~high~volume} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e2} {loudspeaker} {} {Loudspeaker} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e3} {megaphone} {mega} {Megaphone} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ef} {postal-horn} {} {Postal~horn} {1.0}
\__emoji_def:nnnnn {^^^^^^01f514} {bell} {} {Bell} {0.6}
\__emoji_def:nnnnn {^^^^^^01f515} {bell-with-slash} {no-bell} {Bell~with~slash} {1.0}
\__emoji_subgroup:n {Music}
\__emoji_def:nnnnn {^^^^^^01f3bc} {musical-score} {} {Musical~score} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3b5} {musical-note} {} {Musical~note} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3b6} {musical-notes} {notes} {Musical~notes} {0.6}
\__emoji_def:nnnnn {^^^^^^01f399^^^^fe0f} {studio-microphone} {} {Studio~microphone} {0.7}
\__emoji_def:nnnnn {^^^^^^01f39a^^^^fe0f} {level-slider} {} {Level~slider} {0.7}
\__emoji_def:nnnnn {^^^^^^01f39b^^^^fe0f} {control-knobs} {} {Control~knobs} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3a4} {microphone} {} {Microphone} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3a7} {headphone} {headphones} {Headphone} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4fb} {radio} {} {Radio} {0.6}
\__emoji_subgroup:n {Musical-instrument}
\__emoji_def:nnnnn {^^^^^^01f3b7} {saxophone} {} {Saxophone} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa97} {accordion} {} {Accordion} {13.0}
\__emoji_def:nnnnn {^^^^^^01f3b8} {guitar} {} {Guitar} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3b9} {musical-keyboard} {} {Musical~keyboard} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3ba} {trumpet} {} {Trumpet} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3bb} {violin} {} {Violin} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa95} {banjo} {} {Banjo} {12.0}
\__emoji_def:nnnnn {^^^^^^01f941} {drum} {} {Drum} {3.0}
\__emoji_def:nnnnn {^^^^^^01fa98} {long-drum} {} {Long~drum} {13.0}
\__emoji_subgroup:n {Phone}
\__emoji_def:nnnnn {^^^^^^01f4f1} {mobile-phone} {iphone} {Mobile~phone} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4f2} {mobile-phone-with-arrow} {calling} {Mobile~phone~with~arrow} {0.6}
\__emoji_def:nnnnn {^^^^260e^^^^fe0f} {telephone} {phone} {Telephone} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4de} {telephone-receiver} {} {Telephone~receiver} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4df} {pager} {} {Pager} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e0} {fax-machine} {fax} {Fax~machine} {0.6}
\__emoji_subgroup:n {Computer}
\__emoji_def:nnnnn {^^^^^^01f50b} {battery} {} {Battery} {0.6}
\__emoji_def:nnnnn {^^^^^^01f50c} {electric-plug} {} {Electric~plug} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4bb} {laptop} {computer} {Laptop} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5a5^^^^fe0f} {desktop-computer} {} {Desktop~computer} {0.7}
\__emoji_def:nnnnn {^^^^^^01f5a8^^^^fe0f} {printer} {} {Printer} {0.7}
\__emoji_def:nnnnn {^^^^2328^^^^fe0f} {keyboard} {} {Keyboard} {1.0}
\__emoji_def:nnnnn {^^^^^^01f5b1^^^^fe0f} {computer-mouse} {} {Computer~mouse} {0.7}
\__emoji_def:nnnnn {^^^^^^01f5b2^^^^fe0f} {trackball} {} {Trackball} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4bd} {computer-disk} {minidisc} {Computer~disk} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4be} {floppy-disk} {} {Floppy~disk} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4bf} {optical-disk} {cd} {Optical~disk} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c0} {dvd} {} {Dvd} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9ee} {abacus} {} {Abacus} {11.0}
\__emoji_subgroup:n {Light~\&~video}
\__emoji_def:nnnnn {^^^^^^01f3a5} {movie-camera} {} {Movie~camera} {0.6}
\__emoji_def:nnnnn {^^^^^^01f39e^^^^fe0f} {film-frames} {film-strip} {Film~frames} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4fd^^^^fe0f} {film-projector} {} {Film~projector} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3ac} {clapper-board} {clapper} {Clapper~board} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4fa} {television} {tv} {Television} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4f7} {camera} {} {Camera} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4f8} {camera-with-flash} {camera-flash} {Camera~with~flash} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4f9} {video-camera} {} {Video~camera} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4fc} {videocassette} {vhs} {Videocassette} {0.6}
\__emoji_def:nnnnn {^^^^^^01f50d} {magnifying-glass-tilted-left} {mag} {Magnifying~glass~tilted~left} {0.6}
\__emoji_def:nnnnn {^^^^^^01f50e} {magnifying-glass-tilted-right} {mag-right} {Magnifying~glass~tilted~right} {0.6}
\__emoji_def:nnnnn {^^^^^^01f56f^^^^fe0f} {candle} {} {Candle} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4a1} {light-bulb} {bulb} {Light~bulb} {0.6}
\__emoji_def:nnnnn {^^^^^^01f526} {flashlight} {} {Flashlight} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3ee} {red-paper-lantern} {izakaya-lantern, lantern} {Red~paper~lantern} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa94} {diya-lamp} {} {Diya~lamp} {12.0}
\__emoji_subgroup:n {Book-paper}
\__emoji_def:nnnnn {^^^^^^01f4d4} {notebook-with-decorative-cover} {} {Notebook~with~decorative~cover} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d5} {closed-book} {} {Closed~book} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d6} {open-book} {book} {Open~book} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d7} {green-book} {} {Green~book} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d8} {blue-book} {} {Blue~book} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d9} {orange-book} {} {Orange~book} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4da} {books} {} {Books} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d3} {notebook} {} {Notebook} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d2} {ledger} {} {Ledger} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c3} {page-with-curl} {} {Page~with~curl} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4dc} {scroll} {} {Scroll} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c4} {page-facing-up} {} {Page~facing~up} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4f0} {newspaper} {} {Newspaper} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5de^^^^fe0f} {rolled-up-newspaper} {newspaper-roll} {Rolled-up~newspaper} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4d1} {bookmark-tabs} {} {Bookmark~tabs} {0.6}
\__emoji_def:nnnnn {^^^^^^01f516} {bookmark} {} {Bookmark} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3f7^^^^fe0f} {label} {} {Label} {0.7}
\__emoji_subgroup:n {Money}
\__emoji_def:nnnnn {^^^^^^01f4b0} {money-bag} {moneybag} {Money~bag} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa99} {coin} {} {Coin} {13.0}
\__emoji_def:nnnnn {^^^^^^01f4b4} {yen-banknote} {yen} {Yen~banknote} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4b5} {dollar-banknote} {dollar} {Dollar~banknote} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4b6} {euro-banknote} {euro} {Euro~banknote} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4b7} {pound-banknote} {pound} {Pound~banknote} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4b8} {money-with-wings} {} {Money~with~wings} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4b3} {credit-card} {} {Credit~card} {0.6}
\__emoji_def:nnnnn {^^^^^^01f9fe} {receipt} {} {Receipt} {11.0}
\__emoji_def:nnnnn {^^^^^^01f4b9} {chart-increasing-with-yen} {chart} {Chart~increasing~with~yen} {0.6}
\__emoji_subgroup:n {Mail}
\__emoji_def:nnnnn {^^^^2709^^^^fe0f} {envelope} {} {Envelope} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e7} {e-mail} {email} {E-mail} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e8} {incoming-envelope} {} {Incoming~envelope} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e9} {envelope-with-arrow} {} {Envelope~with~arrow} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e4} {outbox-tray} {} {Outbox~tray} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e5} {inbox-tray} {} {Inbox~tray} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4e6} {package} {} {Package} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4eb} {closed-mailbox-with-raised-flag} {mailbox} {Closed~mailbox~with~raised~flag} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ea} {closed-mailbox-with-lowered-flag} {mailbox-closed} {Closed~mailbox~with~lowered~flag} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ec} {open-mailbox-with-raised-flag} {mailbox-with-mail} {Open~mailbox~with~raised~flag} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4ed} {open-mailbox-with-lowered-flag} {mailbox-with-no-mail} {Open~mailbox~with~lowered~flag} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4ee} {postbox} {} {Postbox} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5f3^^^^fe0f} {ballot-box-with-ballot} {ballot-box} {Ballot~box~with~ballot} {0.7}
\__emoji_subgroup:n {Writing}
\__emoji_def:nnnnn {^^^^270f^^^^fe0f} {pencil} {pencil2} {Pencil} {0.6}
\__emoji_def:nnnnn {^^^^2712^^^^fe0f} {black-nib} {} {Black~nib} {0.6}
\__emoji_def:nnnnn {^^^^^^01f58b^^^^fe0f} {fountain-pen} {} {Fountain~pen} {0.7}
\__emoji_def:nnnnn {^^^^^^01f58a^^^^fe0f} {pen} {} {Pen} {0.7}
\__emoji_def:nnnnn {^^^^^^01f58c^^^^fe0f} {paintbrush} {} {Paintbrush} {0.7}
\__emoji_def:nnnnn {^^^^^^01f58d^^^^fe0f} {crayon} {} {Crayon} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4dd} {memo} {} {Memo} {0.6}
\__emoji_subgroup:n {Office}
\__emoji_def:nnnnn {^^^^^^01f4bc} {briefcase} {} {Briefcase} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c1} {file-folder} {} {File~folder} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c2} {open-file-folder} {} {Open~file~folder} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5c2^^^^fe0f} {card-index-dividers} {} {Card~index~dividers} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4c5} {calendar} {date} {Calendar} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c6} {tear-off-calendar} {} {Tear-off~calendar} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5d2^^^^fe0f} {spiral-notepad} {} {Spiral~notepad} {0.7}
\__emoji_def:nnnnn {^^^^^^01f5d3^^^^fe0f} {spiral-calendar} {} {Spiral~calendar} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4c7} {card-index} {} {Card~index} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c8} {chart-increasing} {chart-with-upwards-trend} {Chart~increasing} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4c9} {chart-decreasing} {chart-with-downwards-trend} {Chart~decreasing} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ca} {bar-chart} {} {Bar~chart} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4cb} {clipboard} {} {Clipboard} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4cc} {pushpin} {} {Pushpin} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4cd} {round-pushpin} {} {Round~pushpin} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4ce} {paperclip} {} {Paperclip} {0.6}
\__emoji_def:nnnnn {^^^^^^01f587^^^^fe0f} {linked-paperclips} {paperclips} {Linked~paperclips} {0.7}
\__emoji_def:nnnnn {^^^^^^01f4cf} {straight-ruler} {} {Straight~ruler} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4d0} {triangular-ruler} {} {Triangular~ruler} {0.6}
\__emoji_def:nnnnn {^^^^2702^^^^fe0f} {scissors} {} {Scissors} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5c3^^^^fe0f} {card-file-box} {} {Card~file~box} {0.7}
\__emoji_def:nnnnn {^^^^^^01f5c4^^^^fe0f} {file-cabinet} {} {File~cabinet} {0.7}
\__emoji_def:nnnnn {^^^^^^01f5d1^^^^fe0f} {wastebasket} {} {Wastebasket} {0.7}
\__emoji_subgroup:n {Lock}
\__emoji_def:nnnnn {^^^^^^01f512} {locked} {lock} {Locked} {0.6}
\__emoji_def:nnnnn {^^^^^^01f513} {unlocked} {unlock} {Unlocked} {0.6}
\__emoji_def:nnnnn {^^^^^^01f50f} {locked-with-pen} {lock-with-ink-pen} {Locked~with~pen} {0.6}
\__emoji_def:nnnnn {^^^^^^01f510} {locked-with-key} {closed-lock-with-key} {Locked~with~key} {0.6}
\__emoji_def:nnnnn {^^^^^^01f511} {key} {} {Key} {0.6}
\__emoji_def:nnnnn {^^^^^^01f5dd^^^^fe0f} {old-key} {} {Old~key} {0.7}
\__emoji_subgroup:n {Tool}
\__emoji_def:nnnnn {^^^^^^01f528} {hammer} {} {Hammer} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa93} {axe} {} {Axe} {12.0}
\__emoji_def:nnnnn {^^^^26cf^^^^fe0f} {pick} {} {Pick} {0.7}
\__emoji_def:nnnnn {^^^^2692^^^^fe0f} {hammer-and-pick} {} {Hammer~and~pick} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6e0^^^^fe0f} {hammer-and-wrench} {} {Hammer~and~wrench} {0.7}
\__emoji_def:nnnnn {^^^^^^01f5e1^^^^fe0f} {dagger} {} {Dagger} {0.7}
\__emoji_def:nnnnn {^^^^2694^^^^fe0f} {crossed-swords} {} {Crossed~swords} {1.0}
\__emoji_def:nnnnn {^^^^^^01f52b} {water-pistol} {gun} {Water~pistol} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa83} {boomerang} {} {Boomerang} {13.0}
\__emoji_def:nnnnn {^^^^^^01f3f9} {bow-and-arrow} {} {Bow~and~arrow} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6e1^^^^fe0f} {shield} {} {Shield} {0.7}
\__emoji_def:nnnnn {^^^^^^01fa9a} {carpentry-saw} {} {Carpentry~saw} {13.0}
\__emoji_def:nnnnn {^^^^^^01f527} {wrench} {} {Wrench} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa9b} {screwdriver} {} {Screwdriver} {13.0}
\__emoji_def:nnnnn {^^^^^^01f529} {nut-and-bolt} {} {Nut~and~bolt} {0.6}
\__emoji_def:nnnnn {^^^^2699^^^^fe0f} {gear} {} {Gear} {1.0}
\__emoji_def:nnnnn {^^^^^^01f5dc^^^^fe0f} {clamp} {} {Clamp} {0.7}
\__emoji_def:nnnnn {^^^^2696^^^^fe0f} {balance-scale} {} {Balance~scale} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9af} {white-cane} {probing-cane} {White~cane} {12.0}
\__emoji_def:nnnnn {^^^^^^01f517} {link} {} {Link} {0.6}
\__emoji_def:nnnnn {^^^^26d3^^^^fe0f} {chains} {} {Chains} {0.7}
\__emoji_def:nnnnn {^^^^^^01fa9d} {hook} {} {Hook} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9f0} {toolbox} {} {Toolbox} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9f2} {magnet} {} {Magnet} {11.0}
\__emoji_def:nnnnn {^^^^^^01fa9c} {ladder} {} {Ladder} {13.0}
\__emoji_subgroup:n {Science}
\__emoji_def:nnnnn {^^^^2697^^^^fe0f} {alembic} {} {Alembic} {1.0}
\__emoji_def:nnnnn {^^^^^^01f9ea} {test-tube} {} {Test~tube} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9eb} {petri-dish} {} {Petri~dish} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9ec} {dna} {} {Dna} {11.0}
\__emoji_def:nnnnn {^^^^^^01f52c} {microscope} {} {Microscope} {1.0}
\__emoji_def:nnnnn {^^^^^^01f52d} {telescope} {} {Telescope} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4e1} {satellite-antenna} {} {Satellite~antenna} {0.6}
\__emoji_subgroup:n {Medical}
\__emoji_def:nnnnn {^^^^^^01f489} {syringe} {} {Syringe} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa78} {drop-of-blood} {} {Drop~of~blood} {12.0}
\__emoji_def:nnnnn {^^^^^^01f48a} {pill} {} {Pill} {0.6}
\__emoji_def:nnnnn {^^^^^^01fa79} {adhesive-bandage} {} {Adhesive~bandage} {12.0}
\__emoji_def:nnnnn {^^^^^^01fa7a} {stethoscope} {} {Stethoscope} {12.0}
\__emoji_subgroup:n {Household}
\__emoji_def:nnnnn {^^^^^^01f6aa} {door} {} {Door} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6d7} {elevator} {} {Elevator} {13.0}
\__emoji_def:nnnnn {^^^^^^01fa9e} {mirror} {} {Mirror} {13.0}
\__emoji_def:nnnnn {^^^^^^01fa9f} {window} {} {Window} {13.0}
\__emoji_def:nnnnn {^^^^^^01f6cf^^^^fe0f} {bed} {} {Bed} {0.7}
\__emoji_def:nnnnn {^^^^^^01f6cb^^^^fe0f} {couch-and-lamp} {} {Couch~and~lamp} {0.7}
\__emoji_def:nnnnn {^^^^^^01fa91} {chair} {} {Chair} {12.0}
\__emoji_def:nnnnn {^^^^^^01f6bd} {toilet} {} {Toilet} {0.6}
\__emoji_def:nnnnn {^^^^^^01faa0} {plunger} {} {Plunger} {13.0}
\__emoji_def:nnnnn {^^^^^^01f6bf} {shower} {} {Shower} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c1} {bathtub} {} {Bathtub} {1.0}
\__emoji_def:nnnnn {^^^^^^01faa4} {mouse-trap} {} {Mouse~trap} {13.0}
\__emoji_def:nnnnn {^^^^^^01fa92} {razor} {} {Razor} {12.0}
\__emoji_def:nnnnn {^^^^^^01f9f4} {lotion-bottle} {} {Lotion~bottle} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9f7} {safety-pin} {} {Safety~pin} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9f9} {broom} {} {Broom} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9fa} {basket} {} {Basket} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9fb} {roll-of-paper} {} {Roll~of~paper} {11.0}
\__emoji_def:nnnnn {^^^^^^01faa3} {bucket} {} {Bucket} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9fc} {soap} {} {Soap} {11.0}
\__emoji_def:nnnnn {^^^^^^01faa5} {toothbrush} {} {Toothbrush} {13.0}
\__emoji_def:nnnnn {^^^^^^01f9fd} {sponge} {} {Sponge} {11.0}
\__emoji_def:nnnnn {^^^^^^01f9ef} {fire-extinguisher} {} {Fire~extinguisher} {11.0}
\__emoji_def:nnnnn {^^^^^^01f6d2} {shopping-cart} {} {Shopping~cart} {3.0}
\__emoji_subgroup:n {Other-object}
\__emoji_def:nnnnn {^^^^^^01f6ac} {cigarette} {smoking} {Cigarette} {0.6}
\__emoji_def:nnnnn {^^^^26b0^^^^fe0f} {coffin} {} {Coffin} {1.0}
\__emoji_def:nnnnn {^^^^^^01faa6} {headstone} {} {Headstone} {13.0}
\__emoji_def:nnnnn {^^^^26b1^^^^fe0f} {funeral-urn} {} {Funeral~urn} {1.0}
\__emoji_def:nnnnn {^^^^^^01f5ff} {moai} {moyai} {Moai} {0.6}
\__emoji_def:nnnnn {^^^^^^01faa7} {placard} {} {Placard} {13.0}
\__emoji_group:n {Symbols}
\__emoji_subgroup:n {Transport-sign}
\__emoji_def:nnnnn {^^^^^^01f3e7} {atm-sign} {atm} {ATM~sign} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6ae} {litter-in-bin-sign} {put-litter-in-its-place} {Litter~in~bin~sign} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b0} {potable-water} {} {Potable~water} {1.0}
\__emoji_def:nnnnn {^^^^267f} {wheelchair-symbol} {wheelchair} {Wheelchair~symbol} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6b9} {mens-room} {mens} {Men's~room} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6ba} {womens-room} {womens} {Women's~room} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6bb} {restroom} {} {Restroom} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6bc} {baby-symbol} {} {Baby~symbol} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6be} {water-closet} {wc} {Water~closet} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6c2} {passport-control} {} {Passport~control} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c3} {customs} {} {Customs} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c4} {baggage-claim} {} {Baggage~claim} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6c5} {left-luggage} {} {Left~luggage} {1.0}
\__emoji_subgroup:n {Warning}
\__emoji_def:nnnnn {^^^^26a0^^^^fe0f} {warning} {} {Warning} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6b8} {children-crossing} {} {Children~crossing} {1.0}
\__emoji_def:nnnnn {^^^^26d4} {no-entry} {} {No~entry} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6ab} {prohibited} {no-entry-sign} {Prohibited} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6b3} {no-bicycles} {} {No~bicycles} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6ad} {no-smoking} {} {No~smoking} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6af} {no-littering} {do-not-litter} {No~littering} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b1} {non-potable-water} {} {Non-potable~water} {1.0}
\__emoji_def:nnnnn {^^^^^^01f6b7} {no-pedestrians} {} {No~pedestrians} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4f5} {no-mobile-phones} {} {No~mobile~phones} {1.0}
\__emoji_def:nnnnn {^^^^^^01f51e} {no-one-under-eighteen} {underage} {No~one~under~eighteen} {0.6}
\__emoji_def:nnnnn {^^^^2622^^^^fe0f} {radioactive} {} {Radioactive} {1.0}
\__emoji_def:nnnnn {^^^^2623^^^^fe0f} {biohazard} {} {Biohazard} {1.0}
\__emoji_subgroup:n {Arrow}
\__emoji_def:nnnnn {^^^^2b06^^^^fe0f} {up-arrow} {arrow-up} {Up~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2197^^^^fe0f} {up-right-arrow} {arrow-upper-right} {Up-right~arrow} {0.6}
\__emoji_def:nnnnn {^^^^27a1^^^^fe0f} {right-arrow} {arrow-right} {Right~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2198^^^^fe0f} {down-right-arrow} {arrow-lower-right} {Down-right~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2b07^^^^fe0f} {down-arrow} {arrow-down} {Down~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2199^^^^fe0f} {down-left-arrow} {arrow-lower-left} {Down-left~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2b05^^^^fe0f} {left-arrow} {arrow-left} {Left~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2196^^^^fe0f} {up-left-arrow} {arrow-upper-left} {Up-left~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2195^^^^fe0f} {up-down-arrow} {arrow-up-down} {Up-down~arrow} {0.6}
\__emoji_def:nnnnn {^^^^2194^^^^fe0f} {left-right-arrow} {} {Left-right~arrow} {0.6}
\__emoji_def:nnnnn {^^^^21a9^^^^fe0f} {right-arrow-curving-left} {leftwards-arrow-with-hook} {Right~arrow~curving~left} {0.6}
\__emoji_def:nnnnn {^^^^21aa^^^^fe0f} {left-arrow-curving-right} {arrow-right-hook} {Left~arrow~curving~right} {0.6}
\__emoji_def:nnnnn {^^^^2934^^^^fe0f} {right-arrow-curving-up} {arrow-heading-up} {Right~arrow~curving~up} {0.6}
\__emoji_def:nnnnn {^^^^2935^^^^fe0f} {right-arrow-curving-down} {arrow-heading-down} {Right~arrow~curving~down} {0.6}
\__emoji_def:nnnnn {^^^^^^01f503} {clockwise-vertical-arrows} {arrows-clockwise} {Clockwise~vertical~arrows} {0.6}
\__emoji_def:nnnnn {^^^^^^01f504} {counterclockwise-arrows-button} {arrows-counterclockwise} {Counterclockwise~arrows~button} {1.0}
\__emoji_def:nnnnn {^^^^^^01f519} {back-arrow} {back} {BACK~arrow} {0.6}
\__emoji_def:nnnnn {^^^^^^01f51a} {end-arrow} {end} {END~arrow} {0.6}
\__emoji_def:nnnnn {^^^^^^01f51b} {on!-arrow} {on} {ON!~arrow} {0.6}
\__emoji_def:nnnnn {^^^^^^01f51c} {soon-arrow} {soon} {SOON~arrow} {0.6}
\__emoji_def:nnnnn {^^^^^^01f51d} {top-arrow} {top} {TOP~arrow} {0.6}
\__emoji_subgroup:n {Religion}
\__emoji_def:nnnnn {^^^^^^01f6d0} {place-of-worship} {} {Place~of~worship} {1.0}
\__emoji_def:nnnnn {^^^^269b^^^^fe0f} {atom-symbol} {} {Atom~symbol} {1.0}
\__emoji_def:nnnnn {^^^^^^01f549^^^^fe0f} {om} {} {Om} {0.7}
\__emoji_def:nnnnn {^^^^2721^^^^fe0f} {star-of-david} {} {Star~of~David} {0.7}
\__emoji_def:nnnnn {^^^^2638^^^^fe0f} {wheel-of-dharma} {} {Wheel~of~dharma} {0.7}
\__emoji_def:nnnnn {^^^^262f^^^^fe0f} {yin-yang} {} {Yin~yang} {0.7}
\__emoji_def:nnnnn {^^^^271d^^^^fe0f} {latin-cross} {} {Latin~cross} {0.7}
\__emoji_def:nnnnn {^^^^2626^^^^fe0f} {orthodox-cross} {} {Orthodox~cross} {1.0}
\__emoji_def:nnnnn {^^^^262a^^^^fe0f} {star-and-crescent} {} {Star~and~crescent} {0.7}
\__emoji_def:nnnnn {^^^^262e^^^^fe0f} {peace-symbol} {} {Peace~symbol} {1.0}
\__emoji_def:nnnnn {^^^^^^01f54e} {menorah} {} {Menorah} {1.0}
\__emoji_def:nnnnn {^^^^^^01f52f} {dotted-six-pointed-star} {six-pointed-star} {Dotted~six-pointed~star} {0.6}
\__emoji_subgroup:n {Zodiac}
\__emoji_def:nnnnn {^^^^2648} {aries} {} {Aries} {0.6}
\__emoji_def:nnnnn {^^^^2649} {taurus} {} {Taurus} {0.6}
\__emoji_def:nnnnn {^^^^264a} {gemini} {} {Gemini} {0.6}
\__emoji_def:nnnnn {^^^^264b} {cancer} {} {Cancer} {0.6}
\__emoji_def:nnnnn {^^^^264c} {leo} {} {Leo} {0.6}
\__emoji_def:nnnnn {^^^^264d} {virgo} {} {Virgo} {0.6}
\__emoji_def:nnnnn {^^^^264e} {libra} {} {Libra} {0.6}
\__emoji_def:nnnnn {^^^^264f} {scorpio} {scorpius} {Scorpio} {0.6}
\__emoji_def:nnnnn {^^^^2650} {sagittarius} {} {Sagittarius} {0.6}
\__emoji_def:nnnnn {^^^^2651} {capricorn} {} {Capricorn} {0.6}
\__emoji_def:nnnnn {^^^^2652} {aquarius} {} {Aquarius} {0.6}
\__emoji_def:nnnnn {^^^^2653} {pisces} {} {Pisces} {0.6}
\__emoji_def:nnnnn {^^^^26ce} {ophiuchus} {} {Ophiuchus} {0.6}
\__emoji_subgroup:n {Av-symbol}
\__emoji_def:nnnnn {^^^^^^01f500} {shuffle-tracks-button} {twisted-rightwards-arrows} {Shuffle~tracks~button} {1.0}
\__emoji_def:nnnnn {^^^^^^01f501} {repeat-button} {repeat} {Repeat~button} {1.0}
\__emoji_def:nnnnn {^^^^^^01f502} {repeat-single-button} {repeat-one} {Repeat~single~button} {1.0}
\__emoji_def:nnnnn {^^^^25b6^^^^fe0f} {play-button} {arrow-forward} {Play~button} {0.6}
\__emoji_def:nnnnn {^^^^23e9} {fast-forward-button} {fast-forward} {Fast-forward~button} {0.6}
\__emoji_def:nnnnn {^^^^23ed^^^^fe0f} {next-track-button} {} {Next~track~button} {0.7}
\__emoji_def:nnnnn {^^^^23ef^^^^fe0f} {play-or-pause-button} {} {Play~or~pause~button} {1.0}
\__emoji_def:nnnnn {^^^^25c0^^^^fe0f} {reverse-button} {arrow-backward} {Reverse~button} {0.6}
\__emoji_def:nnnnn {^^^^23ea} {fast-reverse-button} {rewind} {Fast~reverse~button} {0.6}
\__emoji_def:nnnnn {^^^^23ee^^^^fe0f} {last-track-button} {previous-track-button} {Last~track~button} {0.7}
\__emoji_def:nnnnn {^^^^^^01f53c} {upwards-button} {arrow-up-small} {Upwards~button} {0.6}
\__emoji_def:nnnnn {^^^^23eb} {fast-up-button} {arrow-double-up} {Fast~up~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f53d} {downwards-button} {arrow-down-small} {Downwards~button} {0.6}
\__emoji_def:nnnnn {^^^^23ec} {fast-down-button} {arrow-double-down} {Fast~down~button} {0.6}
\__emoji_def:nnnnn {^^^^23f8^^^^fe0f} {pause-button} {} {Pause~button} {0.7}
\__emoji_def:nnnnn {^^^^23f9^^^^fe0f} {stop-button} {} {Stop~button} {0.7}
\__emoji_def:nnnnn {^^^^23fa^^^^fe0f} {record-button} {} {Record~button} {0.7}
\__emoji_def:nnnnn {^^^^23cf^^^^fe0f} {eject-button} {} {Eject~button} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3a6} {cinema} {} {Cinema} {0.6}
\__emoji_def:nnnnn {^^^^^^01f505} {dim-button} {low-brightness} {Dim~button} {1.0}
\__emoji_def:nnnnn {^^^^^^01f506} {bright-button} {high-brightness} {Bright~button} {1.0}
\__emoji_def:nnnnn {^^^^^^01f4f6} {antenna-bars} {signal-strength} {Antenna~bars} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4f3} {vibration-mode} {} {Vibration~mode} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4f4} {mobile-phone-off} {} {Mobile~phone~off} {0.6}
\__emoji_subgroup:n {Gender}
\__emoji_def:nnnnn {^^^^2640^^^^fe0f} {female-sign} {} {Female~sign} {4.0}
\__emoji_def:nnnnn {^^^^2642^^^^fe0f} {male-sign} {} {Male~sign} {4.0}
\__emoji_def:nnnnn {^^^^26a7^^^^fe0f} {transgender-symbol} {} {Transgender~symbol} {13.0}
\__emoji_subgroup:n {Math}
\__emoji_def:nnnnn {^^^^2716^^^^fe0f} {multiply} {heavy-multiplication-x} {Multiply} {0.6}
\__emoji_def:nnnnn {^^^^2795} {plus} {heavy-plus-sign} {Plus} {0.6}
\__emoji_def:nnnnn {^^^^2796} {minus} {heavy-minus-sign} {Minus} {0.6}
\__emoji_def:nnnnn {^^^^2797} {divide} {heavy-division-sign} {Divide} {0.6}
\__emoji_def:nnnnn {^^^^267e^^^^fe0f} {infinity} {} {Infinity} {11.0}
\__emoji_subgroup:n {Punctuation}
\__emoji_def:nnnnn {^^^^203c^^^^fe0f} {double-exclamation-mark} {bangbang} {Double~exclamation~mark} {0.6}
\__emoji_def:nnnnn {^^^^2049^^^^fe0f} {exclamation-question-mark} {interrobang} {Exclamation~question~mark} {0.6}
\__emoji_def:nnnnn {^^^^2753} {red-question-mark} {question} {Red~question~mark} {0.6}
\__emoji_def:nnnnn {^^^^2754} {white-question-mark} {grey-question} {White~question~mark} {0.6}
\__emoji_def:nnnnn {^^^^2755} {white-exclamation-mark} {grey-exclamation} {White~exclamation~mark} {0.6}
\__emoji_def:nnnnn {^^^^2757} {red-exclamation-mark} {exclamation, heavy-exclamation-mark} {Red~exclamation~mark} {0.6}
\__emoji_def:nnnnn {^^^^3030^^^^fe0f} {wavy-dash} {} {Wavy~dash} {0.6}
\__emoji_subgroup:n {Currency}
\__emoji_def:nnnnn {^^^^^^01f4b1} {currency-exchange} {} {Currency~exchange} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4b2} {heavy-dollar-sign} {} {Heavy~dollar~sign} {0.6}
\__emoji_subgroup:n {Other-symbol}
\__emoji_def:nnnnn {^^^^2695^^^^fe0f} {medical-symbol} {} {Medical~symbol} {4.0}
\__emoji_def:nnnnn {^^^^267b^^^^fe0f} {recycling-symbol} {recycle} {Recycling~symbol} {0.6}
\__emoji_def:nnnnn {^^^^269c^^^^fe0f} {fleur-de-lis} {} {Fleur-de-lis} {1.0}
\__emoji_def:nnnnn {^^^^^^01f531} {trident-emblem} {trident} {Trident~emblem} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4db} {name-badge} {} {Name~badge} {0.6}
\__emoji_def:nnnnn {^^^^^^01f530} {japanese-symbol-for-beginner} {beginner} {Japanese~symbol~for~beginner} {0.6}
\__emoji_def:nnnnn {^^^^2b55} {hollow-red-circle} {o} {Hollow~red~circle} {0.6}
\__emoji_def:nnnnn {^^^^2705} {check-mark-button} {white-check-mark} {Check~mark~button} {0.6}
\__emoji_def:nnnnn {^^^^2611^^^^fe0f} {check-box-with-check} {ballot-box-with-check} {Check~box~with~check} {0.6}
\__emoji_def:nnnnn {^^^^2714^^^^fe0f} {check-mark} {heavy-check-mark} {Check~mark} {0.6}
\__emoji_def:nnnnn {^^^^274c} {cross-mark} {x} {Cross~mark} {0.6}
\__emoji_def:nnnnn {^^^^274e} {cross-mark-button} {negative-squared-cross-mark} {Cross~mark~button} {0.6}
\__emoji_def:nnnnn {^^^^27b0} {curly-loop} {} {Curly~loop} {0.6}
\__emoji_def:nnnnn {^^^^27bf} {double-curly-loop} {loop} {Double~curly~loop} {1.0}
\__emoji_def:nnnnn {^^^^303d^^^^fe0f} {part-alternation-mark} {} {Part~alternation~mark} {0.6}
\__emoji_def:nnnnn {^^^^2733^^^^fe0f} {eight-spoked-asterisk} {} {Eight-spoked~asterisk} {0.6}
\__emoji_def:nnnnn {^^^^2734^^^^fe0f} {eight-pointed-star} {eight-pointed-black-star} {Eight-pointed~star} {0.6}
\__emoji_def:nnnnn {^^^^2747^^^^fe0f} {sparkle} {} {Sparkle} {0.6}
\__emoji_def:nnnnn {^^^^00a9^^^^fe0f} {copyright} {} {Copyright} {0.6}
\__emoji_def:nnnnn {^^^^00ae^^^^fe0f} {registered} {} {Registered} {0.6}
\__emoji_def:nnnnn {^^^^2122^^^^fe0f} {trade-mark} {tm} {Trade~mark} {0.6}
\__emoji_subgroup:n {Keycap}
\__emoji_def:nnnnn {^^^^0023^^^^fe0f^^^^20e3} {keycap-hash} {hash} {Keycap:~\#} {0.6}
\__emoji_def:nnnnn {^^^^002a^^^^fe0f^^^^20e3} {keycap-asterisk} {asterisk} {Keycap:~*} {2.0}
\__emoji_def:nnnnn {^^^^0030^^^^fe0f^^^^20e3} {keycap-0} {zero} {Keycap:~0} {0.6}
\__emoji_def:nnnnn {^^^^0031^^^^fe0f^^^^20e3} {keycap-1} {one} {Keycap:~1} {0.6}
\__emoji_def:nnnnn {^^^^0032^^^^fe0f^^^^20e3} {keycap-2} {two} {Keycap:~2} {0.6}
\__emoji_def:nnnnn {^^^^0033^^^^fe0f^^^^20e3} {keycap-3} {three} {Keycap:~3} {0.6}
\__emoji_def:nnnnn {^^^^0034^^^^fe0f^^^^20e3} {keycap-4} {four} {Keycap:~4} {0.6}
\__emoji_def:nnnnn {^^^^0035^^^^fe0f^^^^20e3} {keycap-5} {five} {Keycap:~5} {0.6}
\__emoji_def:nnnnn {^^^^0036^^^^fe0f^^^^20e3} {keycap-6} {six} {Keycap:~6} {0.6}
\__emoji_def:nnnnn {^^^^0037^^^^fe0f^^^^20e3} {keycap-7} {seven} {Keycap:~7} {0.6}
\__emoji_def:nnnnn {^^^^0038^^^^fe0f^^^^20e3} {keycap-8} {eight} {Keycap:~8} {0.6}
\__emoji_def:nnnnn {^^^^0039^^^^fe0f^^^^20e3} {keycap-9} {nine} {Keycap:~9} {0.6}
\__emoji_def:nnnnn {^^^^^^01f51f} {keycap-10} {keycap-ten} {Keycap:~10} {0.6}
\__emoji_subgroup:n {Alphanum}
\__emoji_def:nnnnn {^^^^^^01f520} {input-latin-uppercase} {capital-abcd} {Input~latin~uppercase} {0.6}
\__emoji_def:nnnnn {^^^^^^01f521} {input-latin-lowercase} {abcd} {Input~latin~lowercase} {0.6}
\__emoji_def:nnnnn {^^^^^^01f522} {input-numbers} {1234} {Input~numbers} {0.6}
\__emoji_def:nnnnn {^^^^^^01f523} {input-symbols} {symbols} {Input~symbols} {0.6}
\__emoji_def:nnnnn {^^^^^^01f524} {input-latin-letters} {abc} {Input~latin~letters} {0.6}
\__emoji_def:nnnnn {^^^^^^01f170^^^^fe0f} {a-button-blood-type} {a} {A~button~(blood~type)} {0.6}
\__emoji_def:nnnnn {^^^^^^01f18e} {ab-button-blood-type} {ab} {AB~button~(blood~type)} {0.6}
\__emoji_def:nnnnn {^^^^^^01f171^^^^fe0f} {b-button-blood-type} {b} {B~button~(blood~type)} {0.6}
\__emoji_def:nnnnn {^^^^^^01f191} {cl-button} {cl} {CL~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f192} {cool-button} {cool} {COOL~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f193} {free-button} {free} {FREE~button} {0.6}
\__emoji_def:nnnnn {^^^^2139^^^^fe0f} {information} {information-source} {Information} {0.6}
\__emoji_def:nnnnn {^^^^^^01f194} {id-button} {id} {ID~button} {0.6}
\__emoji_def:nnnnn {^^^^24c2^^^^fe0f} {circled-m} {m} {Circled~M} {0.6}
\__emoji_def:nnnnn {^^^^^^01f195} {new-button} {new} {NEW~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f196} {ng-button} {ng} {NG~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f17e^^^^fe0f} {o-button-blood-type} {o2} {O~button~(blood~type)} {0.6}
\__emoji_def:nnnnn {^^^^^^01f197} {ok-button} {ok} {OK~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f17f^^^^fe0f} {p-button} {parking} {P~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f198} {sos-button} {sos} {SOS~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f199} {up!-button} {up} {UP!~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f19a} {vs-button} {vs} {VS~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f201} {japanese-here-button} {koko} {Japanese~``here''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f202^^^^fe0f} {japanese-service-charge-button} {sa} {Japanese~``service~charge''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f237^^^^fe0f} {japanese-monthly-amount-button} {u6708} {Japanese~``monthly~amount''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f236} {japanese-not-free-of-charge-button} {u6709} {Japanese~``not~free~of~charge''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f22f} {japanese-reserved-button} {u6307} {Japanese~``reserved''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f250} {japanese-bargain-button} {ideograph-advantage} {Japanese~``bargain''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f239} {japanese-discount-button} {u5272} {Japanese~``discount''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f21a} {japanese-free-of-charge-button} {u7121} {Japanese~``free~of~charge''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f232} {japanese-prohibited-button} {u7981} {Japanese~``prohibited''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f251} {japanese-acceptable-button} {accept} {Japanese~``acceptable''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f238} {japanese-application-button} {u7533} {Japanese~``application''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f234} {japanese-passing-grade-button} {u5408} {Japanese~``passing~grade''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f233} {japanese-vacancy-button} {u7a7a} {Japanese~``vacancy''~button} {0.6}
\__emoji_def:nnnnn {^^^^3297^^^^fe0f} {japanese-congratulations-button} {congratulations} {Japanese~``congratulations''~button} {0.6}
\__emoji_def:nnnnn {^^^^3299^^^^fe0f} {japanese-secret-button} {secret} {Japanese~``secret''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f23a} {japanese-open-for-business-button} {u55b6} {Japanese~``open~for~business''~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f235} {japanese-no-vacancy-button} {u6e80} {Japanese~``no~vacancy''~button} {0.6}
\__emoji_subgroup:n {Geometric}
\__emoji_def:nnnnn {^^^^^^01f534} {red-circle} {} {Red~circle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f7e0} {orange-circle} {} {Orange~circle} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7e1} {yellow-circle} {} {Yellow~circle} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7e2} {green-circle} {} {Green~circle} {12.0}
\__emoji_def:nnnnn {^^^^^^01f535} {blue-circle} {large-blue-circle} {Blue~circle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f7e3} {purple-circle} {} {Purple~circle} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7e4} {brown-circle} {} {Brown~circle} {12.0}
\__emoji_def:nnnnn {^^^^26ab} {black-circle} {} {Black~circle} {0.6}
\__emoji_def:nnnnn {^^^^26aa} {white-circle} {} {White~circle} {0.6}
\__emoji_def:nnnnn {^^^^^^01f7e5} {red-square} {} {Red~square} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7e7} {orange-square} {} {Orange~square} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7e8} {yellow-square} {} {Yellow~square} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7e9} {green-square} {} {Green~square} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7e6} {blue-square} {} {Blue~square} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7ea} {purple-square} {} {Purple~square} {12.0}
\__emoji_def:nnnnn {^^^^^^01f7eb} {brown-square} {} {Brown~square} {12.0}
\__emoji_def:nnnnn {^^^^2b1b} {black-large-square} {} {Black~large~square} {0.6}
\__emoji_def:nnnnn {^^^^2b1c} {white-large-square} {} {White~large~square} {0.6}
\__emoji_def:nnnnn {^^^^25fc^^^^fe0f} {black-medium-square} {} {Black~medium~square} {0.6}
\__emoji_def:nnnnn {^^^^25fb^^^^fe0f} {white-medium-square} {} {White~medium~square} {0.6}
\__emoji_def:nnnnn {^^^^25fe} {black-medium-small-square} {} {Black~medium-small~square} {0.6}
\__emoji_def:nnnnn {^^^^25fd} {white-medium-small-square} {} {White~medium-small~square} {0.6}
\__emoji_def:nnnnn {^^^^25aa^^^^fe0f} {black-small-square} {} {Black~small~square} {0.6}
\__emoji_def:nnnnn {^^^^25ab^^^^fe0f} {white-small-square} {} {White~small~square} {0.6}
\__emoji_def:nnnnn {^^^^^^01f536} {large-orange-diamond} {} {Large~orange~diamond} {0.6}
\__emoji_def:nnnnn {^^^^^^01f537} {large-blue-diamond} {} {Large~blue~diamond} {0.6}
\__emoji_def:nnnnn {^^^^^^01f538} {small-orange-diamond} {} {Small~orange~diamond} {0.6}
\__emoji_def:nnnnn {^^^^^^01f539} {small-blue-diamond} {} {Small~blue~diamond} {0.6}
\__emoji_def:nnnnn {^^^^^^01f53a} {red-triangle-pointed-up} {small-red-triangle} {Red~triangle~pointed~up} {0.6}
\__emoji_def:nnnnn {^^^^^^01f53b} {red-triangle-pointed-down} {small-red-triangle-down} {Red~triangle~pointed~down} {0.6}
\__emoji_def:nnnnn {^^^^^^01f4a0} {diamond-with-a-dot} {diamond-shape-with-a-dot-inside} {Diamond~with~a~dot} {0.6}
\__emoji_def:nnnnn {^^^^^^01f518} {radio-button} {} {Radio~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f533} {white-square-button} {} {White~square~button} {0.6}
\__emoji_def:nnnnn {^^^^^^01f532} {black-square-button} {} {Black~square~button} {0.6}
\__emoji_group:n {Flags}
\__emoji_subgroup:n {Flag}
\__emoji_def:nnnnn {^^^^^^01f3c1} {chequered-flag} {checkered-flag} {Chequered~flag} {0.6}
\__emoji_def:nnnnn {^^^^^^01f6a9} {triangular-flag} {triangular-flag-on-post} {Triangular~flag} {0.6}
\__emoji_def:nnnnn {^^^^^^01f38c} {crossed-flags} {} {Crossed~flags} {0.6}
\__emoji_def:nnnnn {^^^^^^01f3f4} {black-flag} {} {Black~flag} {1.0}
\__emoji_def:nnnnn {^^^^^^01f3f3^^^^fe0f} {white-flag} {} {White~flag} {0.7}
\__emoji_def:nnnnn {^^^^^^01f3f3^^^^fe0f^^^^200d^^^^^^01f308} {rainbow-flag} {} {Rainbow~flag} {4.0}
\__emoji_def:nnnnn {^^^^^^01f3f3^^^^fe0f^^^^200d^^^^26a7^^^^fe0f} {transgender-flag} {} {Transgender~flag} {13.0}
\__emoji_def:nnnnn {^^^^^^01f3f4^^^^200d^^^^2620^^^^fe0f} {pirate-flag} {} {Pirate~flag} {11.0}
\__emoji_subgroup:n {Country-flag}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1e8} {flag-ascension-island} {ascension-island} {Flag:~Ascension~Island} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1e9} {flag-andorra} {andorra} {Flag:~Andorra} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1ea} {flag-united-arab-emirates} {united-arab-emirates} {Flag:~United~Arab~Emirates} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1eb} {flag-afghanistan} {afghanistan} {Flag:~Afghanistan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1ec} {flag-antigua-barbuda} {antigua-barbuda} {Flag:~Antigua~\&~Barbuda} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1ee} {flag-anguilla} {anguilla} {Flag:~Anguilla} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1f1} {flag-albania} {albania} {Flag:~Albania} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1f2} {flag-armenia} {armenia} {Flag:~Armenia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1f4} {flag-angola} {angola} {Flag:~Angola} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1f6} {flag-antarctica} {antarctica} {Flag:~Antarctica} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1f7} {flag-argentina} {argentina} {Flag:~Argentina} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1f8} {flag-american-samoa} {american-samoa} {Flag:~American~Samoa} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1f9} {flag-austria} {austria} {Flag:~Austria} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1fa} {flag-australia} {australia} {Flag:~Australia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1fc} {flag-aruba} {aruba} {Flag:~Aruba} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1fd} {flag-aland-islands} {aland-islands} {Flag:~Åland~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e6^^^^^^01f1ff} {flag-azerbaijan} {azerbaijan} {Flag:~Azerbaijan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1e6} {flag-bosnia-herzegovina} {bosnia-herzegovina} {Flag:~Bosnia~\&~Herzegovina} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1e7} {flag-barbados} {barbados} {Flag:~Barbados} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1e9} {flag-bangladesh} {bangladesh} {Flag:~Bangladesh} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1ea} {flag-belgium} {belgium} {Flag:~Belgium} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1eb} {flag-burkina-faso} {burkina-faso} {Flag:~Burkina~Faso} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1ec} {flag-bulgaria} {bulgaria} {Flag:~Bulgaria} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1ed} {flag-bahrain} {bahrain} {Flag:~Bahrain} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1ee} {flag-burundi} {burundi} {Flag:~Burundi} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1ef} {flag-benin} {benin} {Flag:~Benin} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f1} {flag-st-barthelemy} {st-barthelemy} {Flag:~St.~Barthélemy} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f2} {flag-bermuda} {bermuda} {Flag:~Bermuda} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f3} {flag-brunei} {brunei} {Flag:~Brunei} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f4} {flag-bolivia} {bolivia} {Flag:~Bolivia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f6} {flag-caribbean-netherlands} {caribbean-netherlands} {Flag:~Caribbean~Netherlands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f7} {flag-brazil} {brazil} {Flag:~Brazil} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f8} {flag-bahamas} {bahamas} {Flag:~Bahamas} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1f9} {flag-bhutan} {bhutan} {Flag:~Bhutan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1fb} {flag-bouvet-island} {bouvet-island} {Flag:~Bouvet~Island} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1fc} {flag-botswana} {botswana} {Flag:~Botswana} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1fe} {flag-belarus} {belarus} {Flag:~Belarus} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e7^^^^^^01f1ff} {flag-belize} {belize} {Flag:~Belize} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1e6} {flag-canada} {canada} {Flag:~Canada} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1e8} {flag-cocos-keeling-islands} {cocos-islands} {Flag:~Cocos~(Keeling)~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1e9} {flag-congo---kinshasa} {congo-kinshasa} {Flag:~Congo~-~Kinshasa} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1eb} {flag-central-african-republic} {central-african-republic} {Flag:~Central~African~Republic} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1ec} {flag-congo---brazzaville} {congo-brazzaville} {Flag:~Congo~-~Brazzaville} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1ed} {flag-switzerland} {switzerland} {Flag:~Switzerland} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1ee} {flag-cote-divoire} {cote-divoire} {Flag:~Côte~d'Ivoire} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1f0} {flag-cook-islands} {cook-islands} {Flag:~Cook~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1f1} {flag-chile} {chile} {Flag:~Chile} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1f2} {flag-cameroon} {cameroon} {Flag:~Cameroon} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1f3} {flag-china} {cn} {Flag:~China} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1f4} {flag-colombia} {colombia} {Flag:~Colombia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1f5} {flag-clipperton-island} {clipperton-island} {Flag:~Clipperton~Island} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1f7} {flag-costa-rica} {costa-rica} {Flag:~Costa~Rica} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1fa} {flag-cuba} {cuba} {Flag:~Cuba} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1fb} {flag-cape-verde} {cape-verde} {Flag:~Cape~Verde} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1fc} {flag-curacao} {curacao} {Flag:~Curaçao} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1fd} {flag-christmas-island} {christmas-island} {Flag:~Christmas~Island} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1fe} {flag-cyprus} {cyprus} {Flag:~Cyprus} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e8^^^^^^01f1ff} {flag-czechia} {czech-republic} {Flag:~Czechia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e9^^^^^^01f1ea} {flag-germany} {de} {Flag:~Germany} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1e9^^^^^^01f1ec} {flag-diego-garcia} {diego-garcia} {Flag:~Diego~Garcia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e9^^^^^^01f1ef} {flag-djibouti} {djibouti} {Flag:~Djibouti} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e9^^^^^^01f1f0} {flag-denmark} {denmark} {Flag:~Denmark} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e9^^^^^^01f1f2} {flag-dominica} {dominica} {Flag:~Dominica} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e9^^^^^^01f1f4} {flag-dominican-republic} {dominican-republic} {Flag:~Dominican~Republic} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1e9^^^^^^01f1ff} {flag-algeria} {algeria} {Flag:~Algeria} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1e6} {flag-ceuta-melilla} {ceuta-melilla} {Flag:~Ceuta~\&~Melilla} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1e8} {flag-ecuador} {ecuador} {Flag:~Ecuador} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1ea} {flag-estonia} {estonia} {Flag:~Estonia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1ec} {flag-egypt} {egypt} {Flag:~Egypt} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1ed} {flag-western-sahara} {western-sahara} {Flag:~Western~Sahara} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1f7} {flag-eritrea} {eritrea} {Flag:~Eritrea} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1f8} {flag-spain} {es} {Flag:~Spain} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1f9} {flag-ethiopia} {ethiopia} {Flag:~Ethiopia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ea^^^^^^01f1fa} {flag-european-union} {eu, european-union} {Flag:~European~Union} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1eb^^^^^^01f1ee} {flag-finland} {finland} {Flag:~Finland} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1eb^^^^^^01f1ef} {flag-fiji} {fiji} {Flag:~Fiji} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1eb^^^^^^01f1f0} {flag-falkland-islands} {falkland-islands} {Flag:~Falkland~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1eb^^^^^^01f1f2} {flag-micronesia} {micronesia} {Flag:~Micronesia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1eb^^^^^^01f1f4} {flag-faroe-islands} {faroe-islands} {Flag:~Faroe~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1eb^^^^^^01f1f7} {flag-france} {fr} {Flag:~France} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1e6} {flag-gabon} {gabon} {Flag:~Gabon} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1e7} {flag-united-kingdom} {gb, uk} {Flag:~United~Kingdom} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1e9} {flag-grenada} {grenada} {Flag:~Grenada} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1ea} {flag-georgia} {georgia} {Flag:~Georgia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1eb} {flag-french-guiana} {french-guiana} {Flag:~French~Guiana} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1ec} {flag-guernsey} {guernsey} {Flag:~Guernsey} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1ed} {flag-ghana} {ghana} {Flag:~Ghana} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1ee} {flag-gibraltar} {gibraltar} {Flag:~Gibraltar} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f1} {flag-greenland} {greenland} {Flag:~Greenland} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f2} {flag-gambia} {gambia} {Flag:~Gambia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f3} {flag-guinea} {guinea} {Flag:~Guinea} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f5} {flag-guadeloupe} {guadeloupe} {Flag:~Guadeloupe} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f6} {flag-equatorial-guinea} {equatorial-guinea} {Flag:~Equatorial~Guinea} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f7} {flag-greece} {greece} {Flag:~Greece} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f8} {flag-south-georgia-south-sandwich-islands} {south-georgia-south-sandwich-islands} {Flag:~South~Georgia~\&~South~Sandwich~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1f9} {flag-guatemala} {guatemala} {Flag:~Guatemala} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1fa} {flag-guam} {guam} {Flag:~Guam} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1fc} {flag-guinea-bissau} {guinea-bissau} {Flag:~Guinea-Bissau} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ec^^^^^^01f1fe} {flag-guyana} {guyana} {Flag:~Guyana} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ed^^^^^^01f1f0} {flag-hong-kong-sar-china} {hong-kong} {Flag:~Hong~Kong~SAR~China} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ed^^^^^^01f1f2} {flag-heard-mcdonald-islands} {heard-mcdonald-islands} {Flag:~Heard~\&~McDonald~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ed^^^^^^01f1f3} {flag-honduras} {honduras} {Flag:~Honduras} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ed^^^^^^01f1f7} {flag-croatia} {croatia} {Flag:~Croatia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ed^^^^^^01f1f9} {flag-haiti} {haiti} {Flag:~Haiti} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ed^^^^^^01f1fa} {flag-hungary} {hungary} {Flag:~Hungary} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1e8} {flag-canary-islands} {canary-islands} {Flag:~Canary~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1e9} {flag-indonesia} {indonesia} {Flag:~Indonesia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1ea} {flag-ireland} {ireland} {Flag:~Ireland} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f1} {flag-israel} {israel} {Flag:~Israel} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f2} {flag-isle-of-man} {isle-of-man} {Flag:~Isle~of~Man} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f3} {flag-india} {india} {Flag:~India} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f4} {flag-british-indian-ocean-territory} {british-indian-ocean-territory} {Flag:~British~Indian~Ocean~Territory} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f6} {flag-iraq} {iraq} {Flag:~Iraq} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f7} {flag-iran} {iran} {Flag:~Iran} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f8} {flag-iceland} {iceland} {Flag:~Iceland} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ee^^^^^^01f1f9} {flag-italy} {it} {Flag:~Italy} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1ef^^^^^^01f1ea} {flag-jersey} {jersey} {Flag:~Jersey} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ef^^^^^^01f1f2} {flag-jamaica} {jamaica} {Flag:~Jamaica} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ef^^^^^^01f1f4} {flag-jordan} {jordan} {Flag:~Jordan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ef^^^^^^01f1f5} {flag-japan} {jp} {Flag:~Japan} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1ea} {flag-kenya} {kenya} {Flag:~Kenya} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1ec} {flag-kyrgyzstan} {kyrgyzstan} {Flag:~Kyrgyzstan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1ed} {flag-cambodia} {cambodia} {Flag:~Cambodia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1ee} {flag-kiribati} {kiribati} {Flag:~Kiribati} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1f2} {flag-comoros} {comoros} {Flag:~Comoros} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1f3} {flag-st-kitts-nevis} {st-kitts-nevis} {Flag:~St.~Kitts~\&~Nevis} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1f5} {flag-north-korea} {north-korea} {Flag:~North~Korea} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1f7} {flag-south-korea} {kr} {Flag:~South~Korea} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1fc} {flag-kuwait} {kuwait} {Flag:~Kuwait} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1fe} {flag-cayman-islands} {cayman-islands} {Flag:~Cayman~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f0^^^^^^01f1ff} {flag-kazakhstan} {kazakhstan} {Flag:~Kazakhstan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1e6} {flag-laos} {laos} {Flag:~Laos} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1e7} {flag-lebanon} {lebanon} {Flag:~Lebanon} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1e8} {flag-st-lucia} {st-lucia} {Flag:~St.~Lucia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1ee} {flag-liechtenstein} {liechtenstein} {Flag:~Liechtenstein} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1f0} {flag-sri-lanka} {sri-lanka} {Flag:~Sri~Lanka} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1f7} {flag-liberia} {liberia} {Flag:~Liberia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1f8} {flag-lesotho} {lesotho} {Flag:~Lesotho} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1f9} {flag-lithuania} {lithuania} {Flag:~Lithuania} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1fa} {flag-luxembourg} {luxembourg} {Flag:~Luxembourg} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1fb} {flag-latvia} {latvia} {Flag:~Latvia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f1^^^^^^01f1fe} {flag-libya} {libya} {Flag:~Libya} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1e6} {flag-morocco} {morocco} {Flag:~Morocco} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1e8} {flag-monaco} {monaco} {Flag:~Monaco} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1e9} {flag-moldova} {moldova} {Flag:~Moldova} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1ea} {flag-montenegro} {montenegro} {Flag:~Montenegro} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1eb} {flag-st-martin} {st-martin} {Flag:~St.~Martin} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1ec} {flag-madagascar} {madagascar} {Flag:~Madagascar} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1ed} {flag-marshall-islands} {marshall-islands} {Flag:~Marshall~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f0} {flag-north-macedonia} {macedonia} {Flag:~North~Macedonia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f1} {flag-mali} {mali} {Flag:~Mali} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f2} {flag-myanmar-burma} {myanmar} {Flag:~Myanmar~(Burma)} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f3} {flag-mongolia} {mongolia} {Flag:~Mongolia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f4} {flag-macao-sar-china} {macau} {Flag:~Macao~SAR~China} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f5} {flag-northern-mariana-islands} {northern-mariana-islands} {Flag:~Northern~Mariana~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f6} {flag-martinique} {martinique} {Flag:~Martinique} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f7} {flag-mauritania} {mauritania} {Flag:~Mauritania} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f8} {flag-montserrat} {montserrat} {Flag:~Montserrat} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1f9} {flag-malta} {malta} {Flag:~Malta} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1fa} {flag-mauritius} {mauritius} {Flag:~Mauritius} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1fb} {flag-maldives} {maldives} {Flag:~Maldives} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1fc} {flag-malawi} {malawi} {Flag:~Malawi} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1fd} {flag-mexico} {mexico} {Flag:~Mexico} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1fe} {flag-malaysia} {malaysia} {Flag:~Malaysia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f2^^^^^^01f1ff} {flag-mozambique} {mozambique} {Flag:~Mozambique} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1e6} {flag-namibia} {namibia} {Flag:~Namibia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1e8} {flag-new-caledonia} {new-caledonia} {Flag:~New~Caledonia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1ea} {flag-niger} {niger} {Flag:~Niger} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1eb} {flag-norfolk-island} {norfolk-island} {Flag:~Norfolk~Island} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1ec} {flag-nigeria} {nigeria} {Flag:~Nigeria} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1ee} {flag-nicaragua} {nicaragua} {Flag:~Nicaragua} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1f1} {flag-netherlands} {netherlands} {Flag:~Netherlands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1f4} {flag-norway} {norway} {Flag:~Norway} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1f5} {flag-nepal} {nepal} {Flag:~Nepal} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1f7} {flag-nauru} {nauru} {Flag:~Nauru} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1fa} {flag-niue} {niue} {Flag:~Niue} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f3^^^^^^01f1ff} {flag-new-zealand} {new-zealand} {Flag:~New~Zealand} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f4^^^^^^01f1f2} {flag-oman} {oman} {Flag:~Oman} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1e6} {flag-panama} {panama} {Flag:~Panama} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1ea} {flag-peru} {peru} {Flag:~Peru} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1eb} {flag-french-polynesia} {french-polynesia} {Flag:~French~Polynesia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1ec} {flag-papua-new-guinea} {papua-new-guinea} {Flag:~Papua~New~Guinea} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1ed} {flag-philippines} {philippines} {Flag:~Philippines} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1f0} {flag-pakistan} {pakistan} {Flag:~Pakistan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1f1} {flag-poland} {poland} {Flag:~Poland} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1f2} {flag-st-pierre-miquelon} {st-pierre-miquelon} {Flag:~St.~Pierre~\&~Miquelon} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1f3} {flag-pitcairn-islands} {pitcairn-islands} {Flag:~Pitcairn~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1f7} {flag-puerto-rico} {puerto-rico} {Flag:~Puerto~Rico} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1f8} {flag-palestinian-territories} {palestinian-territories} {Flag:~Palestinian~Territories} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1f9} {flag-portugal} {portugal} {Flag:~Portugal} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1fc} {flag-palau} {palau} {Flag:~Palau} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f5^^^^^^01f1fe} {flag-paraguay} {paraguay} {Flag:~Paraguay} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f6^^^^^^01f1e6} {flag-qatar} {qatar} {Flag:~Qatar} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f7^^^^^^01f1ea} {flag-reunion} {reunion} {Flag:~Réunion} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f7^^^^^^01f1f4} {flag-romania} {romania} {Flag:~Romania} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f7^^^^^^01f1f8} {flag-serbia} {serbia} {Flag:~Serbia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f7^^^^^^01f1fa} {flag-russia} {ru} {Flag:~Russia} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1f7^^^^^^01f1fc} {flag-rwanda} {rwanda} {Flag:~Rwanda} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1e6} {flag-saudi-arabia} {saudi-arabia} {Flag:~Saudi~Arabia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1e7} {flag-solomon-islands} {solomon-islands} {Flag:~Solomon~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1e8} {flag-seychelles} {seychelles} {Flag:~Seychelles} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1e9} {flag-sudan} {sudan} {Flag:~Sudan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1ea} {flag-sweden} {sweden} {Flag:~Sweden} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1ec} {flag-singapore} {singapore} {Flag:~Singapore} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1ed} {flag-st-helena} {st-helena} {Flag:~St.~Helena} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1ee} {flag-slovenia} {slovenia} {Flag:~Slovenia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1ef} {flag-svalbard-jan-mayen} {svalbard-jan-mayen} {Flag:~Svalbard~\&~Jan~Mayen} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f0} {flag-slovakia} {slovakia} {Flag:~Slovakia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f1} {flag-sierra-leone} {sierra-leone} {Flag:~Sierra~Leone} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f2} {flag-san-marino} {san-marino} {Flag:~San~Marino} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f3} {flag-senegal} {senegal} {Flag:~Senegal} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f4} {flag-somalia} {somalia} {Flag:~Somalia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f7} {flag-suriname} {suriname} {Flag:~Suriname} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f8} {flag-south-sudan} {south-sudan} {Flag:~South~Sudan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1f9} {flag-sao-tome-principe} {sao-tome-principe} {Flag:~São~Tomé~\&~Príncipe} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1fb} {flag-el-salvador} {el-salvador} {Flag:~El~Salvador} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1fd} {flag-sint-maarten} {sint-maarten} {Flag:~Sint~Maarten} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1fe} {flag-syria} {syria} {Flag:~Syria} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f8^^^^^^01f1ff} {flag-eswatini} {swaziland} {Flag:~Eswatini} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1e6} {flag-tristan-da-cunha} {tristan-da-cunha} {Flag:~Tristan~da~Cunha} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1e8} {flag-turks-caicos-islands} {turks-caicos-islands} {Flag:~Turks~\&~Caicos~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1e9} {flag-chad} {chad} {Flag:~Chad} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1eb} {flag-french-southern-territories} {french-southern-territories} {Flag:~French~Southern~Territories} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1ec} {flag-togo} {togo} {Flag:~Togo} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1ed} {flag-thailand} {thailand} {Flag:~Thailand} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1ef} {flag-tajikistan} {tajikistan} {Flag:~Tajikistan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1f0} {flag-tokelau} {tokelau} {Flag:~Tokelau} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1f1} {flag-timor-leste} {timor-leste} {Flag:~Timor-Leste} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1f2} {flag-turkmenistan} {turkmenistan} {Flag:~Turkmenistan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1f3} {flag-tunisia} {tunisia} {Flag:~Tunisia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1f4} {flag-tonga} {tonga} {Flag:~Tonga} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1f7} {flag-turkey} {tr} {Flag:~Turkey} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1f9} {flag-trinidad-tobago} {trinidad-tobago} {Flag:~Trinidad~\&~Tobago} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1fb} {flag-tuvalu} {tuvalu} {Flag:~Tuvalu} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1fc} {flag-taiwan} {taiwan} {Flag:~Taiwan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1f9^^^^^^01f1ff} {flag-tanzania} {tanzania} {Flag:~Tanzania} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fa^^^^^^01f1e6} {flag-ukraine} {ukraine} {Flag:~Ukraine} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fa^^^^^^01f1ec} {flag-uganda} {uganda} {Flag:~Uganda} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fa^^^^^^01f1f2} {flag-us-outlying-islands} {us-outlying-islands} {Flag:~U.S.~Outlying~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fa^^^^^^01f1f3} {flag-united-nations} {united-nations} {Flag:~United~Nations} {4.0}
\__emoji_def:nnnnn {^^^^^^01f1fa^^^^^^01f1f8} {flag-united-states} {us} {Flag:~United~States} {0.6}
\__emoji_def:nnnnn {^^^^^^01f1fa^^^^^^01f1fe} {flag-uruguay} {uruguay} {Flag:~Uruguay} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fa^^^^^^01f1ff} {flag-uzbekistan} {uzbekistan} {Flag:~Uzbekistan} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fb^^^^^^01f1e6} {flag-vatican-city} {vatican-city} {Flag:~Vatican~City} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fb^^^^^^01f1e8} {flag-st-vincent-grenadines} {st-vincent-grenadines} {Flag:~St.~Vincent~\&~Grenadines} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fb^^^^^^01f1ea} {flag-venezuela} {venezuela} {Flag:~Venezuela} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fb^^^^^^01f1ec} {flag-british-virgin-islands} {british-virgin-islands} {Flag:~British~Virgin~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fb^^^^^^01f1ee} {flag-us-virgin-islands} {us-virgin-islands} {Flag:~U.S.~Virgin~Islands} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fb^^^^^^01f1f3} {flag-vietnam} {vietnam} {Flag:~Vietnam} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fb^^^^^^01f1fa} {flag-vanuatu} {vanuatu} {Flag:~Vanuatu} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fc^^^^^^01f1eb} {flag-wallis-futuna} {wallis-futuna} {Flag:~Wallis~\&~Futuna} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fc^^^^^^01f1f8} {flag-samoa} {samoa} {Flag:~Samoa} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fd^^^^^^01f1f0} {flag-kosovo} {kosovo} {Flag:~Kosovo} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fe^^^^^^01f1ea} {flag-yemen} {yemen} {Flag:~Yemen} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1fe^^^^^^01f1f9} {flag-mayotte} {mayotte} {Flag:~Mayotte} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ff^^^^^^01f1e6} {flag-south-africa} {south-africa} {Flag:~South~Africa} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ff^^^^^^01f1f2} {flag-zambia} {zambia} {Flag:~Zambia} {2.0}
\__emoji_def:nnnnn {^^^^^^01f1ff^^^^^^01f1fc} {flag-zimbabwe} {zimbabwe} {Flag:~Zimbabwe} {2.0}
\__emoji_subgroup:n {Subdivision-flag}
\__emoji_def:nnnnn {^^^^^^01f3f4^^^^^^0e0067^^^^^^0e0062^^^^^^0e0065^^^^^^0e006e^^^^^^0e0067^^^^^^0e007f} {flag-england} {england} {Flag:~England} {5.0}
\__emoji_def:nnnnn {^^^^^^01f3f4^^^^^^0e0067^^^^^^0e0062^^^^^^0e0073^^^^^^0e0063^^^^^^0e0074^^^^^^0e007f} {flag-scotland} {scotland} {Flag:~Scotland} {5.0}
\__emoji_def:nnnnn {^^^^^^01f3f4^^^^^^0e0067^^^^^^0e0062^^^^^^0e0077^^^^^^0e006c^^^^^^0e0073^^^^^^0e007f} {flag-wales} {wales} {Flag:~Wales} {5.0}
%%
%% End of file `emoji-table.def'.
